/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <string.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-akamaru.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

void
kiba_akamaru_init (KibaDock *dock)
{
  AkamaruObject  *anchor_object;

  dock->model.constrain_iterations = 1;

  /* add a polygon around the desktop so that the objects reflect at the borders of the desktop */
  akamaru_model_add_enclosing_rectangle (&dock->model, 0.0,  0.0, dock->geometry.x + dock->geometry.width,
                                          dock->geometry.y + dock->geometry.height);

  anchor_object = akamaru_model_add_object (&dock->model, 0, 0, 0, NULL);

  dock->anchor =
    akamaru_model_add_anchor (&dock->model, anchor_object, dock->geometry.x, dock->geometry.y);

  dock->mouse_anchor =
    akamaru_model_add_anchor (&dock->model, NULL, dock->geometry.x, dock->geometry.y);
}

void
add_spacer (AkamaruObject *object, void *data)
{
  AddSpacerClosure *closure = data;

  /* Skip the anchor object. */
  if (object != closure->dock->anchor->object && object != closure->object)
    akamaru_model_add_spacer (&closure->dock->model,
        closure->object, object, closure->spacer_size / 2.0 + object->radius);
}

void
kiba_reload_spacers (KibaDock *dock)
{
  AddSpacerClosure closure;
  GList            *g_list;

  akamaru_model_delete_spacers (&dock->model);

  for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
  {
    KibaObject *object = g_list->data;

    closure.dock        = dock;
    closure.object      = object->object;
    closure.spacer_size = object->size;

    akamaru_model_for_each_object (&dock->model, add_spacer, &closure);
  }
}

void
kiba_add_object_spring (KibaDock  *dock)
{
  DO_FOR_ALL_OBJECTS
  (
    if (object->spring == NULL)
      object->spring =
        akamaru_model_add_spring (&dock->model, dock->anchor->object, object->object, 0);
  );
}

void
kiba_delete_object_spring (KibaDock  *dock)
{
  DO_FOR_ALL_OBJECTS
  (
    if (object->spring != NULL)
    {
      akamaru_model_delete_spring (&dock->model, object->spring); 
      object->spring = NULL;
    }
  );
}

gboolean
kiba_rope_anchor (KibaDock *dock)
{
  AkamaruObject *rope_left,   *rope_right;
  double         rope_left_x,  rope_right_x,
                 rope_left_y,  rope_right_y;

  rope_left  = akamaru_model_add_object (&dock->model, 0, 0, 0, NULL);
  if (rope_left == NULL)
    return FALSE;
  rope_right = akamaru_model_add_object (&dock->model, 0, 0, 0, NULL);
  if (rope_right == NULL)
    return FALSE;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    rope_left_x  = ((dock->d->halignment - 1 == 1) ? ((double) (dock->geometry.x + dock->geometry.width)) :
                                                     (dock->geometry.x));
    rope_right_x = rope_left_x;
    rope_right_y = dock->geometry.y;
    rope_left_y  = (double) dock->geometry.y + dock->geometry.height - dgconf->icon_size;
  }
  else
  {
    rope_left_x  = dock->geometry.x;
    rope_right_x = (double) dock->geometry.x + dock->geometry.width - dgconf->icon_size;
    rope_left_y  = (double) (dock->geometry.y + (dock->geometry.height / 2)
                         + ((dock->d->valignment - 1) * (dock->geometry.height / 2)));
    rope_right_y = rope_left_y;
  }

  dock->first_rope_anchor =
    akamaru_model_add_anchor (&dock->model, rope_left, rope_left_x,  rope_left_y);
  dock->last_rope_anchor =
    akamaru_model_add_anchor (&dock->model, rope_right, rope_right_x, rope_right_y);
  if (dock->first_rope_anchor == NULL || dock->last_rope_anchor == NULL)
    return FALSE;

  return TRUE;
}

void
kiba_add_rope_spring (KibaDock *dock, double factor)
{
  KibaObject *first_object, *last_object;
  double force_left, force_right;

  if (dock->num_objects < 2)
    return;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    force_left  = factor * dock->num_objects * dgconf->rope_gravity
                / ((dock->d->valignment == 0) ? (2) : (1));
    force_right = factor * dock->num_objects * dgconf->rope_gravity
                / ((dock->d->valignment == 2) ? (2) : (1));
  }
  else
  {
    force_left  = factor * dock->num_objects * dgconf->rope_gravity
                / ((dock->d->halignment == 2) ? (2) : (1));
    force_right = factor * dock->num_objects * dgconf->rope_gravity
                / ((dock->d->halignment == 0) ? (2) : (1));
  }

  first_object = g_list_nth_data (dock->objects, 0);
  last_object  = g_list_nth_data (dock->objects, g_list_length (dock->objects) - 1);

  if (first_object->rope_spring == NULL)
    first_object->rope_spring =
      akamaru_model_add_spring (&dock->model, first_object->object, dock->last_rope_anchor->object, force_left);

  if (last_object->rope_spring == NULL)
    last_object->rope_spring =
      akamaru_model_add_spring (&dock->model, last_object->object, dock->first_rope_anchor->object, force_right);
}

void
kiba_delete_rope_spring (KibaDock *dock)
{
  KibaObject *first_object, *last_object;

  if (dock->num_objects < 2)
    return;

  first_object = g_list_nth_data (dock->objects, 0);
  if (first_object->rope_spring != NULL)
  {
    akamaru_model_delete_spring (&dock->model, first_object->rope_spring);
    first_object->rope_spring = NULL;
  }

  last_object  = g_list_nth_data (dock->objects,
                                    g_list_length (dock->objects) - 1);
  if (last_object->rope_spring != NULL)
  {
    akamaru_model_delete_spring (&dock->model, last_object->rope_spring);
    last_object->rope_spring = NULL;
  }
}

void
kiba_add_rope_strings (KibaDock *dock)
{
  KibaObject *buf_object = NULL;

  DO_FOR_ALL_OBJECTS
  (
    if (buf_object != NULL && object->rope_string == NULL)
      object->rope_string =
        akamaru_model_add_string (&dock->model, object->object,
                 buf_object->object, (ROPE_OBJECT_DISTANCE < 1.0 ?
                ((object->size + buf_object->size)/2.0) :
                ((object->size + buf_object->size)/2.0 * ROPE_OBJECT_DISTANCE)));

    buf_object = object;
  );
}

void
kiba_delete_rope_strings (KibaDock *dock)
{
  DO_FOR_ALL_OBJECTS
  (
    if (object->rope_string != NULL)
    {
      akamaru_model_delete_string (&dock->model, object->rope_string);
      object->rope_string = NULL;
    }
  );
}

void
kiba_add_rope_object (KibaDock *dock, KibaObject *object)
{
  KibaObject *first_object = NULL, *last_object = NULL;

  if (dock->num_objects < 2 || strcmp (dock->gconf->object_model, "rope") != 0)
    return;

  kiba_delete_rope_strings (dock);

  first_object  = g_list_nth_data (dock->objects, 0);

  if (dock->num_objects > 2)
  {
    last_object = g_list_nth_data (dock->objects, dock->num_objects - 2);
     /* delete rope_spring from old last launcher */
    if (last_object->rope_spring != NULL)
    {
      akamaru_model_delete_spring (&dock->model, last_object->rope_spring);
      last_object->rope_spring = NULL;
    }
  }
  else
    last_object = first_object;

  /* string from current to previos launcher */
  object->rope_string =
    akamaru_model_add_string (&dock->model, object->object, last_object->object,
                 (ROPE_OBJECT_DISTANCE < 1.0 ? (dock->gconf->icon_size * 1.0) :
                 (dock->gconf->icon_size * ROPE_OBJECT_DISTANCE)));

  kiba_add_rope_spring (dock, 1.0);
  kiba_add_rope_strings (dock);
}

void
kiba_add_wobbly_spring (KibaDock *dock)
{
  AkamaruObject *last_object = NULL;

  DO_FOR_ALL_OBJECTS
  (
    if (strcmp (dgconf->object_model, "spring_for_all") == 0)
    {
      GList *object_list;

      for (object_list = dock->objects; object_list != NULL;
           object_list = object_list->next)
      {
        KibaObject *check_object = object_list->data;

        if (check_object != object)
          object->wobbly_spring =
            g_list_append (object->wobbly_spring,
                akamaru_model_add_spring (&dock->model, object->object, check_object->object, dgconf->object_wobbly_spring/5));
      }
    }
    else if (last_object != NULL)
      object->wobbly_spring =
        g_list_append (object->wobbly_spring,
          akamaru_model_add_spring (&dock->model, object->object, last_object, -dgconf->object_wobbly_spring));

    last_object = object->object;
  );
}

void
kiba_delete_wobbly_spring (KibaDock *dock)
{
  DO_FOR_ALL_OBJECTS
  (
    GList *wobbly_list;

    for (wobbly_list = object->wobbly_spring; wobbly_list != NULL;
         wobbly_list = wobbly_list->next)
    {
      AkamaruSpring *wobbly_spring = wobbly_list->data;

      akamaru_model_delete_spring (&dock->model, wobbly_spring);
    }

    object->wobbly_spring = NULL;
  );
}

/*
void
kiba_add_spacer_polygons (KibaDock *dock)
{
  GList         *l;
  int i = 1;

  if (dgconf->spacer_size <= 1)
    return;

  for (l = dock->launchers; l != NULL; l = l->next)
  {

    if (l->next != NULL)
    {
      AkamaruPolygon *polygon;

      polygon = akamaru_model_add_rectangle (&dock->model,
        dock->allocation.x + (i + 1.35)*dgconf->icon_size + (i-1.5)*dgconf->spacer_size,
        dock->allocation.y,
        dock->allocation.x + (i + 1.35)*dgconf->icon_size + (i)*dgconf->spacer_size,
        dock->allocation.y + dgconf->spacer_height);
      dock->spacer_polygons = g_list_append (dock->spacer_polygons, polygon);
    }

    i++;
  }
}
*/
