/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <string.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <gtk/gtkprivate.h>
#include <gconf/gconf-client.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-render.h"
#include "kiba-geometry.h"
#include "kiba-events.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

#define max_velocity 50.0

static gint timeout_callback       (gpointer data);

G_DEFINE_TYPE (KibaDock, kiba_dock, GTK_TYPE_WIDGET)

GQuark kiba_error_quark (void)
{
  static GQuark q = 0;

  if (q == 0)
    q = g_quark_from_static_string ("kiba-quark");

  return q;
}

static void
kiba_dock_class_init (KibaDockClass *class)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

  widget_class->realize              = kiba_realize;
  widget_class->show                 = kiba_show;
  widget_class->map                  = kiba_map;
  widget_class->expose_event         = kiba_expose_event;
  widget_class->drag_data_received   = kiba_drag_data_received;
  widget_class->button_press_event   = kiba_button_press_event;
  widget_class->button_release_event = kiba_button_release_event;
  widget_class->scroll_event         = kiba_scroll_event;
  widget_class->motion_notify_event  = kiba_motion_notify_event;
  widget_class->enter_notify_event   = kiba_enter_notify_event;
  widget_class->leave_notify_event   = kiba_leave_notify_event;
}

static void
kiba_dock_init (KibaDock *dock)
{
  GConfClient   *client;
  GError        *error = NULL;

  static const GtkTargetEntry targets[] = {
    { "text/plain", 0, 0 }
  };

  client = gconf_client_get_default ();

  akamaru_model_init (&dock->model);

  dock->text     = g_new0 (KibaText, 1);
  dgconf         = g_new0 (KibaGconf, 1);
  dgconf->client = client;

  kiba_gconf_init_defaults (&dock->model, dgconf);

  GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (dock), GTK_NO_WINDOW);
  GTK_WIDGET_SET_FLAGS   (GTK_WIDGET (dock), GTK_TOPLEVEL);

  GTK_PRIVATE_SET_FLAG   (GTK_WIDGET (dock), GTK_ANCHORED);

  /* correct the dock placement with monitor_number */
  dgconf->icon_size = gconf_client_get_int (client, KIBA_GCONF_PATH "/options/geometry/icon_size", &error);
  if (error != NULL)
  {
    g_clear_error (&error);
    dgconf->icon_size = 0;
  }
  kiba_correct_placement (dock);

  dock->position        = KIBA_DOCK_POSITION_BOTTOM_EDGE_CENTER;
  dgconf->position      = g_strdup ("bottom-edge-center");
  dock->d               = (KibaDockLayoutData *) &kiba_layout_data[dock->position];

  /* init akamaru related stuff */
  kiba_akamaru_init (dock);

  gtk_drag_dest_set (GTK_WIDGET (dock), GTK_DEST_DEFAULT_ALL, targets, G_N_ELEMENTS (targets),
                     GDK_ACTION_MOVE | GDK_ACTION_COPY);
  gtk_drag_dest_add_uri_targets (GTK_WIDGET (dock));

  g_timeout_add (20, timeout_callback, dock);
}

static GtkWidget*
kiba_new (void)
{
  GtkWidget *dock = g_object_new (KIBA_TYPE_DOCK, NULL);

  return dock;
}

static void
kiba_realize (GtkWidget *widget)
{
  KibaDock      *dock = KIBA_DOCK (widget);
  GdkScreen     *screen;
  GdkWindowAttr  attributes;
  GList         *l;
  gint           attributes_mask;

  g_return_if_fail (KIBA_IS_DOCK (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  screen = gdk_screen_get_default ();

/*
  if (!gdk_screen_is_composited (screen))
    fprintf (stderr, "Screen has no running Composite Manager (xcompmgr, beryl ...)!\n"
                      "The dock will have uggly black windows.\n");
*/

  attributes.window_type = GDK_WINDOW_TEMP;

/* commented because of a problem with beryl, where the dock starts as a thin line
  if (dgconf->auto_hide == FALSE)
  {
    attributes.x = ((dock->d->anchor.x == 0) ? (-widget->allocation.width + 1)
                 : ((dock->d->anchor.x == 4) ?
                    (dock->geometry.width - 1) : widget->allocation.x));
    attributes.y = ((dock->d->anchor.y == 0) ? (-widget->allocation.height + 1)
                 : ((dock->d->anchor.y == 4) ?
                    (dock->geometry.height - 1) : widget->allocation.y));
  }
  else
  {
*/
    attributes.x = ((dock->d->anchor.x == 0) ? 0 : ((dock->d->anchor.x == 4) ?
                    (dock->geometry.width - widget->allocation.width) :
                     widget->allocation.x));
    attributes.y = ((dock->d->anchor.y == 0) ? 0 : ((dock->d->anchor.y == 4) ?
                    (dock->geometry.height - widget->allocation.height) :
                     widget->allocation.y));
/*
  }
*/
  widget->allocation.y  = attributes.y;
  widget->allocation.x  = attributes.x;
  attributes.width      = widget->allocation.width;
  attributes.height     = widget->allocation.height;
  attributes.wclass     = GDK_INPUT_OUTPUT;
  attributes.visual     = gdk_screen_get_rgba_visual (screen);
  attributes.colormap   = gdk_screen_get_rgba_colormap (screen);
  attributes.event_mask = gtk_widget_get_events (widget) |
    GDK_EXPOSURE_MASK |
    GDK_BUTTON_PRESS_MASK |
    GDK_BUTTON_RELEASE_MASK |
    GDK_ENTER_NOTIFY_MASK |
    GDK_LEAVE_NOTIFY_MASK |
    GDK_POINTER_MOTION_MASK |
    GDK_POINTER_MOTION_HINT_MASK;

  attributes_mask = GDK_WA_X | GDK_WA_Y;
  if (attributes.visual != NULL && attributes.colormap != NULL)
    attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gdk_screen_get_root_window (screen), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, dock);

  gdk_window_show (widget->window);

  kiba_text_create_window (dock);

  kiba_animate_position_changes (dock);

/*
  kiba_add_spacer_polygons (dock);

  if (dgconf->draw_spacer)
  {
    dock->polygon_window = kiba_polygon_create_window (dock);
    kiba_polygon_paint (dock);
  }
*/
}

/*
GdkWindow*
kiba_polygon_create_window (KibaDock     *dock)
{
  GdkWindow     *window;
  GdkWindowAttr attributes;
  gint          attributes_mask;
  GdkScreen    *screen;

  screen = gdk_screen_get_default ();

  attributes.wclass      = GDK_INPUT_OUTPUT;
  attributes.visual      = gdk_screen_get_rgba_visual (screen);
  attributes.colormap    = gdk_screen_get_rgba_colormap (screen);
  attributes.window_type = GDK_WINDOW_TEMP;
  attributes.x           = dock->geometry.x;
  attributes.y           = dock->geometry.y;
  attributes.width       = dock->geometry.width;
  attributes.height      = dock->geometry.height;
  attributes_mask        = GDK_WA_X | GDK_WA_Y;
  if (attributes.visual != NULL && attributes.colormap != NULL)
      attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;

  window = gdk_window_new (gdk_screen_get_root_window (screen),
                            &attributes, attributes_mask);

  gdk_window_set_type_hint (window, GDK_WINDOW_TYPE_HINT_DOCK);

  gdk_window_show (window);

  return window;
}
*/

GdkWindow*
kiba_create_drag_window (KibaDock *dock)
{
  GdkWindow     *window;
  GdkScreen     *screen;
  GdkWindowAttr  attributes;
  gint           attributes_mask;

  screen = gdk_screen_get_default ();

  attributes.wclass      = GDK_INPUT_OUTPUT;
  attributes.visual      = gdk_screen_get_rgba_visual (screen);
  attributes.colormap    = gdk_screen_get_rgba_colormap (screen);
  attributes.window_type = GDK_WINDOW_TEMP;
  attributes.x           = dock->drag_geometry.x;
  attributes.y           = dock->drag_geometry.y;

  if (dock->d->orientation == GTK_ORIENTATION_HORIZONTAL)
  {
    attributes.width     = dgconf->icon_size * 2;
    attributes.height    = dgconf->icon_size * 3;
  }
  else
  {
    attributes.width     = dgconf->icon_size * 3;
    attributes.height    = dgconf->icon_size * 2;
  }

  attributes_mask = GDK_WA_X | GDK_WA_Y;
  if (attributes.visual != NULL && attributes.colormap != NULL)
      attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;

  window = gdk_window_new (gdk_screen_get_root_window (screen), &attributes, attributes_mask);

  gdk_window_set_type_hint (window, GDK_WINDOW_TYPE_HINT_DOCK);

  gdk_window_show (window);

  return window;
}

static void
kiba_text_create_window (KibaDock *dock)
{
  GdkScreen     *screen;
  GdkWindowAttr  attributes;
  gint           attributes_mask;

  screen = gdk_screen_get_default ();

  attributes.wclass       = GDK_INPUT_OUTPUT;
  attributes.visual       = gdk_screen_get_rgba_visual (screen);
  attributes.colormap     = gdk_screen_get_rgba_colormap (screen);
  attributes.window_type  = GDK_WINDOW_TEMP;
  attributes.width        = 0;
  attributes.height       = 0;

  attributes_mask = GDK_WA_X | GDK_WA_Y;
  if (attributes.visual != NULL && attributes.colormap != NULL)
    attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;

  dock->text->window = gdk_window_new (gdk_screen_get_root_window (screen), &attributes, attributes_mask);
  gdk_window_set_type_hint (dock->text->window, GDK_WINDOW_TYPE_HINT_DOCK);

  dock->text->geometry_size.width  = 10;
  dock->text->geometry_size.height = 10;
  dock->text->showing              = FALSE;
  dock->text->hidden               = TRUE;
}

void
kiba_reload (KibaDock *dock)
{
  AkamaruObject  *object;
  GList          *l;
  GError         *error;

  dock->num_objects = 0;

  DO_FOR_ALL_OBJECTS (kiba_object_free (dock, object);)
  g_list_free (g_list);

  akamaru_model_fini (&dock->model);

  /* check for wrong position values */
  kiba_update_object_positions (dock);

  dock->objects            = NULL;
  dock->dragging_object    = NULL;
  dock->prelight_object    = NULL;
  dock->exec_object        = NULL;
  dock->last_text_object   = NULL;
  dock->plugins            = NULL;

  akamaru_model_init (&dock->model);

  kiba_gconf_init_defaults (&dock->model, dgconf);

  /* correct the dock placement with monitor_number */
  dgconf->icon_size = gconf_client_get_int (dgconf->client, KIBA_GCONF_PATH "/options/geometry/icon_size", &error);
  if (error != NULL)
  {
    g_clear_error (&error);
    dgconf->icon_size = 0;
  }
  kiba_correct_placement (dock);

  /* init akamaru related stuff */
  kiba_akamaru_init (dock, object);

  kiba_gconf_get_settings (dock);

  /* reload all plugins */
  for (l = dock->plugin_functions; l != NULL; l = l->next)
  {
    PluginFunc *plugin_func = l->data;

    plugin_func->init_func (dock, error);

    if (error != NULL) {
      fprintf (stderr, "Failed to load Plugin, notification %s\n", error->message);
    }
  }

  /* setup rope model */
  if (strcmp (dgconf->object_model, "rope") == 0)
  {
    if (kiba_rope_anchor (dock) == FALSE)
    {
      g_message ("failed to create anchor for rope model\n");
      return;
    }

    kiba_add_rope_spring (dock, 1.0);
    kiba_add_rope_strings (dock);
  }
  else if (strcmp (dgconf->object_model, "wobbly") == 0 || strcmp (dgconf->object_model, "spring_for_all") == 0)
    kiba_add_wobbly_spring (dock);

  /* position change is needed for a correct dock repaint
     for icon_size changes */
  KibaDockPosition position = dock->position;
  kiba_set_position (dock, (position != KIBA_DOCK_POSITION_LEFT_EDGE_CENTER ? 
     KIBA_DOCK_POSITION_LEFT_EDGE_CENTER : KIBA_DOCK_POSITION_BOTTOM_EDGE_CENTER));
  kiba_set_position (dock, position);

  if (dgconf->auto_hide && dock->hiding)
    kiba_hide_objects (dock);
}

/* initiall all plugins */
static void
kiba_init_plugins (KibaDock *dock, GError **error)
{
  GDir              *dir;
  gchar             *file_name, *dir_name;
  GSList            *sub_dirs,
                    *dirs;

  dir_name = g_strdup_printf ("/usr/lib/kiba-dock");

  if (!g_file_test (dir_name, G_FILE_TEST_IS_DIR))
    dir_name = g_strdup ("/usr/local/lib/kiba-dock");

  if (!g_file_test (dir_name, G_FILE_TEST_IS_DIR))
  {
    fprintf (stderr, "Unable to load plugins, the libdir doesnt exist\n"
             "u must install the pluginlibraries under /usr/lib/kiba-dock or /usr/local/lib/kiba-dock\n");
    exit (1);
  }

  dir = g_dir_open (dir_name, 0, error);
  if (*error != NULL)
    return;

  file_name = g_strdup_printf ("%s/%s", dir_name, (char *) g_dir_read_name (dir));
  while (!g_strrstr (file_name, "(null)"))
  {
    if (g_str_has_suffix (file_name, ".so"))
    {
      void    *handle = NULL;
      gboolean load_plugin = TRUE;

      handle = dlopen (file_name, RTLD_LAZY);

      if (!handle)
        load_plugin = FALSE;

      switch (load_plugin)
      {
        case FALSE :
        {
          fprintf (stderr, "failed to load %s\n" "%s\n", file_name, dlerror());
          break;
        }
        case TRUE :
        {
          PluginFunc *plugin_func;
          char    *dlsym_error = NULL;
          gint     notify_id;

          plugin_func = g_new0 (PluginFunc, 1);

          dlerror();    // reset errors
          typedef char* (*name_t) (void);
          name_t kiba_plugin_name = (name_t) dlsym(handle, "kiba_plugin_name");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_name': %s\n", dlsym_error);
            g_free (plugin_func);
            break;
          }

          plugin_func->plugin_name = kiba_plugin_name ();

          dlerror();    // reset errors
          typedef void   (*notify_t) (GConfClient*, guint, GConfEntry*, gpointer);
          notify_t kiba_plugin_notify = (notify_t) dlsym(handle, "kiba_plugin_notify");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_notify': %s\n", dlsym_error);
            g_free (plugin_func);
            break;
          }

          notify_id = gconf_client_notify_add (dgconf->client, g_strdup_printf (KIBA_GCONF_PATH"/plugins/%s", plugin_func->plugin_name),  kiba_plugin_notify, dock, NULL, error);
          if (*error != NULL)
          {
            g_message ("Failed to listen for " KIBA_GCONF_PATH "/plugins/%s\n" "notifications: %s\n",
                        plugin_func->plugin_name, (*error)->message);
            g_clear_error (error);
          }

          dlerror();    // reset errors
          plugin_func->init_func = (init_t) dlsym(handle, "kiba_plugin_init");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_init': %s\n", dlsym_error);
            gconf_client_notify_remove (dgconf->client, notify_id);
            g_free (plugin_func);
            break;
          }

          plugin_func->init_func (dock, *error);
          if (*error != NULL)
          {
            fprintf (stderr, "failed to load plugin %s\n" "notifications%s\n",
                      plugin_func->plugin_name, (*error)->message);
            g_clear_error (error);
            gconf_client_notify_remove (dgconf->client, notify_id);
            g_free (plugin_func->plugin_name);
            g_free (plugin_func);
            break;
          }

          if (plugin_func->plugin_name == NULL)
          {
            gconf_client_notify_remove (dgconf->client, notify_id);
            g_free (plugin_func->plugin_name);
            g_free (plugin_func);
            break;
          }

          dlerror();    // reset errors
          plugin_func->paint_func = (paint_t) dlsym(handle, "kiba_plugin_paint");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_paint': %s\n", dlsym_error);
            gconf_client_notify_remove (dgconf->client, notify_id);
            g_free (plugin_func->plugin_name);
            g_free (plugin_func);
            break;
          }

          dlerror();    // reset errors
          plugin_func->timeout_func = (timeout_t) dlsym(handle, "kiba_plugin_timeout");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_timeout': %s\n", dlsym_error);
            gconf_client_notify_remove (dgconf->client, notify_id);
            dlsym_error = NULL;
            g_free (plugin_func->plugin_name);
            g_free (plugin_func);
            break;
          }

          dlerror();    // reset errors
          plugin_func->menu_func = (menu_t) dlsym(handle, "kiba_plugin_menu");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_menu': %s\n", dlsym_error);
            gconf_client_notify_remove (dgconf->client, notify_id);
            dlsym_error = NULL;
            g_free (plugin_func->plugin_name);
            g_free (plugin_func);
            break;
          }

          dlerror();    // reset errors
          plugin_func->release_func = (release_t) dlsym(handle, "kiba_plugin_release_event");
          dlsym_error = dlerror();
          if (dlsym_error)
          {
            printf ("Cannot load symbol 'kiba_plugin_release_event': %s\n", dlsym_error);
            gconf_client_notify_remove (dgconf->client, notify_id);
            dlsym_error = NULL;
            g_free (plugin_func->plugin_name);
            g_free (plugin_func);
            break;
          }

          dock->plugin_functions = g_list_append (dock->plugin_functions, plugin_func);

          dock->plugin_handles   = g_list_append (dock->plugin_handles, handle);
        }
      }
    }

    file_name = g_strdup_printf ("%s/%s", dir_name, (char *) g_dir_read_name (dir));
  }

  /* setup rope moel */
  if (strcmp (dock->gconf->object_model, "rope") == 0)
  {
    if (kiba_rope_anchor (dock) == FALSE)
    {
      g_message ("failed to create anchor for rope model\n");
      return;
    }

    kiba_add_rope_spring (dock, 1.0);
    kiba_add_rope_strings (dock);
  }
  else if (strcmp (dock->gconf->object_model, "wobbly") == 0 ||
           strcmp (dock->gconf->object_model, "spring_for_all") == 0)
    kiba_add_wobbly_spring (dock);


  g_dir_close (dir);
  free (file_name);
  free (dir_name);
}

static gint
timeout_callback (gpointer data)
{
  KibaDock   *dock = data;
  GList      *l;
  GError     *error = NULL;
  gint        mouse_x, mouse_y, mouse_l_x, mouse_l_y;
  gboolean    fast_motion = FALSE, object_size_modified = FALSE, objects_at_dock = TRUE;

  if (dock->num_objects == 0)
    return TRUE;

  gdk_window_get_pointer (dock->widget.window, &mouse_x, &mouse_y, NULL);

  /* rerender dock bg (for animation) */
  if (dgconf->animated_background == TRUE && !g_str_has_suffix (dgconf->dock_background, ".png") &&
      dock->hiding == FALSE)
  {
    if (dock->old_bg_offset != dock->bg_offset)
      dock->rerender = TRUE;

    kiba_modify_bg_offset (dock, mouse_x, mouse_y);
  }

  /* animate dock position changes */
  if (dock->animated_placement == TRUE && dock->hiding == FALSE && dock->unhiding == FALSE)
    kiba_animate_position_changes (dock);

  for (l = dock->objects; l != NULL; l = l->next)
  {
    KibaObject *object = l->data;

    if ((dock->d->anchor.y == 4 &&
         object->object->position.y < dock->geometry.height - object->normal_size/1.5) ||
        (dock->d->anchor.y == 0 &&
         object->object->position.y > object->normal_size/1.5) ||
        (dock->d->anchor.x == 4 &&
         object->object->position.x < dock->geometry.height - object->normal_size/1.5) ||
        (dock->d->anchor.x == 0 &&
         object->object->position.x > object->normal_size/1.5))
    {
      objects_at_dock  = FALSE;
      object->at_dock = FALSE;
    }
    else
      object->at_dock = TRUE;

    if ((int) object->size != (int) object->normal_size)
      object_size_modified = TRUE;

    object->velocity_x = (object->object->position.x - object->object->previous_position.x) / 0.02;
    object->velocity_y = (object->object->position.y - object->object->previous_position.y) / 0.02;

    if (object->velocity_x < 0.0) object->velocity_x *= -1.0;
    if (object->velocity_y < 0.0) object->velocity_y *= -1.0;

    if (object->velocity_x > max_velocity || object->velocity_y > max_velocity)
      fast_motion = TRUE;

    if (object->prelight && dock->prelight_object != object)
    {
      object->prelight = FALSE;
      object->rerender = TRUE;
    }
  }

  for (l = dock->objects; l != NULL; l = l->next)
  {
    KibaObject *object = l->data;
    GList      *plugin_list;

    gdk_window_move (object->window, object->win_position.x + 0.5 - object->size / 2.0,
                                     object->win_position.y + 0.5 - object->size / 2.0);

    kiba_object_timeout (dock, object, object_size_modified, objects_at_dock, fast_motion);

    for (plugin_list = dock->plugin_functions; plugin_list != NULL; plugin_list = plugin_list->next)
    {
      PluginFunc *plugin_func = plugin_list->data;

      if (g_strrstr (plugin_func->plugin_name, object->plugin->name))
      {
        plugin_func->timeout_func (dock, object, error);

        if (object->rerender && !dock->hiding && !dock->dragging_dock)
          kiba_object_paint (dock, object, plugin_func);
      }
    }
  }

  /* calculate new object positions */
  if (objects_at_dock == FALSE || object_size_modified == FALSE ||
      dock->dragging_object != NULL || dock->zoom_exec == TRUE)
  {
    gint i;

    for (i = 0; i < 6; i++)
      akamaru_model_step (&dock->model, 0.01/dock->gconf->inertia);
  }

  /* hide/unhide the dock */
  if (dgconf->auto_hide && dock->dragging_dock == FALSE && dock->dragging_object == NULL)
    kiba_check_hide (dock, mouse_x, mouse_y, fast_motion, object_size_modified);

  /* show drag animation */
  if (dock->drag_motion == TRUE)
    kiba_drag_animation_paint (dock);

  dock->last_mouse_position.x = mouse_x;
  dock->last_mouse_position.y = mouse_y;

  if (dgconf->text_enable && dock->text->showing &&
      dock->prelight_object == NULL && dock->dragging_object == NULL)
  {
    dock->text->showing = FALSE;
    dock->text->requests++;
    g_timeout_add (dgconf->text_hide_delay, text_timeout, dock);
  }

  if (error != NULL)
    g_error_free (error);

  /* rerender dock */
  if (dock->rerender == TRUE)
    kiba_dock_paint (dock);

  return TRUE;
}

void*
kiba_get_plugin_struct (KibaDock *dock, gchar *plugin_name)
{
  GList *l;

  for (l = dock->plugins; l != NULL; l = l->next)
  {
    KibaPlugin *plugin = l->data;

    if (g_strrstr (plugin->name, plugin_name))
      return plugin;
  }

  return NULL;
}

char *
merge_desktop_to_home (KibaDock  *dock, char *system_path, char *key)
{
  char *launcher_path = NULL;

  if (system_path == NULL)
    return NULL;

  launcher_path = g_strdup_printf ("%s/.kiba-dock%s", g_get_home_dir (), g_strrstr (system_path, "/"));

  if (g_file_test (launcher_path, (G_FILE_TEST_EXISTS)) == TRUE)
    return launcher_path;

  if (strcmp (launcher_path, system_path) == 0)
  {
    GConfUnsetFlags  flags;

    g_message ("Cant find %s\nDeleting all %s related gconf values...\n", launcher_path, launcher_path);

    gconf_client_recursive_unset (dgconf->client, key, flags, NULL);

    g_free (launcher_path);

    return NULL;
  }
  else if (copyfile (system_path, launcher_path) == -1)
  {
    g_message ("Cant copy %s to %s\n", system_path, launcher_path);
    g_free (launcher_path);

    return NULL;
  }

  if (strcmp (launcher_path, system_path) != 0)
  {
    GError *error = NULL;

    gconf_client_set_string (dgconf->client, g_strdup_printf ("%s/file", key), launcher_path, &error);
    if (error != NULL)
    {
      g_message ("Failed to set file Option for launcher notifications: %s",
                 error->message);
      g_error_free (error);
    }
  }

  return launcher_path;
}

static void
launch_programm (GtkWidget* widget, gpointer data)
{
  KibaDock *dock  = data;
  GError   *error = NULL;
  char     *command;

  /* first try working dir */
  if (g_strrstr (dock->cwd, "/") != NULL)
  {
    command = g_strdup_printf ("%s/%s", g_strndup (dock->cwd, strlen (dock->cwd) -
                           strlen (g_strrstr (dock->cwd, "/"))), gtk_widget_get_name (widget));
    g_spawn_command_line_async (command, &error);
    if (error == NULL)
    {
      g_free (command);
      return;
    }
    else
      error = NULL;
  }

  /* try BINPATH */
  command = g_strdup (gtk_widget_get_name (widget));
  g_spawn_command_line_async (command, &error);
  g_free (command);
  if (error != NULL)
  {
    g_message ("failed to launch %s: %s\n", command, error->message);
    g_error_free (error);
  }
}

static int
copyfile (char *fromfile, char *tofile)
{
  FILE *ifp, *ofp;
  int   c;

  if ((ifp = fopen (fromfile, "r")) == NULL)
    return -1;
  if ((ofp = fopen (tofile, "w")) == NULL)
  {
    fclose(ifp);
    return -1;
  }

  while ((c = getc (ifp)) != EOF)
    putc(c, ofp);

  fclose (ifp);
  fclose (ofp);

  return 0;
}

void
convert_color (const char *html_color, double *r, double *g, double *b)
{
  int red,green,blue;
  sscanf (html_color, "#%02X%02X%02X", &red, &green, &blue);
  *r = (double) red   / 256;
  *g = (double) green / 256;
  *b = (double) blue  / 256;
}

static gboolean
spit_out_redraws (gpointer data)
{
  KibaDock *dock = data;

  g_print ("%d fps, %d events\n", dock->redraws, dock->events);

  dock->redraws = 0;
  dock->events  = 0;

  return TRUE;
}

void
kiba_create_menu (KibaDock *dock, KibaObject *object, GdkEventButton *event)
{
  GtkWidget *settings, *systray, *icon_editor,
            *rope_model, *dock_model, *wobbly_model, *spring_for_all_model,
            *hide, *launch_menu, *model, *separator, *quit;
  GtkMenu   *menu, *sub_menu, *model_menu;

  menu       = GTK_MENU (gtk_menu_new ());
  sub_menu   = (GtkMenu *) gtk_menu_new();
  model_menu = (GtkMenu *) gtk_menu_new();

  launch_menu = gtk_menu_item_new_with_label ("Kiba Utils");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), launch_menu);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (launch_menu), GTK_WIDGET(sub_menu));
  gtk_widget_show (launch_menu);

  settings = gtk_menu_item_new_with_label ("gset-kiba");
  gtk_widget_set_name (settings, "gset-kiba");
  gtk_menu_shell_append (GTK_MENU_SHELL (sub_menu), settings);
  g_signal_connect (G_OBJECT (settings), "activate", G_CALLBACK (launch_programm), dock);
  gtk_widget_show (settings);

  icon_editor = gtk_menu_item_new_with_label ("icon-editor");
//   if (object != NULL &&
//       g_strrstr (object->plugin->name, "launcher"))
//     gtk_widget_set_name (icon_editor,
//       g_strdup_printf ("kiba-icon-editor.py %s", object->desktop_key));
//   else
    gtk_widget_set_name (icon_editor, "kiba-icon-editor.py");

  gtk_menu_shell_append (GTK_MENU_SHELL (sub_menu), icon_editor);
  g_signal_connect (G_OBJECT (icon_editor), "activate", G_CALLBACK (launch_programm), dock);
  gtk_widget_show (icon_editor);

  systray = gtk_menu_item_new_with_label ("systray");
  gtk_widget_set_name (systray, "kiba-systray.py");
  gtk_menu_shell_append (GTK_MENU_SHELL (sub_menu), systray);
  g_signal_connect (G_OBJECT (systray), "activate", G_CALLBACK (launch_programm), dock);
  gtk_widget_show (systray);

  separator = gtk_separator_menu_item_new();
  gtk_menu_shell_append (GTK_MENU_SHELL (menu),separator);
  gtk_widget_show (separator);

  /* add plugin related menu entries */
  if (object != NULL)
  {
    GList      *plugin_list;

    for (plugin_list = dock->plugin_functions; plugin_list != NULL; plugin_list = plugin_list->next)
    {
      PluginFunc *plugin_func = plugin_list->data;

      if (object != NULL && g_strrstr (plugin_func->plugin_name, object->plugin->name))
        plugin_func->menu_func (dock, object, menu);
    }
  }

  separator = gtk_separator_menu_item_new();
  gtk_menu_shell_append (GTK_MENU_SHELL (menu),separator);
  gtk_widget_show (separator);

  model = gtk_menu_item_new_with_label ("Change Model");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), model);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (model), GTK_WIDGET(model_menu));
  gtk_widget_show (model);

  if (strcmp (dgconf->object_model, "dock") != 0)
  {
    dock_model = gtk_menu_item_new_with_label ("dock");
    gtk_menu_shell_append (GTK_MENU_SHELL (model_menu), dock_model);
    g_signal_connect (G_OBJECT (dock_model), "activate", G_CALLBACK (kiba_switch_model), "dock");
    gtk_widget_show (dock_model);
  }

  if (strcmp (dgconf->object_model, "rope") != 0)
  {
    rope_model = gtk_menu_item_new_with_label ("rope");
    gtk_menu_shell_append (GTK_MENU_SHELL (model_menu), rope_model);
    g_signal_connect (G_OBJECT (rope_model), "activate", G_CALLBACK (kiba_switch_model), "rope");
    gtk_widget_show (rope_model);
  }

  if (strcmp (dgconf->object_model, "wobbly") != 0)
  {
    wobbly_model = gtk_menu_item_new_with_label ("wobbly");
    gtk_menu_shell_append (GTK_MENU_SHELL (model_menu), wobbly_model);
    g_signal_connect (G_OBJECT (wobbly_model), "activate", G_CALLBACK (kiba_switch_model), "wobbly");
    gtk_widget_show (wobbly_model);
  }

  if (strcmp (dgconf->object_model, "spring_for_all") != 0)
  {
    spring_for_all_model = gtk_menu_item_new_with_label ("spring for all");
    gtk_menu_shell_append (GTK_MENU_SHELL (model_menu), spring_for_all_model);
    g_signal_connect (G_OBJECT (spring_for_all_model), "activate", G_CALLBACK (kiba_switch_model), "spring_for_all");
    gtk_widget_show (spring_for_all_model);
  }

  hide = gtk_check_menu_item_new_with_label ("Autohide");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), hide);
  if (dgconf->auto_hide == TRUE)
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (hide), TRUE);
  else
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (hide), FALSE);
  g_signal_connect (G_OBJECT (hide), "activate", G_CALLBACK (toggle_auto_hide), dock);
  gtk_widget_show (hide);

  separator = gtk_separator_menu_item_new();
  gtk_menu_shell_append (GTK_MENU_SHELL (menu),separator);
  gtk_widget_show (separator);

  quit = gtk_menu_item_new_with_label ("Quit");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), quit);
  g_signal_connect (G_OBJECT (quit), "activate", G_CALLBACK (gtk_main_quit), "spring_for_all");
  gtk_widget_show (quit);

  gtk_menu_popup (menu, NULL, NULL, NULL, NULL, event->button, event->time);
}

int main (int argc, char *argv[])
{
  GtkWidget   *dock;
  GConfClient *gconf_client;
  GError      *error = NULL;
  GSList      *sub_dirs, *dirs;
  char        *kiba_home_dir;
  gboolean     benchmark = FALSE, verbose = FALSE;

  gtk_init (&argc, &argv);

  int arguments = 0;
  while (arguments < argc)
  {
    if (g_strrstr (argv[arguments], "--help") != NULL || g_strrstr (argv[arguments], "-h") != NULL)
    {
      printf ("Usage: kiba-dock [--enable-tray] [--benchmark] ...\n");
      return 0;
    }
    /* enable benchmarking */
    else if (g_strrstr (argv[arguments], "--benchmark") != NULL)
      benchmark = TRUE;
    /* launch kiba-menu.py */
    else if (g_strrstr (argv[arguments], "--enable-tray") != NULL)
    {
      char *command;

      if (argv[arguments+1] == NULL)
        command = g_strdup_printf ("%s/kiba-systray.py", g_strndup (argv[0], strlen (argv[0]) -
                               strlen (g_strrstr (argv[0], "/"))));
      else
        command = g_strdup_printf ("python %s", argv[arguments+1]);

      g_spawn_command_line_async (command, &error);
      if (error != NULL)
      {
        g_clear_error (&error);
        command = g_strdup ("python kiba-systray.py");

        g_spawn_command_line_async (command, &error);
        if (error != NULL)
        {
          g_message ("failed to launch %s: %s\n", command, error->message);
          g_clear_error (&error);
        }
      }
      g_free (command);
    }
    arguments++;
  }

#ifdef HAVE_SVG
  rsvg_init ();
#endif

  kiba_home_dir = g_strdup_printf ("%s/.kiba-dock/", g_get_home_dir ());
  if (!g_file_test (kiba_home_dir, G_FILE_TEST_IS_DIR))
  {
    if (g_mkdir (kiba_home_dir, 7664) != 0)
      g_error ("failed to create kiba dir at %s\n", kiba_home_dir);
  }
  g_free (kiba_home_dir);

  gconf_client = gconf_client_get_default ();
  dock         = kiba_new ();

  (KIBA_DOCK (dock))->cwd           = argv[0];
  (KIBA_DOCK (dock))->benchmark     = benchmark;

  gconf_client_add_dir (gconf_client, KIBA_GCONF_PATH, GCONF_CLIENT_PRELOAD_NONE, &error);
  if (error != NULL)
  {
    g_error ("Failed to add gconf dir " KIBA_GCONF_PATH "\nnotifications: %s", error->message);
    g_error_free (error);
    exit (1);
  }

  /* add notify for all subdirs in /apps/kiba */
  sub_dirs = gconf_client_all_dirs (gconf_client, KIBA_GCONF_PATH "/options", &error);
  if (error != NULL)
  {
    g_error ("Failed to get subdirs from " KIBA_GCONF_PATH "/options notifications: %s", error->message);
    g_error_free (error);
    exit (1);
  }
  for (dirs = sub_dirs; dirs != NULL; dirs = dirs->next)
  {
    gconf_client_notify_add (gconf_client, dirs->data, options_changed_callback, dock, NULL, &error);
    if (error != NULL)
    {
      g_error ("Failed to listen for %s notifications: %s", (char *) dirs->data, error->message);
      g_error_free (error);
      exit (1);
    }
  }

  gtk_widget_show (dock);

  kiba_gconf_get_settings (KIBA_DOCK (dock));

  gtk_widget_show (dock);

  kiba_init_plugins (KIBA_DOCK (dock), &error);
  if (error != NULL)
  {
    g_error ("Failed to load plugins: %s", error->message);
    g_error_free (error);
    exit (1);
  }

  gdk_window_show (KIBA_DOCK (dock)->text->window);

  /* connect signals for drag animation */
  g_signal_connect (G_OBJECT (dock), "drag-motion", G_CALLBACK (kiba_drag_motion), dock);
  g_signal_connect (G_OBJECT (dock), "drag-leave", G_CALLBACK (kiba_drag_leave), dock);

  if ((KIBA_DOCK (dock))->benchmark)
    g_timeout_add (1000, spit_out_redraws, (KIBA_DOCK (dock)));

  gtk_main ();

  GList *g_list;

  for (g_list = (KIBA_DOCK (dock))->objects; g_list != NULL; g_list = g_list->next)
  {
    KibaObject *object = g_list->data;
    kiba_object_free ((KIBA_DOCK (dock)), object);
  }

  for (g_list = (KIBA_DOCK (dock))->plugin_handles; g_list != NULL; g_list = g_list->next)
  {
    void *handle = g_list->data;
    dlclose (handle);
  }

  g_list_free (g_list);

  if ((KIBA_DOCK (dock))->text->window != NULL)
    gdk_window_destroy ((KIBA_DOCK (dock))->text->window);

  if ((KIBA_DOCK (dock))->drag_window != NULL)
    gdk_window_destroy ((KIBA_DOCK (dock))->drag_window);

  if ((KIBA_DOCK (dock))->pixmap != NULL)
    gdk_pixmap_unref ((KIBA_DOCK (dock))->pixmap);

  if ((KIBA_DOCK (dock))->drag_pixmap != NULL)
    gdk_pixmap_unref ((KIBA_DOCK (dock))->drag_pixmap);

  if ((KIBA_DOCK (dock))->text->pixmap != NULL)
    gdk_pixmap_unref ((KIBA_DOCK (dock))->text->pixmap);

  if ((KIBA_DOCK (dock))->anchor != NULL)
    akamaru_model_delete_anchor (&(KIBA_DOCK (dock))->model, (KIBA_DOCK (dock))->anchor);

  if ((KIBA_DOCK (dock))->mouse_anchor != NULL)
    akamaru_model_delete_anchor (&(KIBA_DOCK (dock))->model, (KIBA_DOCK (dock))->mouse_anchor);

  if ((KIBA_DOCK (dock))->last_rope_anchor != NULL)
    akamaru_model_delete_anchor (&(KIBA_DOCK (dock))->model, (KIBA_DOCK (dock))->last_rope_anchor);

  if ((KIBA_DOCK (dock))->first_rope_anchor != NULL)
    akamaru_model_delete_anchor (&(KIBA_DOCK (dock))->model, (KIBA_DOCK (dock))->first_rope_anchor);

  akamaru_model_fini (&(KIBA_DOCK (dock))->model);

#ifdef HAVE_SVG
  rsvg_term ();
#endif

  return 0;
}
