/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <string.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-geometry.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

void
kiba_correct_placement (KibaDock *dock)
{
  GdkScreen *screen;

  screen = gdk_screen_get_default ();

  /* set screen geometry when monitor number cannot be found */
  if (gdk_screen_get_n_monitors (screen) - 1 > dgconf->monitor_number)
  {
    dock->geometry.x      = 0;
    dock->geometry.y      = 0;
    dock->geometry.width  = gdk_screen_get_width (screen);
    dock->geometry.height = gdk_screen_get_height (screen);
  }
  /* set monitor geometry */
  else
    gdk_screen_get_monitor_geometry (screen, dgconf->monitor_number, &dock->geometry);
}

void
kiba_layout (KibaDock *dock)
{
  GList     *g_list;
  int       cap_size,
            dock_lenght = 0;

  cap_size = dgconf->cap_size;

  for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
  {
    KibaObject *buf_object = g_list->data;

    dock_lenght += buf_object->size + dgconf->object_space;

    /* set object gravity point */
    if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    {
      buf_object->object->gravity.x = (dock->d->halignment - 1) * dgconf->gravity_multiplier;
      buf_object->object->gravity.y = 0;
    }
    else
    {
      buf_object->object->gravity.x = 0;
      buf_object->object->gravity.y = (dock->d->valignment - 1) * dgconf->gravity_multiplier;
    }
  }

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    dock->allocation.width  = 2*dock->gconf->icon_size / 3 + 2*dock->gconf->radius + dgconf->margin;
    dock->allocation.height = 2*cap_size + 2*dock->gconf->radius
      + ((strcmp (dgconf->object_model, "rope") == 0) ?
         (dock_lenght * ROPE_OBJECT_DISTANCE) : (dock_lenght));
  }
  else
  {
    dock->allocation.width  = 2*cap_size + 2*dock->gconf->radius
      + ((strcmp (dgconf->object_model, "rope") == 0) ?
         (dock_lenght * ROPE_OBJECT_DISTANCE) : (dock_lenght));
    dock->allocation.height = 2*dock->gconf->icon_size / 3
      + 2*dock->gconf->radius + dgconf->margin;
  }

  if (dock->allocation.width > dock->geometry.width)
    dock->allocation.width = dock->geometry.width;
  else if (dock->allocation.width <= 0)
    dock->allocation.width = 1;
  if (dock->allocation.height > dock->geometry.height)
    dock->allocation.height = dock->geometry.height;
  else if (dock->allocation.height <= 0)
    dock->allocation.height = 1;

  dock->allocation.x = dock->geometry.x +
    dock->d->halignment * (dock->geometry.width - dock->allocation.width) / 2;
  dock->allocation.y = dock->geometry.y +
    dock->d->valignment * (dock->geometry.height - dock->allocation.height) / 2;

  /* move the dock anchor (object spring is connected with it) */
  dock->anchor->x = dock->allocation.x + dock->allocation.width / 2;
  dock->anchor->y = dock->allocation.y + dock->allocation.height / 2;

  if (dock->widget.allocation.x >= 0 && dock->widget.allocation.y >= 0
      && !dock->animated_placement)
    dock->animated_placement = TRUE;
  else if (!dock->hiding && !dock->animated_placement)
  {
    gtk_widget_size_allocate (GTK_WIDGET (dock), &dock->allocation);
    gtk_widget_queue_draw (GTK_WIDGET (dock));
  }
}

void
kiba_animate_position_changes (KibaDock *dock)
{
  GtkAllocation allocation;
  gint          step_x, step_y, step_width, step_height;

  allocation = dock->widget.allocation;

  /* calculate stepsize for position changes */
  step_x = (dock->allocation.x - allocation.x) / dgconf->dock_position_stepsize;
  if (step_x == 0)
    step_x = ((dock->allocation.x - allocation.x) > 0 ? 1 : -1);

  step_y = (dock->allocation.y - allocation.y) / dgconf->dock_position_stepsize;
  if (step_y == 0)
    step_y = ((dock->allocation.y - allocation.y) > 0 ? 1 : -1);

  /* calculate new dock position */
  if (dock->allocation.x != allocation.x)
    allocation.x += step_x;
  if (dock->allocation.y != allocation.y)
    allocation.y += step_y;

  /* calculate stepsize for size changes */
  step_width = (dock->allocation.width - allocation.width) / dgconf->dock_position_stepsize;
  if (step_width == 0)
    step_width = ((dock->allocation.width - allocation.width) > 0 ? 1 : -1);

  step_height = (dock->allocation.height - allocation.height) / dgconf->dock_position_stepsize;
  if (step_height == 0)
    step_height = ((dock->allocation.height - allocation.height) > 0 ? 1 : -1);

  /* calculate new dock size */
  if (dock->allocation.width != allocation.width)
    allocation.width += step_width;
  if (dock->allocation.height != allocation.height)
    allocation.height += step_height;

  /* allocate size/position changes */
  gtk_widget_size_allocate (GTK_WIDGET (dock), &allocation);
  if (allocation.width  != dock->last_allocation.width || allocation.height != dock->last_allocation.height)
    dock->rerender = TRUE;

  if (dock->allocation.x == allocation.x && dock->allocation.y == allocation.y &&
      allocation.width == dock->allocation.width && allocation.height == dock->allocation.height)
    dock->animated_placement = FALSE;
}

void
set_position_from_string (KibaDock *dock, const char *string)
{
  int i;

  for (i = 0; i < G_N_ELEMENTS (kiba_layout_data); i++)
  {
    if (strcmp (kiba_layout_data[i].name, string) == 0)
    {
      kiba_set_position (dock, i);
      return;
    }
  }

  fprintf (stderr, "invalid position\n");
}

void
kiba_set_position (KibaDock *dock, KibaDockPosition position)
{
  dock->position = position;
  dock->d        = (KibaDockLayoutData *) &kiba_layout_data[dock->position];
  kiba_layout (dock);
}

int
kiba_find_closest_position (KibaDock *dock, int x, int y)
{
  int i, dx, dy, min_d, d, position = 0;

  min_d = INT_MAX;
  for (i = 0; i < G_N_ELEMENTS (kiba_layout_data); i++)
  {
    dx = x - dock->geometry.x
       - dock->geometry.width * kiba_layout_data[i].anchor.x / 4;
    dy = y - dock->geometry.y
       - dock->geometry.height * kiba_layout_data[i].anchor.y / 4;
    d = dx * dx + dy * dy;
    if (d < min_d)
    {
      min_d    = d;
      position = i;
    }
  }

  return position;
}

gint
kiba_hide (gpointer data)
{
  KibaDock *dock = data;

  if (!dock->hiding)
    return FALSE;

  kiba_hide_objects (dock);

  if ((dock->widget.allocation.y >= dock->geometry.height - 1) ||
      (dock->widget.allocation.y <= 1 - dock->widget.allocation.height) ||
      (dock->widget.allocation.x >= dock->geometry.width - 1) ||
      (dock->widget.allocation.x <= 1 - dock->widget.allocation.width))
  {
    gtk_widget_hide (&dock->widget);
    return FALSE;
  }

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    dock->widget.allocation.x += dock->d->halignment - 1;
  else
    dock->widget.allocation.y += dock->d->valignment - 1;

  gdk_window_move (dock->widget.window, dock->widget.allocation.x,
                                        dock->widget.allocation.y);

  return TRUE;
}

gint
kiba_unhide (gpointer data)
{
  KibaDock *dock = data;
  gint      destination;

  if (dock->hiding)
    return FALSE;

  if (dock->d->anchor.y == 4)
    destination = dock->geometry.height - dock->widget.allocation.height;
  else if (dock->d->anchor.x == 4)
    destination = dock->geometry.width - dock->widget.allocation.width;
  else
    destination = 0;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    /* show objects when dock is at half height */
    if ((dock->d->anchor.x == 4) ?
(dock->widget.allocation.x - dock->widget.allocation.width/2 <= destination) :
(dock->widget.allocation.x + dock->widget.allocation.width/2 >= destination))
      kiba_show_objects (dock);

    if ((dock->d->anchor.x == 4) ? (dock->widget.allocation.x <= destination) :
                             (dock->widget.allocation.x >= destination))
    {
      dock->unhiding = FALSE;
      return FALSE;
    }
    dock->widget.allocation.x += -1*(dock->d->halignment - 1);
  }
  else
  {
    /* show objects when dock is at half height */
    if ((dock->d->anchor.y == 4) ?
(dock->widget.allocation.y - dock->widget.allocation.height/2 <= destination) :
(dock->widget.allocation.y + dock->widget.allocation.height/2 >= destination))
      kiba_show_objects (dock);

    if ((dock->d->anchor.y == 4) ? (dock->widget.allocation.y <= destination) :
                                   (dock->widget.allocation.y >= destination))
    {
      dock->unhiding = FALSE;
      return FALSE;
    }
    dock->widget.allocation.y += -1 * (dock->d->valignment - 1);
  }

  gdk_window_move (dock->widget.window, dock->widget.allocation.x,
                                        dock->widget.allocation.y);

  return TRUE;
}

void
kiba_show (GtkWidget *widget)
{
  KibaDock *dock = KIBA_DOCK (widget);

  GTK_WIDGET_SET_FLAGS (widget, GTK_VISIBLE);

  if (!GTK_WIDGET_REALIZED (widget))
  {
    kiba_layout (dock);
    gtk_widget_realize (widget);
  }

  gtk_widget_map (widget);
}

void
kiba_map (GtkWidget *widget)
{
  KibaDock *dock = KIBA_DOCK (widget);

  GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);
  gdk_window_show (widget->window);

  if (!dock->unhiding)
    kiba_show_objects (dock);
}

void
kiba_check_hide (KibaDock *dock, int mouse_x, int mouse_y,
                 gboolean fast_motion, gboolean object_size_modified)
{
  gint width, height, min_x, min_y, max_x, max_y;

  gdk_window_get_size (dock->widget.window, &width, &height);
  min_x = -dock->gconf->auto_hide_area;
  min_y = min_x;
  max_x = width + dgconf->auto_hide_area;
  max_y = height + dgconf->auto_hide_area;

  /* just hide when objects arent in fast motion */
  if ((mouse_y < min_y || mouse_x < min_x || mouse_y > max_y || mouse_x > max_x) &&
      !dock->hiding  && !object_size_modified && !dock->zoom_exec && !fast_motion)
  {
      dock->hide_requests++;

    if (dock->hide_requests > dgconf->auto_hide_delay)
    {
      dock->hide_requests = 0;
      dock->hiding = TRUE;
      g_timeout_add (dgconf->auto_hide_speed, kiba_hide, dock);
    }
  }
  else if (mouse_x > min_x && mouse_x < max_x && mouse_y >= min_y && mouse_y <= max_y && dock->hiding)
  {
    dock->hiding   = FALSE;
    dock->unhiding = TRUE;
    gtk_widget_show (&dock->widget);
    g_timeout_add (dgconf->auto_hide_speed, kiba_unhide, dock);
  }
}

gint
text_animate_callback (gpointer data)
{
  KibaDock *dock = data;
  gint      step_x, step_y, destination_x, destination_y;

  destination_x = dock->text->geometry.x - dock->text->geometry_animate.x;
  destination_y = dock->text->geometry.y - dock->text->geometry_animate.y;

  /* move the text wndow */
  if (dock->text->geometry.x != dock->text->geometry_animate.x ||
      dock->text->geometry.y != dock->text->geometry_animate.y)
    gdk_window_move (GDK_WINDOW (dock->text->window),
                     dock->text->geometry_animate.x, dock->text->geometry_animate.y);

  /* calculate stepsize for position changes */
  step_x = (destination_x) / dgconf->text_position_stepsize;
  if (step_x == 0)
    step_x = (destination_x > 0 ? 1 : -1);
  step_y = (destination_y) / dgconf->text_position_stepsize;
  if (step_y == 0)
    step_y = (destination_y > 0 ? 1 : -1);

  /* set new text_window position */
  if (dock->text->geometry.x != dock->text->geometry_animate.x)
    dock->text->geometry_animate.x += step_x;
  if (dock->text->geometry.y != dock->text->geometry_animate.y)
    dock->text->geometry_animate.y += step_y;

  /* hide window when its off screen */
  if ((dock->text->geometry_animate.y <= (0.0 - dock->text->geometry_size.height) ||
       dock->text->geometry_animate.x <= (0.0 - dock->text->geometry_size.width) ||
       dock->text->geometry_animate.y >= dock->geometry.height ||
       dock->text->geometry_animate.x >= dock->geometry.width) &&
       gdk_window_get_state (dock->text->window) != GDK_WINDOW_STATE_WITHDRAWN)
  {
    dock->text->hide_startpoint = 0;
//     gdk_window_hide (dock->text->window);
    dock->last_text_object = NULL;
  }
  else if ((dock->text->geometry.x >= dock->geometry.width ||
            dock->text->geometry.x < 0 ||
            dock->text->geometry.y >= dock->geometry.height ||
            dock->text->geometry.y < 0) &&
            gdk_window_get_state (dock->text->window) != GDK_WINDOW_STATE_WITHDRAWN)
  {
    if (destination_x < 0) destination_x *= -1;
    if (destination_y < 0) destination_y *= -1;

    if (dock->text->hide_startpoint == 0)
      dock->text->hide_startpoint = destination_x + destination_y;

    if (dock->last_text_object != NULL)
      kiba_text_paint (dock, NULL, (double) (destination_x + destination_y) / dock->text->hide_startpoint);
  }

  return TRUE;
}

gint
text_timeout (gpointer data)
{
  KibaDock *dock = data;

  dock->text->requests--;

  if (dock->text->requests > 0)
    return FALSE;

  if (dock->dragging_object && dgconf->text_while_dragging)
  {
    dock->text->showing = FALSE;
    dock->text->requests++;
    return TRUE;
  }

  if (!dock->text->showing)
  {
    dock->text->hidden = TRUE;
    if (dock->d->anchor.y == 4)
      dock->text->geometry.y = dock->geometry.height;
    else if (dock->d->anchor.y == 0)
      dock->text->geometry.y = 0.0 - dock->text->geometry_size.height;
    else if (dock->d->anchor.x == 4)
      dock->text->geometry.x = dock->geometry.width;
    else
      dock->text->geometry.x = 0.0 - dock->text->geometry_size.width;
    dock->text->requests = 0;
  }

  return FALSE;
}

void
kiba_modify_bg_offset (KibaDock *dock, int mouse_x, int mouse_y)
{
  if (mouse_x > - (dgconf->icon_size*dgconf->zoom_factor)/2 && mouse_y > - (dgconf->icon_size*dgconf->zoom_factor)/2 &&
       mouse_x < dock->widget.allocation.width + (dgconf->icon_size*dgconf->zoom_factor)/2 &&
       mouse_y < dock->widget.allocation.height + (dgconf->icon_size*dgconf->zoom_factor)/2)
  {
    double new_offset, x_motion, y_motion;

    if (dock->bg_offset > 300.0f) dock->bg_offset = -300.0f;
    else if (dock->bg_offset < -300.0f) dock->bg_offset = 300.0f;

    x_motion   = mouse_x - dock->last_mouse_position.x;
    y_motion   = mouse_y - dock->last_mouse_position.y;
    new_offset = dock->d->orientation == GTK_ORIENTATION_VERTICAL ?
      ((mouse_y - dock->last_mouse_position.y) * dgconf->mouse_translation) :
      ((mouse_x - dock->last_mouse_position.x) * dgconf->mouse_translation);

    /* maximum stepsize at 5.0 */
    if (new_offset > 7.5) new_offset = 7.5;
    else if (new_offset < -7.5) new_offset = -7.5;

    dock->bg_offset -= new_offset;
  }
}
