/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <glib/gstdio.h>
#include <gtk/gtkprivate.h>
#include <gconf/gconf-client.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include "../akamaru/akamaru.h"

#include "kiba-dock.h"
#include "kiba-render.h"
#include "kiba-geometry.h"
#include "kiba-gconf.h"
#include "kiba-object.h"

#define max_object_velocity 50.0

void
kiba_object_free (KibaDock *dock, KibaObject *object)
{
  GList *g_list = NULL;

  if (object->spacer != NULL)
    akamaru_model_delete_spacer (&dock->model, object->spacer);

  if (object->spring != NULL)
    akamaru_model_delete_spring (&dock->model, object->spring);

  if (object->rope_spring != NULL)
    akamaru_model_delete_spring (&dock->model, object->rope_spring);

  if (object->rope_string != NULL)
    akamaru_model_delete_string (&dock->model, object->rope_string);

  if (object->object != NULL)
    akamaru_model_delete_object (&dock->model, object->object);

  for (g_list = object->wobbly_spring; g_list != NULL; g_list = g_list->next)
  {
    AkamaruSpring *spring = g_list->data;

    akamaru_model_delete_spring (&dock->model, spring);
  }

  if (object->pango_layout != NULL)
  {
    g_object_unref (object->pango_layout);
    object->pango_layout = NULL;
  }

  if (object->background_surface != NULL)
    cairo_surface_destroy (object->background_surface);

  if (object->foreground_surface != NULL)
    cairo_surface_destroy (object->foreground_surface);

  if (object->pixmap != NULL)
  {
    gdk_pixmap_unref (object->pixmap);
    object->pixmap = NULL;
  }

  if (object->pulse_pixmap != NULL)
  {
    gdk_pixmap_unref (object->pulse_pixmap);
    object->pulse_pixmap = NULL;
  }

  if (object->bg_pixbuf != NULL)
    g_object_unref (object->bg_pixbuf);

  if (object->fg_pixbuf != NULL)
    g_object_unref (object->fg_pixbuf);

  gdk_window_destroy (object->window);

  if (object->pulse_window != NULL)
    gdk_window_destroy (object->pulse_window);
}

void
kiba_object_timeout (KibaDock *dock, KibaObject *object, gboolean object_size_modified, gboolean objects_at_dock, gboolean fast_motion)
{
  gint mouse_x, mouse_y;

  /* mouse position related to object */
  gdk_window_get_pointer (object->window, &mouse_x, &mouse_y, NULL);

  /* gtk leave notify event replacement because of the pulse window */
  if (object->prelight && !object->dragging &&
     (mouse_x > object->size || mouse_x < 0 || mouse_y > object->size || mouse_y < 0))
  {
    object->prelight = FALSE;
    dock->prelight_object = NULL;

    if (strcmp (dgconf->mouse_over_effect, "zoom") != 0 || object->size == object->normal_size)
      object->rerender = TRUE;
  }

  mouse_x -= object->size / 2;
  if (mouse_x < 0) mouse_x = mouse_x * -1;
  mouse_y -= object->size / 2;
  if (mouse_y < 0) mouse_y = mouse_y * -1;

  /* rotate effect */
  if ((!object->dragging || (object->dragging && object->rotate_angle != 0.0)) &&
      ((object->exec && dgconf->exec_rotate == TRUE) ||
       (object->prelight && dgconf->mouseover_rotating == TRUE) || object->rotate_angle != 0.0 ||
       /* mouse cursor related rotation */
      (dgconf->mouse_cursor_related_rotating && mouse_y < ROTATE_AREA && mouse_x < ROTATE_AREA)))
  {
    object->rotate_angle += (object->exec) ? (dgconf->exec_rotate_velocity / 100.0) :
                            (dgconf->mouseover_rotate_velocity / 100.0);

   /* rotation over */
   if (object->at_dock && object->exec && object->drawn_rotate_angle == 0.0)
       object->exec = FALSE;

    if (object->rotate_angle >= 2.0)
      object->rotate_angle = 0.0;
  }

  /* enlightenment like pulse */
  if (strcmp (dgconf->mouse_over_effect, "pulse") == 0 &&
     (((!object->dragging || dgconf->pulse_while_dragging) &&
      object->prelight) || (object->pulse_size != object->normal_size)))
    object->rerender_pulse = TRUE;
  /* zoom the object */
  else if (strcmp (dgconf->mouse_over_effect, "zoom") == 0)
  {
    gboolean pointer_in_zoom_area = FALSE;

    if (mouse_y < (int) ZOOM_AREA && mouse_x < (int) ZOOM_AREA)
      pointer_in_zoom_area = TRUE;

    if (dock->zoom_exec == TRUE && object->exec)
    {
      object->size -= dgconf->zoom_max_stepsize;

      if (object->rotate_angle == 0.0 && object->at_dock == TRUE)
      {
        object->exec = FALSE;
        dock->zoom_exec = FALSE;

        kiba_zoom_calculate_size (dock, object, mouse_x, mouse_y, object_size_modified);

        dock->exec_object = NULL;
      }
      else
        kiba_zoom_calculate_size (dock, object, mouse_x, mouse_y, object_size_modified);
    }
    else if ((!object->exec && ((pointer_in_zoom_area == TRUE &&
       (objects_at_dock == TRUE || dock->dragging_object != NULL)) ||
       (pointer_in_zoom_area == FALSE && object_size_modified == TRUE))) ||
        object->rotate_angle != object->drawn_rotate_angle)
      kiba_zoom_calculate_size (dock, object, mouse_x, mouse_y, object_size_modified);
  }

  /* modify object win position for zoomed objects */
  if (object_size_modified && objects_at_dock == TRUE)
   kiba_change_object_win_position (dock, object, mouse_x, mouse_y);
  else
  {
    object->win_position.x = object->object->position.x;
    object->win_position.y = object->object->position.y;
  }

  /* show text */
  if (dgconf->text_enable && object->prelight &&
    ((object->velocity_x < max_object_velocity && object->velocity_y < max_object_velocity) ||
      dock->dragging_object != NULL) &&
     (dgconf->text_while_dragging ? 1 : (dock->dragging_object == NULL)))
    object->rerender_text = TRUE;

  if (object->rerender_pulse)
    kiba_pulse_paint (dock, object);

  if (object->rerender_text)
    kiba_text_paint (dock, object, 1.0);
}

void
kiba_object_create_window   (KibaDock   *dock, KibaObject *object)
{
//TODO introduce glitz, it has atm many bugs, the icons starts as black windows
//     and Xgl says goodbye when hovering an icon
  GdkScreen      *screen;
  GdkWindowAttr   attributes;
  gint            attributes_mask;

#ifdef HAVE_GLITZ
  glitz_drawable_format_t templ, *format;
  unsigned long           mask = 0;
  XVisualInfo            *vinfo = NULL;
  GdkVisual              *visual;
  GdkColormap            *colormap;
  GdkDisplay             *gdkdisplay;
  Display                *xdisplay;
  glitz_format_t          templ_f;
  Window                  xid;
#endif

  screen = gdk_screen_get_default ();

/* find draw format */
#ifdef HAVE_GLITZ
  gdkdisplay = gdk_display_get_default ();
  xdisplay   = gdk_x11_display_get_xdisplay (gdkdisplay);

  templ.doublebuffer = 1;
  mask |= GLITZ_FORMAT_DOUBLEBUFFER_MASK;

  int i = 0;
  do
  {
    format = glitz_glx_find_window_format (xdisplay,
                                           DefaultScreen (xdisplay),
                                           mask,
                                          &templ,
                                           i++);
    if (format)
    {
      vinfo = glitz_glx_get_visual_info_from_format (xdisplay,
                                                     DefaultScreen (xdisplay),
                                                     format);
      if (vinfo->depth == 32)
      {
        object->glitz_draw_format = format;
        break;
      }
      else
      {
        if (!object->glitz_draw_format)
          object->glitz_draw_format = format;
      }
    }
  } while (format);
#endif

  /* create window */
  attributes.wclass      = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_TEMP;
  attributes.x           = object->object->position.x;
  attributes.y           = object->object->position.y;
  attributes.width       = object->normal_size;
  attributes.height      = object->normal_size;
  attributes.event_mask  =
      GDK_EXPOSURE_MASK |
      GDK_BUTTON_PRESS_MASK |
      GDK_BUTTON_RELEASE_MASK |
      GDK_ENTER_NOTIFY_MASK |
      GDK_LEAVE_NOTIFY_MASK |
      GDK_POINTER_MOTION_MASK |
      GDK_POINTER_MOTION_HINT_MASK;

#ifdef HAVE_GLITZ
  if (!object->glitz_draw_format)
    fprintf (stderr, "no double buffered GLX visual\n");
  else
  {
    vinfo = glitz_glx_get_visual_info_from_format (xdisplay,
                                                   DefaultScreen (xdisplay),
                                                   object->glitz_draw_format);
    visual              = gdkx_visual_get (vinfo->visualid);
    colormap            = gdk_colormap_new (visual, TRUE);
    attributes.visual   = visual;
    attributes.colormap = colormap;
  }
#else
  attributes.visual   = gdk_screen_get_rgba_visual (screen);
  attributes.colormap = gdk_screen_get_rgba_colormap (screen);
#endif

  attributes_mask = GDK_WA_X | GDK_WA_Y;
  if (attributes.visual != NULL && attributes.colormap != NULL)
    attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;

  object->window = gdk_window_new (gdk_screen_get_root_window (screen),
                                    &attributes, attributes_mask);
  gdk_window_set_user_data (object->window, dock);
  gdk_window_set_type_hint (object->window, GDK_WINDOW_TYPE_HINT_NORMAL);

/* create a glitz drawable for this window */
#ifdef HAVE_GLITZ
  if (object->glitz_draw_format)
  {
//what to do for GdkWindows?
//    gtk_widget_set_double_buffered (launcher, FALSE);
    xid = gdk_x11_drawable_get_xid (GDK_DRAWABLE (object->window));
    object->glitz_drawable = glitz_glx_create_drawable_for_window (xdisplay,
                                 DefaultScreen (xdisplay),
                                 object->glitz_draw_format,
                                 xid, object->size,
                                 object->size);

    if (!object->glitz_drawable)
      fprintf (stderr, "failed to create drawable\n");
    else
    {
      templ_f.color        = object->glitz_draw_format->color;
      templ_f.color.fourcc = GLITZ_FOURCC_RGB;

      object->glitz_format =
         glitz_find_format (object->glitz_drawable,
                            GLITZ_FORMAT_RED_SIZE_MASK   |
                            GLITZ_FORMAT_GREEN_SIZE_MASK |
                            GLITZ_FORMAT_BLUE_SIZE_MASK  |
                            GLITZ_FORMAT_ALPHA_SIZE_MASK |
                            GLITZ_FORMAT_FOURCC_MASK,
                            &templ_f,
                            0);
      if (!object->glitz_format)
      {
        fprintf (stderr, "couldn't find surface format\n");
        object->glitz_drawable = NULL;
      }
    }
  }
#endif

  object->size                   = object->normal_size;
  object->pulse_size             = object->normal_size;
  object->rerender               = TRUE;
  object->rotate_angle           = 0.0;
  object->drawn_rotate_angle     = 0.5;
  object->dragging               = FALSE;
  object->bgsurface_needs_update = TRUE;
  object->fgsurface_needs_update = TRUE;

  /* create pango layout for object */
  if (dock->text->window)
    kiba_text_update_pango_for_object (object, dock->gconf, dock->text);

  /* create pulse window */
  object->pulse_window = kiba_object_create_pulse_window (object);
}

GdkWindow*
kiba_object_create_pulse_window (KibaObject *object)
{
  GdkWindow    *window;
  GdkWindowAttr attributes;
  gint          attributes_mask;
  GdkScreen    *screen;

  screen = gdk_screen_get_default ();

  attributes.wclass      = GDK_INPUT_OUTPUT;
  attributes.visual      = gdk_screen_get_rgba_visual (screen);
  attributes.colormap    = gdk_screen_get_rgba_colormap (screen);
  attributes.window_type = GDK_WINDOW_TEMP;

  attributes.x           = -object->normal_size / 2;
  attributes.y           = -object->normal_size / 2;
  attributes.width       = object->normal_size;
  attributes.height      = object->normal_size;
  attributes_mask        = GDK_WA_X | GDK_WA_Y;
  if (attributes.visual != NULL && attributes.colormap != NULL)
      attributes_mask |= GDK_WA_VISUAL | GDK_WA_COLORMAP;

  window = gdk_window_new (gdk_screen_get_root_window (screen),
                            &attributes, attributes_mask);

  gdk_window_set_type_hint (window, GDK_WINDOW_TYPE_HINT_DOCK);

  return window;
}

KibaObject *
kiba_get_object_for_window   (KibaDock  *dock, GdkWindow *window)
{
  GList *g_list;

  for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
  {
    KibaObject *object = g_list->data;

    if (object->window == window)
      return object;
  }

  return NULL;
}

static void
kiba_change_object_win_position (KibaDock *dock, KibaObject *object, gint mouse_x, gint mouse_y)
{
  GList        *g_list;
  GdkRectangle  offset;

  offset.y = 0;
  offset.x = 0;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
  {
    for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
    {
      KibaObject *buf_object = g_list->data;

      if (buf_object->object->position.y < object->object->position.y && buf_object != object)
        offset.y += (buf_object->size - buf_object->normal_size) / 2;
      if (buf_object->object->position.y > object->object->position.y && buf_object != object)
        offset.y -= 2*(buf_object->size - buf_object->normal_size) / 2;
     }

     if (mouse_y != dock->last_mouse_position.y)
       object->win_position.y = (double) object->object->position.y + 0.5
                                - (object->size - object->normal_size) / 2 + offset.y;
     object->win_position.x = (double) object->object->position.x + 0.5 + (dock->d->anchor.x == 0 ?
                                ((object->size - object->normal_size) / 2) :
                                (-(object->size - object->normal_size) / 2)) + offset.x;
  }
  else
  {
    for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
    {
      KibaObject *buf_object = g_list->data;

      if (buf_object->object->position.x < object->object->position.x && buf_object != object)
        offset.x += (buf_object->size - buf_object->normal_size) / 2;
      if (buf_object->object->position.x > object->object->position.x && buf_object != object)
        offset.x -= 2*(buf_object->size - buf_object->normal_size) / 2;
    }

    if (mouse_x != dock->last_mouse_position.x)
      object->win_position.x = (double) object->object->position.x + 0.5
                               - (object->size - object->normal_size) / 2 + offset.x;
    object->win_position.y = (double) object->object->position.y + 0.5 + (dock->d->anchor.y == 0 ?
                               ((object->size - object->normal_size) / 2) :
                               (-(object->size - object->normal_size) / 2)) + offset.y;
  }
}

#ifdef HAVE_SVG
gboolean
kiba_load_svg (KibaObject *object, char* path, RsvgHandle **handle, GError **error)
{
  if (path == NULL || strstr (path, ".svg") == NULL)
    return FALSE;

  if (*handle != NULL)
    rsvg_handle_free (*handle);

  if (!object->size)
    object->size = object->normal_size;

  object->buffer_size = object->normal_size;

  *handle = rsvg_handle_new_from_file (path, NULL);
  if (*error != NULL)
  {
    fprintf (stderr, "Failed to load Svg Icon for launcher %s\n" "notifications: %s\n",
              object->name, (*error)->message);
    g_clear_error (error);
    return FALSE;
  }
  if (handle != NULL)
  {
    rsvg_handle_set_size_callback (*handle, rsvg_resize_func, object, NULL);
    return TRUE;
  }

  *handle = NULL;

  return FALSE;
}
#endif

gboolean
kiba_load_png (KibaDock *dock, KibaObject *object, gchar *icon, GdkPixbuf **pixbuf, GError **error)
{
  GtkIconInfo *icon_info = NULL;
  gchar       *path, *suffix;

  if (icon == NULL)
    return FALSE;

  if (*pixbuf != NULL)
    g_object_unref (*pixbuf);

  if (!object->size)
    object->size = object->normal_size;

  object->buffer_size = object->normal_size;

  if (strstr (icon, "svg")  != NULL)
  {
    g_message ("Cant handle %s\n" "When u wanna use svgs u need svg support in cairo", object->name);

    *pixbuf = NULL;

    return FALSE;
  }

  /* try png with absolute path */
  g_clear_error (error);
  *pixbuf = gdk_pixbuf_new_from_file_at_size (icon, object->buffer_size, -1, error);
  if (*pixbuf != NULL)
    return TRUE;

  /* try png at /usr/share/pixmaps path */
  g_clear_error (error);
  path = g_strdup_printf (PIXMAP_PATH "/%s", icon);
  *pixbuf = gdk_pixbuf_new_from_file_at_size (path, object->buffer_size, -1, error);
  if (*pixbuf != NULL)
  {
    g_free (path);
    return TRUE;
  }

  path = NULL;
  g_clear_error (error);
  if (g_file_test ("/usr/share/icons/hicolor/48x48/apps", G_FILE_TEST_IS_DIR))
    path = g_strdup_printf ("/usr/share/icons/hicolor/48x48/apps/%s.png", icon);
  if (path != NULL)
  {
    *pixbuf = gdk_pixbuf_new_from_file_at_size (path, object->buffer_size, -1, error);

    if (*pixbuf != NULL)
    {
      g_free (path);
      return TRUE;
    }
  }

  /* try kde dirs */
  path = NULL;
  g_clear_error (error);
  if (g_file_test ("/usr/kde/3.5/share/icons/hicolor/48x48/apps", G_FILE_TEST_IS_DIR))
    path = g_strdup_printf ("/usr/kde/3.5/share/icons/hicolor/48x48/apps/%s.png", icon);
  else if (g_file_test ("/usr/kde/3.4/share/icons/hicolor/48x48/apps", G_FILE_TEST_IS_DIR))
    path = g_strdup_printf ("/usr/kde/3.4/share/icons/hicolor/48x48/apps/%s.png", icon);
  if (path != NULL)
  {
    *pixbuf = gdk_pixbuf_new_from_file_at_size (path, object->buffer_size, -1, error);

    if (*pixbuf != NULL)
    {
      g_free (path);
      return TRUE;
    }
  }

  if (path != NULL)
    g_free (path);

  /* try png from icon theme */
  g_clear_error (error);
  icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), icon, object->buffer_size, 0);
  if (icon_info != NULL)
  {
    *pixbuf = gdk_pixbuf_new_from_file_at_size (gtk_icon_info_get_filename (icon_info), object->buffer_size, -1, error);
    gtk_icon_info_free (icon_info);
  }
  if (*pixbuf != NULL)
    return TRUE;

  /* try png at */
  g_clear_error (error);
  suffix = strstr (icon, ".png");
  if (suffix == NULL)
    return FALSE;
  *suffix = '\0';
  icon_info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), icon, object->buffer_size, 0);
  if (icon_info != NULL)
  {
    *pixbuf = gdk_pixbuf_new_from_file_at_size (gtk_icon_info_get_filename (icon_info), object->buffer_size, -1, error);
    gtk_icon_info_free(icon_info);
  }
  if (error != NULL)
    g_clear_error (error);
  if (*pixbuf != NULL)
    return TRUE;

  return FALSE;
}

void
fill_surface_buffer (cairo_t *pSourceContext, KibaObject *object, gboolean background)
{
  cairo_surface_t* pNewSurface;
  cairo_t* drawing_context;

  /*create surface with the zoomed icon size*/
  pNewSurface =
    cairo_surface_create_similar (cairo_get_target (pSourceContext), CAIRO_CONTENT_COLOR_ALPHA,
                                  object->buffer_size, object->buffer_size);

  if (cairo_surface_status (pNewSurface) != CAIRO_STATUS_SUCCESS)
    return;

  drawing_context = cairo_create (pNewSurface);
  if (cairo_status (drawing_context) != CAIRO_STATUS_SUCCESS)
    return;
  cairo_set_source_rgba (drawing_context, 1.0f, 1.0f, 1.0f, 0.0f);
  cairo_set_operator (drawing_context, CAIRO_OPERATOR_OVER);
  cairo_paint (drawing_context);

  /* draw stuff */
#ifdef HAVE_SVG
  if ((background == TRUE ? object->bg_svg_handle : object->fg_svg_handle) != NULL)
    rsvg_handle_render_cairo ((background == TRUE ? object->bg_svg_handle :
                               object->fg_svg_handle), drawing_context);
  else if ((background == TRUE ? object->bg_pixbuf : object->fg_pixbuf) != NULL)
  {
#endif
    gdk_cairo_set_source_pixbuf (drawing_context, (background == TRUE ? object->bg_pixbuf : object->fg_pixbuf), 0, 0);
    cairo_set_operator (drawing_context, CAIRO_OPERATOR_SOURCE);
    cairo_paint (drawing_context);
#ifdef HAVE_SVG
  }
#endif

  cairo_destroy (drawing_context);

  if (background)
  {
    if (object->background_surface)
    {
      cairo_surface_destroy (object->background_surface);
      object->background_surface = NULL;
    }

    object->background_surface = pNewSurface;
  }

  if (!background)
  {
    if (object->foreground_surface)
    {
      cairo_surface_destroy (object->foreground_surface);
      object->foreground_surface = NULL;
    }

    object->foreground_surface = pNewSurface;
  }
}

void
kiba_update_surface_buffer (KibaObject *object, gboolean background)
{
  cairo_t *cr = NULL;

  cr = dock_cairo_create (object);
  if (cr == NULL)
    return;
  fill_surface_buffer (cr, object, background);
  cairo_destroy (cr);
}

void
kiba_object_paint (KibaDock *dock, KibaObject *object, PluginFunc *plugin_func)
{
  cairo_t *cairo_context = NULL;

  object->rerender = FALSE;

  if (gdk_window_get_state (object->window) == GDK_WINDOW_STATE_WITHDRAWN)
    return;

  #ifdef HAVE_GLITZ
  if (object->glitz_drawable)
    glitz_drawable_update_size (object->glitz_drawable, object->size, object->size);
  #endif

  cairo_context = dock_cairo_create (object);
  if (cairo_context == NULL)
    return;
  cairo_set_operator (cairo_context, CAIRO_OPERATOR_CLEAR);
  cairo_paint (cairo_context);

  cairo_scale (cairo_context, object->size / object->normal_size, object->size / object->normal_size);

  cairo_save (cairo_context);

  if (object->rotate_angle != 0.0)
    kiba_rotate_object (dock, object, cairo_context);

  plugin_func->paint_func (dock, object, cairo_context);

  object->drawn_rotate_angle = object->rotate_angle;

  #ifdef HAVE_GLITZ
  if (object->glitz_draw_format && object->glitz_draw_format->doublebuffer)
      glitz_drawable_swap_buffers (object->glitz_drawable);
  else if (object->glitz_draw_format)
      glitz_drawable_flush (object->glitz_drawable);
  #else

  /* update window with offscreen pixmap content */
  if (object->pixmap != NULL)
    gdk_draw_drawable  (object->window, object->gc, object->pixmap,
                        0, 0, 0, 0, object->size, object->size);
/*
  update_input_shape_from_cairo (object->window, cairo_get_target (cairo_context),
                                 object->size, object->size, FALSE);
*/

  #endif

  cairo_destroy (cairo_context);

  if (object->prelight && dgconf->mouseover_bouncing_height > 0.0)
  {
    if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
      object->object->position.x += (dock->d->halignment - 1) * dgconf->mouseover_bouncing_height * -1;
    else
      object->object->position.y += (dock->d->valignment - 1) * dgconf->mouseover_bouncing_height * -1;
  }

  if (dock->benchmark)
    dock->redraws++;
}

void
kiba_pulse_paint (KibaDock *dock, KibaObject *object)
{
  if (object->pulse_window == NULL ||
      gdk_window_get_state (object->window) == GDK_WINDOW_STATE_WITHDRAWN)
    return;

  if (gdk_window_get_state (object->pulse_window) == GDK_WINDOW_STATE_WITHDRAWN)
    gdk_window_show (object->pulse_window);

  if (gdk_window_get_state (object->pulse_window) ==
      GDK_WINDOW_STATE_WITHDRAWN && dock->exec_pulse == FALSE &&
      object->prelight == TRUE)
    object->pulse_size = object->normal_size;
  else if (gdk_window_get_state (object->pulse_window) ==
           GDK_WINDOW_STATE_WITHDRAWN && object->pulse_exec == TRUE)
  {
    object->pulse_exec = FALSE;
    object->pulse_size = object->normal_size;
  }

  cairo_t *pixmap_context = NULL;

  /* modify pulse window geometry */
  gdk_window_resize (object->pulse_window, object->pulse_size, object->pulse_size);

  gdk_window_move (object->pulse_window,
    object->object->position.x - object->pulse_size / 2.0,
    object->object->position.y - object->pulse_size / 2.0);

  /* update pixmap size */
  if (object->pulse_pixmap != NULL)
  {
    gdk_pixmap_unref (object->pulse_pixmap);
    object->pulse_pixmap = NULL;
  }
  object->pulse_pixmap = create_pixmap (object->pulse_size, object->pulse_size);
  if (object->pulse_pixmap == NULL)
    return;

  if (object->pulse_gc)
  {
    gdk_gc_unref (object->pulse_gc);
    object->pulse_gc = NULL;
  }
  object->pulse_gc = gdk_gc_new (object->pulse_pixmap);

  pixmap_context = gdk_cairo_create (object->pulse_pixmap);
  if (pixmap_context == NULL)
    return;

  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_CLEAR);
  cairo_paint (pixmap_context);

  double scale_factor = object->pulse_size / object->size;

  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    cairo_translate (pixmap_context,
        (object->width < object->height ?
        (scale_factor + dock->d->halignment - 1)*(object->height - object->width)/2.0 : 0.0),
        (object->width > object->height ?  scale_factor*(object->width - object->height)/2.0 : 0.0));
  else
    cairo_translate (pixmap_context,
        (object->width < object->height ?  scale_factor*(object->height - object->width)/2.0 : 0.0),
        (object->width > object->height ?
        (scale_factor + (double) dock->d->valignment - 1.0)*(object->width - object->height)/2.0 : 0.0));

  cairo_scale (pixmap_context, object->pulse_size / object->normal_size,
    object->pulse_size / object->normal_size);

  /* rotate the object */
//TODO rotation and pulse doesnt work
  double offset_x, offset_y;
  if (object->rotate_angle != 0.0)
  {
    cairo_translate (pixmap_context, object->buffer_size/2.0, object->buffer_size/2.0);
    cairo_rotate (pixmap_context, M_PI * object->rotate_angle);

    offset_x = -object->buffer_size/2.0;
    offset_y = offset_x;
  }
  else
  {
    offset_x = 0.0;
    offset_y = 0.0;
  }

  cairo_set_source_surface (pixmap_context, object->background_surface, offset_x, offset_y);
  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_SOURCE);
  cairo_paint_with_alpha (pixmap_context, (dgconf->pulse_max_alpha
            * (MAX_PULSE_SIZE - object->pulse_size)) / (MAX_PULSE_SIZE - object->normal_size) - 0.1);

  if (object->prelight || dock->exec_pulse == TRUE)
  {
    double r,g,b;
    convert_color (dgconf->prelight_color, &r, &g, &b);
    cairo_set_source_rgba (pixmap_context, r, g, b, dgconf->prelight_alpha);
    cairo_set_operator (pixmap_context, CAIRO_OPERATOR_ATOP);
    cairo_paint (pixmap_context);
  }
  cairo_destroy (pixmap_context);

  /* update window with offscreen pixmap */
  gdk_draw_drawable  (object->pulse_window, object->pulse_gc, object->pulse_pixmap,
                      0, 0, 0, 0, object->pulse_size, object->pulse_size);

  /* shape pulse window complete */
  update_input_shape_from_cairo (object->pulse_window, NULL,
    object->pulse_size, object->pulse_size, TRUE);

  object->pulse_size += (object->pulse_exec == TRUE ? dgconf->pulse_exec_stepsize : dgconf->pulse_stepsize);

  if (object->pulse_size > MAX_PULSE_SIZE)
  {
    object->rerender_pulse = FALSE;

    object->pulse_size = object->normal_size;

    if (dock->exec_pulse == TRUE)
      dock->exec_pulse = FALSE;
  }
}

gboolean
kiba_text_paint (KibaDock *dock, KibaObject *object, double fade_factor)
{
  PangoRectangle  ink, log;
  gboolean        render_text;
  double          x_displacement, y_displacement,
                  margin, radius,
                  r, g, b;
  int             w_x, w_y;

  /* ignore text paint calls while they come from animation callback */
  if (dock->text->hide_startpoint != 0 && fade_factor == 1.0)
    return FALSE;

  if (object == NULL)
  {
    object = dock->last_text_object;
    render_text = TRUE;
  }
  /* assume rendering is not needed */
  else
    render_text = FALSE;

  object->rerender_text = FALSE;

  /* just show text when object is shown */
  if (gdk_window_get_state (object->window) == GDK_WINDOW_STATE_WITHDRAWN &&
      !dock->hiding)
    return FALSE;

//   if (gdk_window_get_state (dock->text->window) == GDK_WINDOW_STATE_WITHDRAWN)
//     gdk_window_show (dock->text->window);

  if (object->pango_layout == NULL)
    return FALSE;

  pango_layout_get_pixel_extents (object->pango_layout, &ink, &log);
  w_x = log.width + 10 + 2*dock->gconf->text_border_width;
  w_y = log.height + 2 + 2*dock->gconf->text_border_width;

  if (fade_factor == 1.0)
  {
    double x0, y0;

    dock->last_text_object = object;

    /* resize window and store window size w_x, w_y */
    if (w_x != dock->text->geometry_size.width ||
        w_y != dock->text->geometry_size.height)
    {
      render_text = TRUE;           //rerender
      dock->text->geometry_size.width  = w_x;
      dock->text->geometry_size.height = w_y;
      gdk_window_resize (GDK_WINDOW (dock->text->window),
        dock->text->geometry_size.width, dock->text->geometry_size.height);
    }

    if (strcmp (dgconf->text_placement_x, "left") == 0)
      x_displacement = dock->text->geometry_size.width;
    else if (strcmp (dgconf->text_placement_x, "right") == 0)
      x_displacement = 0;
    else
      x_displacement = dock->text->geometry_size.width/2;

    if (strcmp (dgconf->text_placement_y, "top") == 0)
      y_displacement = dock->text->geometry_size.height;
    else if (strcmp (dgconf->text_placement_y, "bottom") == 0)
      y_displacement = 0;
    else
      y_displacement = dock->text->geometry_size.height/2;

    /* calculate new position */
    if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    {
      x0 = object->object->position.x + (dock->d->halignment - 1) *
           (object->normal_size - object->size) / 2.0 - (dock->d->halignment - 1) * object->size / 2.0 +
          ((dock->d->halignment - 1 == 1) ? -dock->text->geometry_size.width : 0) +
          -(dock->d->valignment - 1) * dgconf->text_icon_padding;
      y0 = object->object->position.y + (object->normal_size - object->size) / 2.0 - y_displacement;
    }
    else
    {
      x0 = object->object->position.x + (object->normal_size - object->size) / 2.0 - x_displacement;
      y0 = object->object->position.y + (dock->d->valignment - 1) *
           (object->normal_size - object->size) / 2.0 - (dock->d->valignment - 1) * object->size / 2.0 +
          ((dock->d->valignment - 1 == 1) ? -dock->text->geometry_size.height : 0) +
          -(dock->d->valignment - 1) * dgconf->text_icon_padding;
    }

    /* check if textwindow is on screen */
    if (x0 < 0) x0 = 0;
    else if (x0 > (dock->geometry.x + dock->geometry.width - dock->text->geometry_size.width))
      x0 = dock->geometry.x + dock->geometry.width - dock->text->geometry_size.width;
    if (y0 < 0) y0 = 0;
    else if (y0 > (dock->geometry.y + dock->geometry.height - dock->text->geometry_size.height))
      y0 = dock->geometry.y + dock->geometry.height - dock->text->geometry_size.height;

    dock->text->geometry.x = x0;
    dock->text->geometry.y = y0;

    /* move GDK window */
    if (dock->text->animate == FALSE)
    {
      dock->text->animate              = TRUE;
      dock->text->animation_stepsize.x = 5.0;
      dock->text->animation_stepsize.y = 5.0;

      gdk_window_move (GDK_WINDOW (dock->text->window), dock->geometry.width, dock->geometry.height);

      if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
      {
        if (dock->d->anchor.x == 0)
        {
          dock->text->geometry_animate.x = (double) - dock->text->geometry_size.width;
        }
        else
          dock->text->geometry_animate.x = (double) dock->geometry.width;
        dock->text->geometry_animate.y   = (double) dock->text->geometry.y;
      }
      else
      {
        if (dock->d->anchor.y == 4)
          dock->text->geometry_animate.y = (double) dock->geometry.height;
        else
          dock->text->geometry_animate.y = -dock->text->geometry_size.height;
        dock->text->geometry_animate.x   = (double) dock->text->geometry.x;
      }
      g_timeout_add (20, text_animate_callback, dock);
    }

    if (dock->text->hidden)
    {
      dock->text->hidden = FALSE;
      render_text        = TRUE;
//       gdk_window_show (dock->text->window);
    }
  }

  if (dock->text->rendered_text && strcmp (object->name, dock->text->rendered_text) != 0)
    render_text = TRUE;

  if (render_text)
  {
    cairo_t *cr = NULL;
    double   x0, y0,
             x1, y1;

    dock->text->showing = TRUE;

    if (dock->text->pixmap != NULL)
    {
      gdk_pixmap_unref (dock->text->pixmap);
      object->pixmap = NULL;
    }
    dock->text->pixmap = create_pixmap (w_x, w_y);
    if (dock->text->pixmap == NULL)
      return FALSE;

    if (dock->text->gc)
    {
      gdk_gc_unref (dock->text->gc);
      dock->text->gc = NULL;
    }
    dock->text->gc     = gdk_gc_new (dock->text->pixmap);
    if (dock->text->gc == NULL)
      return FALSE;

    cr = gdk_cairo_create (dock->text->pixmap);
    if (cr == NULL)
      return FALSE;

    /* clear canvas */
    cairo_set_source_rgba (cr, 1.0f, 1.0f, 1.0f, 0.0f);
    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    cairo_paint (cr);

    /* border points */
    margin = dgconf->text_border_width;
    x0   = margin / 2;
    y0   = dock->text->geometry_size.height - margin / 2;
    x1   = dock->text->geometry_size.width - margin / 2;
    y1   = margin / 2;
    radius = ((double) log.height / 3);

    /* draw dock shape */
    cairo_save (cr);
    kiba_paint_shape (dock, cr,
         dgconf->text_border_style, dgconf->text_border_style,
         dgconf->text_border_style, dgconf->text_border_style,
         FALSE, TRUE, x0, y0, x1, y1, radius, margin);

    /* stroke border */
    convert_color (dgconf->text_border_color, &r, &g, &b);
    cairo_set_source_rgba (cr, r, g, b, dgconf->text_border_alpha * fade_factor);
    cairo_set_line_width (cr, margin);
    cairo_fill_preserve (cr);
    cairo_stroke (cr);

    cairo_pattern_t *pattern = NULL;
    /* png background */
    if (g_str_has_suffix (dgconf->text_background, ".png"))
      pattern = kiba_get_png_pattern (dock, cr, dgconf->text_background, FALSE);
    /* gradient backround */
    else if (strcmp (dgconf->dock_background, "gradient") == 0)
      pattern = kiba_get_gradient_pattern (dock,
                  dgconf->text_bg_color_1, dgconf->text_bg_color_2,
                  dgconf->text_bg_color_3,
                  dgconf->text_bg_alpha_1 * fade_factor,
                  dgconf->text_bg_alpha_2 * fade_factor,
                  dgconf->text_bg_alpha_3 * fade_factor,
                  x0, x1, y0, y1);
    else
      pattern = kiba_get_bg_pattern (dock,
                  strcmp (dgconf->text_background, "linear") == 0 ?
                  TRUE : FALSE,
                  dgconf->text_bg_color_1, dgconf->text_bg_color_2,
                  dgconf->text_bg_color_3, dgconf->text_bg_color_4,
                  dgconf->text_bg_color_5,
                  dgconf->text_bg_alpha_1 * fade_factor,
                  dgconf->text_bg_alpha_2 * fade_factor,
                  dgconf->text_bg_alpha_3 * fade_factor,
                  dgconf->text_bg_alpha_4 * fade_factor,
                  dgconf->text_bg_alpha_5 * fade_factor,
                  dgconf->text_bg_affection,
                  dgconf->dock_bg_step_size / 10.0,
                  dgconf->text_bg_radius_1,
                  dgconf->text_bg_radius_2);

    if (pattern == NULL)
    {
      cairo_destroy (cr);
      return FALSE;
    }

    cairo_save (cr);
      cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
      cairo_set_source (cr, pattern);
      cairo_paint (cr);
      cairo_pattern_destroy (pattern);
    cairo_restore (cr);
    cairo_restore (cr);

    /* create textshadow */
    cairo_save (cr);
    convert_color (dgconf->text_shadow_color, &r, &g, &b);
    cairo_set_source_rgba (cr, r, g, b, dgconf->text_shadow_alpha * fade_factor);
    cairo_translate (cr, margin + 5 + dgconf->text_shadow_offset, margin + 1 + dgconf->text_shadow_offset);
    pango_cairo_update_layout(cr, object->pango_layout);
    pango_cairo_show_layout (cr, object->pango_layout);

    /* layout text */
    cairo_restore (cr);
    convert_color (dgconf->text_color, &r, &g, &b);
    cairo_set_source_rgba (cr, r, g, b, dgconf->text_alpha * fade_factor);
    cairo_translate (cr, margin+5, margin+1);  //+5 & +1 for the +10 & +2 in w_x  & w_y above
    pango_cairo_update_layout (cr, object->pango_layout);
    pango_cairo_show_layout (cr, object->pango_layout);

    /* shape the text window */
    update_input_shape_from_cairo (dock->text->window, NULL,
       dock->text->geometry_size.width, dock->text->geometry_size.height, TRUE);

    /* update window with offscreen pixmap */
    gdk_draw_drawable  (dock->text->window, dock->text->gc, dock->text->pixmap,
                        0, 0, 0, 0, dock->text->geometry_size.width, dock->text->geometry_size.height);

    cairo_destroy (cr);

    dock->text->rendered_text = object->name;
  }

  return TRUE;
}

void
kiba_text_update_pango_for_object (KibaObject *object, KibaGconf *gconf, KibaText *text)
{
  cairo_t               *cr = NULL;
  cairo_font_options_t  *font_options;
  PangoFontDescription  *font_description;
  PangoStyle             font_style;

  cr = gdk_cairo_create (text->window);
  if (!cr || object->name == NULL)
    return;

  if (object->pango_layout != NULL)
  {
    g_object_unref (object->pango_layout);
    object->pango_layout = NULL;
  }

  text->last_size = gconf->text_size;

  object->pango_layout = pango_cairo_create_layout (cr);
  font_options           = cairo_font_options_create ();
  font_description       = pango_font_description_new ();

  cairo_font_options_set_hint_metrics (font_options, CAIRO_HINT_METRICS_ON);
  cairo_font_options_set_hint_style (font_options, CAIRO_HINT_STYLE_NONE);
  cairo_font_options_set_subpixel_order (font_options, CAIRO_SUBPIXEL_ORDER_RGB);
  cairo_font_options_set_antialias (font_options, CAIRO_ANTIALIAS_SUBPIXEL);
  pango_cairo_context_set_font_options (pango_layout_get_context
                                         (object->pango_layout), font_options);
  cairo_font_options_destroy (font_options);
  pango_font_description_set_absolute_size (font_description,
               gconf->text_size* PANGO_SCALE);
  pango_font_description_set_family_static (font_description,
                                            gconf->text_type);

  if (g_strrstr (gconf->text_weight, "old"))
    font_style = PANGO_STYLE_OBLIQUE;
  else if (g_strrstr (gconf->text_weight, "talic"))
    font_style = PANGO_STYLE_ITALIC;
  else
    font_style = PANGO_STYLE_NORMAL;
  pango_font_description_set_style (font_description, font_style);

  pango_layout_set_font_description (object->pango_layout, font_description);
  pango_font_description_free (font_description);

  /* we need to scale the content before we set the pango layout */
  if (text->geometry.height <
      text->geometry.width)
    cairo_scale (cr, 1.0, text->geometry.height 
                        / text->geometry.width);
  else if (text->geometry.height >
           text->geometry.width)
    cairo_scale (cr, text->geometry.width 
                   / text->geometry.height , 1.0);

  /* set Pango text */
  pango_layout_set_text (object->pango_layout, object->name,-1);

  cairo_destroy (cr);
}

void
kiba_rotate_object (KibaDock *dock, KibaObject *object, cairo_t*pixmap_context)
{
  double angle;

  if (dgconf->rotate_scale_factor != 1.0)
    cairo_scale (pixmap_context, object->scale_factor, object->scale_factor);

  cairo_translate (pixmap_context, (2.0 - object->scale_factor)*object->normal_size / 2.0,
                   (2.0 - object->scale_factor)*object->normal_size / 2.0 );
  cairo_rotate (pixmap_context, M_PI * object->rotate_angle);

  if (dgconf->rotate_scale_factor != 1.0)
  {
    angle = object->rotate_angle - 1.0;
    if (angle < 0.0) angle *= -1.0;

    object->scale_factor = (1.0 - dgconf->rotate_scale_factor)*(angle) + dgconf->rotate_scale_factor;
  }
}

void
kiba_zoom_calculate_size (KibaDock *dock, KibaObject *object,
                          int mouse_l_x, int mouse_l_y, gboolean object_size_modified)
{
  double new_size;

  /* calculate new icon_size depending on the mousecursor position */
  if (mouse_l_y < ZOOM_AREA && mouse_l_x < ZOOM_AREA &&
      (dock->dragging_object == NULL || dock->dragging_object != NULL &&
       object->rotate_angle) && !dock->zoom_exec)
  {
    double mouse_l_x2 = ((double) mouse_l_x - ZOOM_AREA) * -1;
    double mouse_l_y2 = ((double) mouse_l_y - ZOOM_AREA) * -1;

    new_size = object->normal_size * (1.0 + ((dgconf->zoom_factor - 1.0)
//TODO avoid oscillating new_size values, this just avoids it for prelighted object, probably a better algorythm is needed
* (object->prelight ? 1.0 : ((mouse_l_x2 < mouse_l_y2 ? mouse_l_x2 : mouse_l_y2)/ZOOM_AREA))));
  }
  /* calculate new icon_size depending on gconf->zoom_max_stepsize */
  else
    new_size = (object->size >= (object->normal_size
              + dgconf->zoom_max_stepsize) ?
               (object->size - dgconf->zoom_max_stepsize) :
                object->normal_size);

  /* start zooming just when a object is prelighted */
  if (dock->prelight_object == NULL && !object_size_modified)
    return;

//   if (new_size < object->size)
//     object->zoom_out = TRUE;
//   else if (new_size > object->size)
//     object->zoom_out = FALSE;
// 
//   if (new_size < object->size && object->zoom_out = FALSE &&

  if (new_size != object->size || object->rotate_angle != 0.0 ||
      object->rotate_angle != object->drawn_rotate_angle)
  {
    /* avoid big stepsizes for zooming */
    if (new_size >= object->size + dgconf->zoom_max_stepsize)
      new_size = object->size + dgconf->zoom_max_stepsize;
    else if (new_size <= object->size - dgconf->zoom_max_stepsize)
      new_size = object->size - dgconf->zoom_max_stepsize;

//     object->old_size = object->size;
    object->size = new_size;

    gdk_window_resize (object->window, object->size, object->size);

    /* repaint the object with new size */
    object->rerender = TRUE;

    /* correct dock width */
    kiba_layout (dock);
  }
}

void
kiba_object_position_changed (KibaDock *dock, KibaObject *object, GConfEntry *entry)
{
  GError *error = NULL;
  int new_position, offset;

  if (dock->changing_position == TRUE)
    return;
/*
  dock->changing_position = TRUE;

  if (strcmp (dgconf->object_model, "rope") == 0)
  {
    kiba_delete_rope_strings (dock);
    kiba_delete_rope_spring (dock);
  }
  else if (strcmp (dgconf->object_model, "wobbly") == 0)
    kiba_delete_wobbly_spring (dock);

  new_position = gconf_client_get_int (dock->gconf->client, entry->key, &error);
  if (new_position > g_list_length (dock->objects) || new_position < 1)
    {
      if (new_position > g_list_length (dock->objects))
        new_position = g_list_length (dock->objects);
      else if (new_position < 1)
        new_position = 1;

      gconf_client_set_int (dgconf->client, g_strdup_printf ("%s/%s_position", object->key, object->plugin->name), new_position, &error);
      if (error != NULL)
      {
        g_message ("failed to create position key for %s notification: %s",
                    object->name, error->message);
        error = NULL;
      }
    }
    offset = new_position - object->position;

    while (offset != 0)
    {
      KibaObject *buf_object;

      buf_object = g_list_nth_data (dock->objects, object->position + offset - 1);

      buf_object->position += (offset > 0 ? -1 : 1);

      gconf_client_set_int (dgconf->client,
        g_strdup_printf ("%s/%s_position", buf_object->key, buf_object->plugin->name), buf_object->position, &error);
      if (error != NULL)
      {
        g_message ("failed to create position key for %s notification: %s",
                    buf_object->name, error->message);
        error = NULL;
      }
      offset += (offset > 0 ? -1 : 1);
    }
    object->position = new_position;

    kiba_reorder_objectlist (dock);

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_add_rope_spring (dock, 1.0);
      kiba_add_rope_strings (dock);
    }
    else if (strcmp (dgconf->object_model, "wobbly") == 0 ||
             strcmp (dgconf->object_model, "spring_for_all") == 0)
      kiba_add_wobbly_spring (dock);

    dock->changing_position = FALSE;
*/
}


void
kiba_object_update_position (KibaDock *dock, KibaObject *object, gboolean reload)
{
/*
  object->position = gconf_client_get_int (dgconf->client, g_strdup_printf ("%s/%s_position", object->key, object->plugin->name) , NULL);
  if (object->position == 0)
  {
    GError *error = NULL;

    object->position = dock->num_objects;

    gconf_client_set_int (dgconf->client,
      g_strdup_printf ("%s/%s_position", object->key, object->plugin->name) , object->position, &error);
    if (error != NULL)
    {
      g_message ("failed to create position key for %s notification: %s",
                  object->name, error->message);
      g_error_free (error);
    }
  }
  else
    kiba_reorder_objectlist (dock);

  if (reload == FALSE && dock->num_objects > 1)
  {
    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      KibaObject *first_object = NULL, *last_object = NULL;

      first_object  = g_list_nth_data (dock->objects, 0);

      if (dock->num_objects > 2)
      {
        last_object = g_list_nth_data (dock->objects, dock->num_objects - 2);

        if (last_object->rope_spring != NULL)
        {
          akamaru_model_delete_spring (&dock->model, last_object->rope_spring);
          last_object->rope_spring = NULL;
        }
      }
      else
        last_object = first_object;


      object->rope_string =
        akamaru_model_add_string (&dock->model, object->object, last_object->object,
                     (ROPE_OBJECT_DISTANCE < 1.0 ? (object->normal_size * 1.0) :
                     (object->normal_size * ROPE_OBJECT_DISTANCE)));

      kiba_add_rope_spring (dock, 1.0);
    }
  }
*/
}

void
kiba_update_rope_positions (KibaDock *dock, KibaObject *object)
{
  KibaObject    *next_object = NULL,
                *prev_object = NULL;
  int position = g_list_index (dock->objects, object);

  kiba_delete_rope_spring (dock);

  if (position > 0)
    prev_object = g_list_nth_data (dock->objects, position - 1);
  if (position < g_list_position (dock->objects, g_list_last (dock->objects)))
    next_object = g_list_nth_data (dock->objects, position + 1);

  if (prev_object != NULL)
  {
    if (object->rope_string != NULL)
    {
      akamaru_model_delete_string (&dock->model, object->rope_string);
      object->rope_string = NULL;
    }
  }
  if (next_object != NULL)
  {
    if (next_object->rope_string != NULL)
    {
      akamaru_model_delete_string (&dock->model, next_object->rope_string);
      next_object->rope_string = NULL;
    }
  }

  if (prev_object != NULL && next_object != NULL)
    next_object->rope_string =
      akamaru_model_add_string (&dock->model, next_object->object,
          prev_object->object, dgconf->icon_size *
         (dgconf->rope_object_distance / 10));
}

void
kiba_reorder_objectlist (KibaDock *dock)
{
  GList *buf_list = NULL, *check_list;
/*
  DO_FOR_ALL_OBJECTS
  (
    KibaObject *next_object = NULL;
    int offset = 0;

    for (check_list = buf_list; check_list != NULL; check_list = check_list->next)
    {
      KibaObject *check_object = check_list->data;

      if ((check_object->position - object->position < offset || offset == 0)
        && check_object->position > object->position)
      {
        next_object = check_object;
        offset = check_object->position - object->position;
      }
    }

    if (next_object == NULL)
      buf_list = g_list_insert (buf_list, object, object->position - 1);
    else
      buf_list = g_list_insert_before (buf_list,
                 g_list_find (buf_list, next_object), object);
  )

  if (g_list_length (dock->objects) != g_list_length (buf_list))
    return;

  dock->objects = buf_list;
*/
}

void
kiba_update_object_positions (KibaDock *dock)
{
  int position = 1;
/*
  DO_FOR_ALL_OBJECTS
  (
    if (object->position != position)
    {
      GError *error = NULL;

      object->position_notify = FALSE;

      gconf_client_set_int (dgconf->client,
        g_strdup_printf ("%s/%s_position", object->key, object->plugin->name) , position, &error);
      if (error != NULL)
      {
        g_error ("Failed to set gconf position key for %s. notifications: %s",
                 object->name, error->message);
        g_error_free (error);
      }

      object->position = position;

      object->position_notify = TRUE;
    }

    position++;
  )
*/
}

void
kiba_hide_objects (KibaDock *dock)
{
  GList     *g_list;

  for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
  {
    KibaObject *object = g_list->data;

    if (gdk_window_get_state (object->window) != GDK_WINDOW_STATE_WITHDRAWN)
      gdk_window_hide (object->window);

    if (gdk_window_get_state (object->pulse_window) != GDK_WINDOW_STATE_WITHDRAWN)
      gdk_window_hide (object->pulse_window);
  }
}

void
kiba_show_objects (KibaDock *dock)
{
  GList     *g_list;

  for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
  {
    KibaObject *object = g_list->data;

    if (gdk_window_get_state (object->window) == GDK_WINDOW_STATE_WITHDRAWN)
      gdk_window_show (object->window);

    if (gdk_window_get_state (object->pulse_window) == GDK_WINDOW_STATE_WITHDRAWN)
      gdk_window_show (object->pulse_window);
  }
}

void
kiba_object_show_window (GtkWidget* widget, gpointer data)
{
  GdkWindow *window = data;
  Display   *display;
  Window     xwindow;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  xwindow = gdk_x11_drawable_get_xid (window);

  XMapWindow (display, xwindow);

  XCloseDisplay (display);

  kiba_object_focus_window (widget, window);
}

void
kiba_object_focus_window (GtkWidget* widget, gpointer data)
{
  GdkWindow *window = data;
  Display   *display;
  Window     xwindow;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  xwindow = gdk_x11_drawable_get_xid (window);

  gdk_window_focus (window, gtk_get_current_event_time ());

//   XRaiseWindow (display, xwindow);
  XSetInputFocus (display, xwindow, 1, CurrentTime);

  XCloseDisplay (display);
}

void
kiba_object_iconify_window (GtkWidget* widget, gpointer data)
{
  GdkWindow *window = data;
  Display   *display;
  Window     xwindow;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  xwindow = gdk_x11_drawable_get_xid (window);

  XIconifyWindow (display, xwindow, DefaultScreen(display));

  XCloseDisplay (display);
}

void
kiba_object_kill_window (GtkWidget* widget, gpointer data)
{
  GdkWindow *window = data;
  Display   *display;
  Window     xwindow;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  xwindow = gdk_x11_drawable_get_xid (window);

  XKillClient (display, xwindow);

  XCloseDisplay (display);
}

/* returns current windowstate, -1=not running, 0=systray/not mapped, 1=shown, 3=iconified */
int
kiba_object_get_winstate (Display *display, GdkWindow *window)
{
  Window            xwindow;
  Atom              type_return, state_atom;
  int               format_return;
  unsigned long int nitems_return;
  unsigned long int bytes_after_return;
  unsigned char    *state;

  xwindow    = gdk_x11_drawable_get_xid (window);
  state_atom = XInternAtom (display, "WM_STATE", TRUE);

  gdk_error_trap_push();
  if ((XGetWindowProperty (display, xwindow, state_atom, 0, G_MAXLONG, False, AnyPropertyType,
      &type_return, &format_return, &nitems_return, &bytes_after_return, &state) == Success) &&
       state != NULL)
    return (int)(((unsigned long *) state)[0]);

  return -1;
}

Window
kiba_object_get_active_window (Display *display)
{
  Window  window;
  Atom    type_return;
  gint    format_return;
  gulong  nitems_return;
  gulong  bytes_after_return;
  guchar *data = NULL;

  Atom active_atom = XInternAtom (display, "_NET_ACTIVE_WINDOW", TRUE);

  if (XGetWindowProperty (display, XDefaultRootWindow (display), active_atom,
        0, 1, False, XA_WINDOW, &type_return, &format_return, &nitems_return,
        &bytes_after_return, &data) == Success)
  {
    if ((type_return == XA_WINDOW) && (format_return == 32) && (data))
      window = *(Window *) data;
  }

  if (data)
    XFree (data);

  return window;
}

int
kiba_get_active_workspace_number (Display *display)
{
  static Atom _NET_CURRENT_DESKTOP = None;
  static Atom _WIN_WORKSPACE = None;
  Atom type_ret = None;
  int format_ret = 0;
  unsigned long nitems_ret = 0;
  unsigned long bytes_after_ret = 0;
  unsigned int *prop_ret = NULL;
  gint ws_num = 0;

  if (G_UNLIKELY (!_NET_CURRENT_DESKTOP))
  {
    _NET_CURRENT_DESKTOP = XInternAtom (display, "_NET_CURRENT_DESKTOP", False);
    _WIN_WORKSPACE = XInternAtom (display, "_WIN_WORKSPACE", False);
  }

  gdk_error_trap_push();
  if(XGetWindowProperty(display, XDefaultRootWindow (display), _NET_CURRENT_DESKTOP,
          0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
          &bytes_after_ret, (unsigned char **)&prop_ret) != Success)
  {
    if(XGetWindowProperty(display, XDefaultRootWindow (display), _WIN_WORKSPACE,
        0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
        &bytes_after_ret, (unsigned char **)&prop_ret) != Success)
      return 0;
  }

  gdk_error_trap_pop();

  if (type_ret == None || format_ret == 0 || !prop_ret)
  {
    if(prop_ret)
      XFree(prop_ret);
    return 0;
  }

  ws_num = *prop_ret;
  XFree(prop_ret);

  return ws_num;
}

int
kiba_get_win_workspace_number (Display *display, Window window)
{
  static Atom _NET_WM_DESKTOP = None;
  static Atom _WIN_WORKSPACE = None;
  Atom type_ret = None;
  int format_ret = 0;
  unsigned long nitems_ret = 0;
  unsigned long bytes_after_ret = 0;
  unsigned int *prop_ret = NULL;
  gint ws_num = 0;

  if (G_UNLIKELY (!_NET_WM_DESKTOP))
  {
    _NET_WM_DESKTOP = XInternAtom (display, "_NET_WM_DESKTOP", False);
    _WIN_WORKSPACE = XInternAtom (display, "_WIN_WORKSPACE", False);
  }

  gdk_error_trap_push();
  if(XGetWindowProperty(display, window, _NET_WM_DESKTOP,
          0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
          &bytes_after_ret, (unsigned char **)&prop_ret) != Success)
  {
    if(XGetWindowProperty(display, window, _WIN_WORKSPACE,
       0, 32, False, XA_CARDINAL, &type_ret, &format_ret, &nitems_ret,
       &bytes_after_ret, (unsigned char **)&prop_ret) != Success)
      return 0;
  }

  gdk_error_trap_pop();

  if (type_ret == None || format_ret == 0 || !prop_ret)
  {
    if(prop_ret)
      XFree(prop_ret);
    return 0;
  }

  ws_num = *prop_ret;
  XFree(prop_ret);

  return ws_num;
}

char*
kiba_object_get_wmclass (Display *display, Window window)
{
  XTextProperty text_property;
  Atom          cmd_atom;

  cmd_atom = XInternAtom (display, "WM_CLASS", TRUE);
  if (XGetTextProperty (display, window, &text_property, cmd_atom))
      return text_property.value;

  return NULL;
}

/* returns true if urgency flag is set */
gboolean
kiba_object_window_urgency (Display *display, GdkWindow *window)
{
  XWMHints *wm_hints;

  wm_hints = XGetWMHints (display,  gdk_x11_drawable_get_xid (window));

  if (wm_hints)
    return wm_hints->flags & XUrgencyHint;

  return FALSE;
}
