/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <time.h>
#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>
#include <cairo.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#include <akamaru.h>

#include <kiba-dock.h>
#include <kiba-akamaru.h>
#include <kiba-render.h>
#include <kiba-gconf.h>
#include <kiba-object.h>

#define SECOND_INTERVAL  1000
#define MINUTE_INTERVAL 60000
#define CLOCK_GCONF_PATH KIBA_GCONF_PATH"/plugins/clock"

typedef struct _KibaClock   KibaClock;

struct _KibaClock {
  KibaDock        *dock;
  KibaObject      *object;

  GdkPixbuf       *mark_pixbuf;
  GdkPixbuf       *shadow_pixbuf;
  GdkPixbuf       *glass_pixbuf;
  GdkPixbuf       *hour_hand_pixbuf;
  GdkPixbuf       *minute_hand_pixbuf;
  GdkPixbuf       *second_hand_pixbuf;

#ifdef HAVE_SVG
  RsvgHandle      *mark_svg_handle;
  RsvgHandle      *shadow_svg_handle;
  RsvgHandle      *glass_svg_handle;
  RsvgHandle      *hour_hand_svg_handle;
  RsvgHandle      *minute_hand_svg_handle;
  RsvgHandle      *second_hand_svg_handle;
#endif

  gint             timeout_id;
};

void
kiba_plugin_init            (KibaDock *dock, GError *error);
char*
kiba_plugin_name            (void);
void
kiba_plugin_timeout         (KibaDock *dock, KibaObject *object);
void
kiba_plugin_paint           (KibaDock *dock, KibaObject *object, cairo_t *pixmap_context);
void
kiba_plugin_notify          (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data);
void
kiba_plugin_menu            (KibaDock *dock, KibaObject *object, GtkMenu *menu);
void
kiba_plugin_release_event (KibaDock *dock, KibaObject *object);

static KibaClock*
kiba_clock_new              (KibaDock *dock);
static gboolean
kiba_add_clock              (KibaDock *dock, KibaClock *clock, gint x, gint y, GError *error);
static gboolean
kiba_clock_time_handler     (KibaDock *dock);
static KibaClock *
kiba_get_clock              (KibaDock *dock);
static void
kiba_clock_disable          (void);
static void
kiba_clock_fill_surface_buffer (KibaDock *dock, KibaObject *object, gboolean background);

void
kiba_plugin_init (KibaDock *dock, GError *error)
{
  KibaPlugin  *plugin;
  KibaClock   *clock;

  if (!dgconf->clock_enable)
    return;

  plugin = g_new0 (KibaPlugin, 1);
  clock  = kiba_clock_new (dock);
  if (clock == NULL || plugin == NULL)
    return;

  if (error != NULL)
    g_clear_error (&error);

  plugin->name          = g_strdup ("clock");
  plugin->plugin_struct = clock;

  clock->object->plugin = plugin;

  kiba_add_clock (dock, clock, g_random_int_range (1, dock->geometry.width),
                      g_random_int_range (1, dock->geometry.height), error);

  clock->timeout_id =
      g_timeout_add (SECOND_INTERVAL, (GSourceFunc) kiba_clock_time_handler, (gpointer) dock);

  dock->plugins = g_list_append (dock->plugins, clock->object->plugin);

  return;
}

char*
kiba_plugin_name (void)
{
  return "clock";
}

void
kiba_plugin_paint (KibaDock *dock, KibaObject *object, cairo_t *pixmap_context)
{
  KibaClock *clock;
  cairo_text_extents_t textExtents;
  char      g_acDate[6];
  time_t    g_timeOfDay;
  struct    tm*  g_pTime;
  gint      seconds, minutes, hours, day, month;
  double    r, g, b, offset;

  if (dock->hiding || !dgconf->clock_enable)
    return;

  clock = kiba_get_clock (dock);

  time (&g_timeOfDay);
  g_pTime = localtime (&g_timeOfDay);
  seconds = g_pTime->tm_sec;
  minutes = g_pTime->tm_min;
  hours   = g_pTime->tm_hour;

  if (!dgconf->clock_24mode)
    hours = hours >= 12 ? hours - 12 : hours;

  day   = g_pTime->tm_mday;
  month = g_pTime->tm_mon + 1;
  sprintf (g_acDate, "%02d/%02d", day, month);

  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_SOURCE);

  if (object->rotate_angle != 0.0)
    offset = -object->normal_size / 2.0;
  else
    offset =  .0f;

  cairo_set_source_surface (pixmap_context, object->background_surface, offset, offset);

  cairo_paint_with_alpha (pixmap_context, dgconf->clock_alpha);

  cairo_restore (pixmap_context);

  if (object->prelight)
  {
    cairo_save (pixmap_context);
      convert_color (dgconf->prelight_color, &r, &g, &b);
      cairo_set_source_rgba (pixmap_context, r, g, b, dgconf->prelight_alpha);
      cairo_set_operator (pixmap_context, CAIRO_OPERATOR_ATOP);
      cairo_paint (pixmap_context);
    cairo_restore (pixmap_context);
  }

  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_OVER);

  cairo_save (pixmap_context);

  cairo_translate (pixmap_context, object->normal_size / 2.0, object->normal_size / 2.0);
  cairo_rotate (pixmap_context, -M_PI/2.0f);

  if (dgconf->clock_show_date)
  {
    cairo_save (pixmap_context);
    convert_color (dgconf->clock_date_text_color, &r, &g, &b);
    cairo_set_source_rgba (pixmap_context, r, g, b, 1.0);
    cairo_set_line_width (pixmap_context, 5.0f);
    cairo_text_extents (pixmap_context, g_acDate, &textExtents);
    cairo_rotate (pixmap_context, (M_PI/180.0f) * 90.0f);
    cairo_move_to (pixmap_context, -textExtents.width / 2.0f, 2.0f * textExtents.height);
    cairo_show_text (pixmap_context, g_acDate);
    cairo_restore (pixmap_context);
  }

  /* hour hand */
  cairo_save (pixmap_context);
    cairo_rotate (pixmap_context, (M_PI/ (dgconf->clock_24mode ? 12.0f : 6.0f)) * hours +
                                  (M_PI/ (dgconf->clock_24mode ? 720.0f : 360.0f)) * minutes);

    if (clock->hour_hand_pixbuf != NULL)
    {
      gdk_cairo_set_source_pixbuf (pixmap_context, clock->hour_hand_pixbuf, 0, 0);
      cairo_paint (pixmap_context);
    }
    #ifdef HAVE_SVG
    else if (clock->hour_hand_svg_handle != NULL)
      rsvg_handle_render_cairo (clock->hour_hand_svg_handle, pixmap_context);
    #endif
    else
    {
      convert_color (dgconf->clock_hour_hand_color, &r, &g, &b);
      cairo_set_source_rgba (pixmap_context, r, g, b, 1.0);
      cairo_set_line_width (pixmap_context, 3.5);
      cairo_move_to (pixmap_context, 0.5, 0.5);
      cairo_rel_line_to (pixmap_context, 0.25*object->normal_size, 0.0);
      cairo_stroke (pixmap_context);
    }
  cairo_restore (pixmap_context);

  /* minute hand */
  cairo_save (pixmap_context);
    cairo_rotate (pixmap_context, (M_PI/30.0f) * minutes);
    if (clock->minute_hand_pixbuf != NULL)
    {
      gdk_cairo_set_source_pixbuf (pixmap_context, clock->minute_hand_pixbuf, 0, 0);
      cairo_paint (pixmap_context);
    }
    #ifdef HAVE_SVG
    else if (clock->minute_hand_svg_handle != NULL)
      rsvg_handle_render_cairo (clock->minute_hand_svg_handle, pixmap_context);
    #endif
    else
    {
      convert_color (dgconf->clock_minute_hand_color, &r, &g, &b);
      cairo_set_source_rgba (pixmap_context, r, g, b, 1.0);
      cairo_set_line_width (pixmap_context, 2.0);
      cairo_move_to (pixmap_context, 0.5, 0.5);
      cairo_rel_line_to (pixmap_context, 0.35*object->normal_size, 0.0);
      cairo_stroke (pixmap_context);
    }
  cairo_restore (pixmap_context);

  /* second hand */
  if (dgconf->clock_show_seconds)
  {
    cairo_save (pixmap_context);
    cairo_rotate (pixmap_context, (M_PI/30.0f) * seconds);
    if (clock->second_hand_pixbuf != NULL)
    {
      gdk_cairo_set_source_pixbuf (pixmap_context, clock->second_hand_pixbuf, 0, 0);
      cairo_paint (pixmap_context);
    }
    #ifdef HAVE_SVG
    else if (clock->second_hand_svg_handle != NULL)
      rsvg_handle_render_cairo (clock->second_hand_svg_handle, pixmap_context);
    #endif
    else
    {
      convert_color (dgconf->clock_second_hand_color, &r, &g, &b);
      cairo_set_source_rgba (pixmap_context, r, g, b, 1.0);
      cairo_set_line_width (pixmap_context, 0.75);
      cairo_move_to (pixmap_context, 0.5, 0.5);
      cairo_rel_line_to (pixmap_context, 0.4*object->normal_size, 0.0);
      cairo_stroke (pixmap_context);
    }
  }

  /* inner circle */
  cairo_save (pixmap_context);
    convert_color (dgconf->clock_circle_color, &r, &g, &b);
    cairo_set_source_rgba (pixmap_context, r, g, b, 1.0);
    cairo_set_line_width (pixmap_context, 0.25);
    cairo_arc (pixmap_context, 0.5, 0.5, dgconf->clock_circle_radius, 0, 2 * M_PI);
    cairo_fill_preserve (pixmap_context);
  cairo_restore (pixmap_context);

/*
#define clock_night_blend_alpha 0.3
  cairo_save (pixmap_context);
    convert_color ("#000000", &r, &g, &b);
    double blend_alpha;

    blend_alpha = (((double)hours * 60.0 + (double) minutes - 1220.0) / 1220.0) * clock_night_blend_alpha;
    if (blend_alpha < 0.0) blend_alpha *= -1.0;

    cairo_set_source_rgba (pixmap_context, r, g, b, blend_alpha);

    cairo_set_operator (pixmap_context, CAIRO_OPERATOR_ATOP);
    cairo_paint (pixmap_context);
  cairo_restore (pixmap_context);
*/

  cairo_restore (pixmap_context);

  if (object->foreground_surface != NULL)
  {
    cairo_set_source_surface (pixmap_context, object->foreground_surface, 0.0f, 0.0f);
    cairo_paint (pixmap_context);
  }
}

void
kiba_plugin_timeout (KibaDock *dock, KibaObject *object)
{
  KibaClock *clock = NULL;

  if (!dgconf->clock_enable || object->window == NULL)
   return;

  DO_FOR_ALL_OBJECTS
  (
    if (strcmp (object->plugin->name, "clock") == 0)
      clock = object->plugin->plugin_struct;
  );

  if (clock->object->bgsurface_needs_update == TRUE || clock->object->fgsurface_needs_update == TRUE)
  {
    cairo_t*         clock_context = NULL;

    clock_context = dock_cairo_create (clock->object);
    cairo_set_operator (clock_context, CAIRO_OPERATOR_SOURCE);

    if (clock->object->bgsurface_needs_update)
    {
      kiba_clock_fill_surface_buffer (dock, clock->object, TRUE);

      clock->object->bgsurface_needs_update = FALSE;
    }
    else
    {
      if (dgconf->clock_glass != NULL)
        kiba_clock_fill_surface_buffer (dock, clock->object, FALSE);

      clock->object->fgsurface_needs_update = FALSE;
    }

    clock->object->rerender = TRUE;

    cairo_destroy (clock_context);
  }

  if (clock->object->drawn_rotate_angle != 0.0 ||
      clock->object->drawn_rotate_angle != clock->object->rotate_angle)
    clock->object->rerender = TRUE;
}

void
kiba_plugin_release_event (KibaDock *dock, KibaObject *object)
{

}

/* gconf notify event */
void
kiba_plugin_notify (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
  KibaDock  *dock;
  KibaClock *clock;
  GError    *error = NULL;

  dock  = user_data;
  clock = kiba_get_clock (dock);

  set_option (dock, entry);

  if (strcmp (entry->key, CLOCK_GCONF_PATH "/clock_size") == 0)
  {
    kiba_reload (dock);
/*
    clock->object->normal_size   = (double) gconf_value_get_float (entry->value);
    clock->object->size          = (double) gconf_value_get_float (entry->value);

    clock->object->bgsurface_needs_update = TRUE;
    clock->object->fgsurface_needs_update = TRUE;

    gdk_window_resize (clock->object->window, clock->object->size, clock->object->size);

    kiba_reload_spacers (dock);

    kiba_layout (dock);
*/
  }
  else if (strcmp (entry->key, CLOCK_GCONF_PATH "/clock_enable") == 0)
  {
    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_delete_rope_spring (dock);
      kiba_delete_rope_strings (dock);
    }

    if (clock != NULL && dgconf->clock_enable == FALSE)
    {
      dock->plugins = g_list_remove (dock->plugins, clock->object->plugin);

      g_source_remove (clock->timeout_id);

      kiba_object_free (dock, clock->object);
      dock->objects = g_list_remove (dock->objects, clock->object);
      clock->object = NULL;
      dock->num_objects--;

      clock = NULL;

      kiba_reload_spacers (dock);

      kiba_update_object_positions (dock);
    }
    else
       kiba_plugin_init (dock, error);

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_add_rope_spring (dock, 1.0);
      kiba_add_rope_strings (dock);
    }

    kiba_layout (dock);
  }
  else if (g_strrstr (entry->key, "/plugins/clock/clock_show_seconds"))
  {
    if (dgconf->clock_show_seconds == TRUE)
    {
      g_source_remove (clock->timeout_id);
      clock->timeout_id = g_timeout_add (SECOND_INTERVAL, (GSourceFunc) kiba_clock_time_handler, (gpointer) dock);
    }
    else
    {
      g_source_remove (clock->timeout_id);
      clock->timeout_id = g_timeout_add (MINUTE_INTERVAL, (GSourceFunc) kiba_clock_time_handler, (gpointer) dock);
      clock->object->rerender = TRUE;
    }
  }
  else if (dgconf->clock_enable &&
          (g_strrstr (entry->key, "/plugins/clock/clock_alpha") ||
           g_strrstr (entry->key, "/plugins/clock/clock_color") ||
           g_strrstr (entry->key, "/plugins/clock/clock_border_") ||
           g_strrstr (entry->key, "/plugins/clock/clock_marker_") ||
           g_strrstr (entry->key, "/plugins/clock/clock_bg_")))
    clock->object->bgsurface_needs_update = TRUE;
  else if (g_strrstr (entry->key, "/plugins/clock/clock_show_date") ||
           g_strrstr (entry->key, "_hand_color") ||
           g_strrstr (entry->key, "clock_circle_color") ||
           g_strrstr (entry->key, "/plugins/clock/clock_circle_radius") ||
           g_strrstr (entry->key, "/plugins/clock/clock_24mode"))
    clock->object->rerender = TRUE;
  else if (g_strrstr (entry->key, "clock_hour_hand"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_hour_hand, &clock->hour_hand_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_hour_hand, &clock->hour_hand_pixbuf, &error);
    g_clear_error (&error);

    clock->object->rerender = TRUE;
  }
  else if (g_strrstr (entry->key, "clock_minute_hand"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_minute_hand, &clock->minute_hand_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_minute_hand, &clock->minute_hand_pixbuf, &error);
    g_clear_error (&error);

    clock->object->rerender = TRUE;
  }
  else if (g_strrstr (entry->key, "clock_second_hand"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_second_hand, &clock->second_hand_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_second_hand, &clock->second_hand_pixbuf, &error);

    clock->object->rerender = TRUE;
  }
  else if (g_str_has_suffix (entry->key, "/clock_face"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_face, &clock->object->bg_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_face, &clock->object->bg_pixbuf, &error);

    clock->object->bgsurface_needs_update = TRUE;
  }
  else if (g_str_has_suffix (entry->key, "/clock_shadow"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_shadow, &clock->shadow_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_shadow, &clock->shadow_pixbuf, &error);

    clock->object->bgsurface_needs_update = TRUE;
  }
  else if (g_str_has_suffix (entry->key, "/clock_marks"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_marks, &clock->mark_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_marks, &clock->mark_pixbuf, &error);

    clock->object->bgsurface_needs_update = TRUE;
  }
  else if (g_str_has_suffix (entry->key, "/clock_glass"))
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (clock->object, clock->dgconf->clock_glass, &clock->glass_svg_handle, &error))
    #endif
      kiba_load_png (dock, clock->object, clock->dgconf->clock_glass, &clock->glass_pixbuf, &error);

    clock->object->fgsurface_needs_update = TRUE;
  }
  else if (g_str_has_suffix (entry->key, "/clock_position"))
    kiba_object_position_changed (dock, clock->object, entry);

  if (error != NULL)
    g_error_free (error);
}

void
kiba_plugin_menu (KibaDock *dock, KibaObject *object, GtkMenu *menu)
{
  GtkWidget  *disable_item;

  disable_item = gtk_menu_item_new_with_label ("Disable Clock");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), disable_item);
  g_signal_connect (G_OBJECT (disable_item), "activate", G_CALLBACK (kiba_clock_disable), NULL);
  gtk_widget_show (disable_item);
}

static KibaClock *
kiba_clock_new (KibaDock *dock)
{
  KibaClock *clock;

  clock                           = g_new0 (KibaClock, 1);
  clock->object                   = g_new0 (KibaObject, 1);
//   clock->object->position_notify  = TRUE;

  return clock;
}

static gboolean
kiba_add_clock (KibaDock *dock, KibaClock *clock, gint x, gint y, GError *error)
{
  AddSpacerClosure closure;

  closure.dock           = dock;
  closure.object         = akamaru_model_add_object (&dock->model, x, y, 12, NULL);
  closure.spacer_size    = (double) dgconf->clock_size + dgconf->object_space;
  closure.object->radius = closure.spacer_size / 2.0;

  clock->object->spring =
    akamaru_model_add_spring (&dock->model, dock->anchor->object, closure.object, 0);
  dock->objects         = g_list_append (dock->objects, clock->object);
  dock->num_objects++;

  akamaru_model_for_each_object (&dock->model, add_spacer, &closure);

  clock->object->object           = closure.object;
  clock->dock                     = dock;
  clock->object->object->friction = dock->model.friction;
  clock->object->win_position.x   = clock->object->object->position.x;
  clock->object->win_position.y   = clock->object->object->position.y;
  clock->object->normal_size      = dgconf->clock_size;
  clock->object->scale_factor     = 1.0;
  clock->object->key              = g_strdup (CLOCK_GCONF_PATH);
  clock->object->name             = g_strdup ("clock");

  /* load svgs and pngs */
  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_face, &clock->object->bg_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_face, &clock->object->bg_pixbuf, &error);

  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_shadow, &clock->shadow_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_shadow, &clock->shadow_pixbuf, &error);

  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_marks, &clock->mark_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_marks, &clock->mark_pixbuf, &error);

  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_glass, &clock->glass_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_glass, &clock->glass_pixbuf, &error);

  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_hour_hand, &clock->hour_hand_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_hour_hand, &clock->hour_hand_pixbuf, &error);

  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_minute_hand, &clock->minute_hand_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_minute_hand, &clock->minute_hand_pixbuf, &error);

  #ifdef HAVE_SVG
  if (!kiba_load_svg (clock->object, clock->dgconf->clock_second_hand, &clock->second_hand_svg_handle, &error))
  #endif
    kiba_load_png (dock, clock->object, clock->dgconf->clock_second_hand, &clock->second_hand_pixbuf, &error);

  clock->object->bgsurface_needs_update = TRUE;
  clock->object->fgsurface_needs_update = TRUE;

//   kiba_object_update_position (dock, clock->object, reload);

  if (GTK_WIDGET_REALIZED (GTK_WIDGET (dock)))
    kiba_object_create_window (dock, clock->object);

  if (GTK_WIDGET_MAPPED (GTK_WIDGET (dock)))
    gdk_window_show (clock->object->window);

  kiba_layout (dock);
}

static KibaClock *
kiba_get_clock (KibaDock *dock)
{
  DO_FOR_ALL_OBJECTS
  (
    if (g_strrstr (object->plugin->name, "clock"))
      return object->plugin->plugin_struct;
  )

  return NULL;
}

static void
kiba_clock_fill_surface_buffer (KibaDock *dock, KibaObject *object, gboolean background)
{
  KibaClock *clock;
  cairo_t *drawing_context, *cr = NULL;
  cairo_surface_t* pNewSurface;
  double   r, g, b;

  cr = dock_cairo_create (object);
  if (cr == NULL)
    return;

  clock = kiba_get_clock (dock);
  if (clock == NULL)
  {
    cairo_destroy (cr);
    return;
  }

  pNewSurface = cairo_surface_create_similar (cairo_get_target (cr), CAIRO_CONTENT_COLOR_ALPHA,
                                  object->buffer_size, object->buffer_size);

  if (cairo_surface_status (pNewSurface) != CAIRO_STATUS_SUCCESS)
  {
    cairo_destroy (cr);
    return;
  }

  drawing_context = cairo_create (pNewSurface);
  if (cairo_status (drawing_context) != CAIRO_STATUS_SUCCESS)
  {
    cairo_surface_destroy (pNewSurface);
    cairo_destroy (cr);
    return;
  }
  cairo_set_source_rgba (drawing_context, 1.0f, 1.0f, 1.0f, 0.0f);
  cairo_set_operator (drawing_context, CAIRO_OPERATOR_OVER);
  cairo_paint (drawing_context);

  /* draw shadow */
  if (background == TRUE && clock->shadow_pixbuf != NULL)
  {
    gdk_cairo_set_source_pixbuf (drawing_context, clock->shadow_pixbuf, 0, 0);
    cairo_set_operator (drawing_context, CAIRO_OPERATOR_SOURCE);
    cairo_paint (drawing_context);
  }
#ifdef HAVE_SVG
  else if (background == TRUE && clock->shadow_svg_handle != NULL)
    rsvg_handle_render_cairo (clock->shadow_svg_handle, drawing_context);
#endif

  /* draw face */
  if ((background == TRUE ? object->bg_pixbuf : object->fg_pixbuf) != NULL)
  {
    gdk_cairo_set_source_pixbuf (drawing_context, (background == TRUE ? object->bg_pixbuf : object->fg_pixbuf), 0, 0);
    cairo_set_operator (drawing_context, CAIRO_OPERATOR_SOURCE);
    cairo_paint (drawing_context);
  }
#ifdef HAVE_SVG
  else if ((background == TRUE ? object->bg_svg_handle : object->fg_svg_handle) != NULL)
    rsvg_handle_render_cairo ((background == TRUE ? object->bg_svg_handle :
                               object->fg_svg_handle), drawing_context);
#endif
  else if (background)
  {
    double radius = clock->object->size / 2.0;

    cairo_arc (drawing_context, radius, radius, radius - (clock->dgconf->clock_border_width / 2.0), 0.0, 2 * M_PI);
    cairo_close_path (drawing_context);

    convert_color (clock->dgconf->clock_color, &r, &g, &b);
    cairo_set_source_rgba (drawing_context, r, g, b, clock->dgconf->clock_alpha);
    cairo_fill_preserve (drawing_context);

    convert_color (clock->dgconf->clock_border_color, &r, &g, &b);
    cairo_set_source_rgba (drawing_context, r, g, b, clock->dgconf->clock_border_alpha);
    cairo_set_line_width (drawing_context, clock->dgconf->clock_border_width);
    cairo_stroke (drawing_context);
  }

  /* draw marker */
  if (background && clock->mark_pixbuf != NULL)
  {
    gdk_cairo_set_source_pixbuf (drawing_context, clock->mark_pixbuf, 0, 0);
    cairo_set_operator (drawing_context, CAIRO_OPERATOR_SOURCE);
    cairo_paint (drawing_context);
  }
#ifdef HAVE_SVG
  else if (background && clock->mark_svg_handle != NULL)
    rsvg_handle_render_cairo (clock->mark_svg_handle, drawing_context);
#endif
  else if (background)
  {
    int fnord;

    convert_color (clock->dgconf->clock_marker_color, &r, &g, &b);
    cairo_set_source_rgba (cr, r, g, b, .5);
    for (fnord = 1; fnord < 13; fnord++)
    {
      cairo_save (cr);
        cairo_translate (cr, 0.5*clock->object->size, 0.5*clock->object->size);
        cairo_rotate (cr, (M_PI * fnord)/6.0);
        if (fnord == 3 || fnord == 6 || fnord == 9 || fnord == 12)
        {
          cairo_set_line_width (cr, 1.2*clock->dgconf->clock_marker_width);
          cairo_move_to (cr, 0.25*clock->object->size, 0.5);
        }
        else
        {
           cairo_set_line_width (cr, clock->dgconf->clock_marker_width);
          cairo_move_to (cr, 0.3*clock->object->size, 0.5);
        }
        cairo_line_to (cr, 0.4*clock->object->size, 0.0);
        cairo_stroke (cr);
      cairo_restore (cr);
    }
  }

  /* draw glass */
  if (!background && clock->glass_pixbuf != NULL)
  {
    gdk_cairo_set_source_pixbuf (drawing_context, clock->glass_pixbuf, 0, 0);
    cairo_set_operator (drawing_context, CAIRO_OPERATOR_SOURCE);
    cairo_paint (drawing_context);
  }
#ifdef HAVE_SVG
  else if (!background && clock->glass_svg_handle != NULL)
    rsvg_handle_render_cairo (clock->glass_svg_handle, drawing_context);
#endif

  cairo_destroy (drawing_context);

  if (background)
  {
    if (object->background_surface)
    {
      cairo_surface_destroy (object->background_surface);
      object->background_surface = NULL;
    }

    object->background_surface = pNewSurface;
  }
  else
  {
    if (object->foreground_surface)
    {
      cairo_surface_destroy (object->foreground_surface);
      object->foreground_surface = NULL;
    }

    object->foreground_surface = pNewSurface;
  }

  cairo_destroy (cr);
}


static void
kiba_clock_disable (void)
{
  GConfClient *gconf_client;
  GError *error = NULL;

  gconf_client = gconf_client_get_default ();

  if (gconf_client_set_bool (gconf_client, CLOCK_GCONF_PATH "/clock_enable", FALSE, &error) == FALSE)
    fprintf (stderr, "failed to disable the clock\n");

  if (error != NULL)
  {
    fprintf (stderr, "%s\n", error->message);
    g_error_free (error);
  }
}

static gboolean
kiba_clock_time_handler (KibaDock *dock)
{
  KibaClock  *clock;
  gchar      *new_string;
  int         seconds, minutes, hours;
  static      time_t systime;
  struct      tm* g_pTime;

  clock = kiba_get_clock (dock);
  if (clock == NULL)
    return FALSE;

  if (dock->num_objects == 0)
    return TRUE;

  time (&systime);
  g_pTime = localtime (&systime);
  seconds = g_pTime->tm_sec;
  minutes = g_pTime->tm_min;
  hours = g_pTime->tm_hour;

  new_string = g_strdup_printf ("%s%i:%s%i:%s%i",
               (hours   < 10) ? "0" : "", hours,
               (minutes < 10) ? "0" : "", minutes,
               (seconds < 10) ? "0" : "", seconds);

  if (strcmp (new_string, clock->object->name) != 0)
  {
    clock->object->name = g_strdup (new_string);

    kiba_text_update_pango_for_object (clock->object, dock->gconf, dock->text);
  }
  g_free (new_string);

  if (clock != NULL && !dock->hiding)
    clock->object->rerender = TRUE;

  return TRUE;
}
