/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <gconf/gconf-client.h>
#include <cairo.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include <akamaru.h>

#include <kiba-dock.h>
#include <kiba-akamaru.h>
#include <kiba-render.h>
#include <kiba-gconf.h>
#include <kiba-object.h>

#define LAUNCHER_GCONF_PATH KIBA_GCONF_PATH"/plugins/launcher"
#define SECOND_INTERVAL  1000
#define FADE_STEP        0.005
#define ldock            launcher->dock
#define lgconf           ldock->gconf

typedef struct _KibaLauncher      KibaLauncher;
typedef struct _KibaLauncherIcon  KibaLauncherIcon;

struct _KibaLauncherIcon {
  KibaObject   *object;

  char         *desktop_key;   /* path to the desktop file */
  char         *exec_command;  /* command line for exec */
  char         *active_icon;   /* path to the icon */
  char         *inactive_icon; /* path to the icon */
  char         *name;          /* object name (source for textrendering) */

  double        blend_alpha;
  gboolean      running;       /* stores if the app is running */
  gint          notify_id;     /* gconf notify id */
};


struct _KibaLauncher {
  KibaDock        *dock;
  GList           *launcher_icons;   /* list with all KibaLauncherIcon's */

  gint             timeout_id;
};

void
kiba_plugin_init                (KibaDock *dock, GError *error);
char*
kiba_plugin_name                (void);
void
kiba_plugin_paint               (KibaDock *dock, KibaObject *object, cairo_t *cairo_context);
void
kiba_plugin_timeout             (KibaDock *dock, KibaObject *object, GError *error);
void
kiba_plugin_notify              (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data);
void
kiba_plugin_menu                (KibaDock *dock, KibaObject *object, GtkMenu *menu);
void
kiba_plugin_release_event       (KibaDock *dock, KibaObject *object);

static KibaLauncherIcon*
kiba_launcher_new               (KibaDock *dock, KibaLauncher *launcher, char *gconf_path, GError *error);
static void
kiba_add_launcher               (KibaDock *dock, KibaLauncherIcon *launcher, int x, int y, GError *error);
static gboolean
kiba_launcher_remove            (GtkWidget *widget, gpointer data);
static gboolean
kiba_replace_desktop_entry      (KibaLauncherIcon *launcher, char *desktop_entry, char *new_value, GError *error);
static void
kiba_merge_new_desktop_file     (KibaLauncherIcon *launcher, GConfClient *gconf_client, GError *error);
static void
launcher_data_changed_callback  (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data);
static KibaLauncherIcon *
kiba_get_launcher_icon          (KibaObject *object);
static KibaLauncher *
kiba_get_launcher               (KibaDock *dock);
static gboolean
kiba_launcher_time_handler      (KibaLauncher *launcher);
GdkWindow *
kiba_launcher_cmd_running       (Display *display, KibaLauncherIcon *launcher_icon);
static char*
kiba_launcher_load_new_icon     (KibaDock *dock, KibaLauncherIcon *launcher_icon, char* icon, char* key, GError *error);
static void
kiba_launcher_disable           (void);
static void
parse_desktop_file              (KibaLauncherIcon *launcher_icon, const char *uri, GError **error);
static gboolean
kiba_launcher_exec              (gpointer data);
GList *
kiba_launcher_get_winlist       (Display *display, KibaLauncherIcon *launcher_icon);

void
kiba_plugin_init (KibaDock *dock, GError *error)
{
  KibaPlugin   *plugin;
  KibaLauncher *launcher;
  GSList *launchers, *l;
  GDir   *kiba_home_dir;
  char   *file_name;

  if (!dgconf->launcher_enable)
    return;

  plugin   = g_new0 (KibaPlugin, 1);
  launcher = g_new0 (KibaLauncher, 1);
  if (plugin == NULL || launcher == NULL)
    return;

  launcher->dock         = dock;
  plugin->name           = g_strdup ("launcher");
  plugin->plugin_struct  = launcher;

  launchers = gconf_client_all_dirs (dock->gconf->client, LAUNCHER_GCONF_PATH "/launchers", &error);
  if (error != NULL)
  {
    fprintf (stderr, "Failed to open" LAUNCHER_GCONF_PATH "/launchers\n" "notifications: %s\n",
             error->message);

    g_clear_error (&error);
  }

  for (l = launchers; l != NULL; l = l->next)
  {
    KibaLauncherIcon *launcher_icon;
    char *path = g_strdup (l->data);

    launcher_icon = kiba_launcher_new (dock, launcher, path, error);
    g_free (path);
    if (launcher_icon == NULL)
  continue;

    launcher_icon->object->plugin = plugin;

    kiba_add_launcher (dock, launcher_icon,
                       g_random_int_range (1, dock->geometry.width),
                       g_random_int_range (1, dock->geometry.height), error);

    launcher->launcher_icons = g_list_append (launcher->launcher_icons, launcher_icon);
  }

  dock->plugins = g_list_append (dock->plugins, plugin);

  /* check homedir for unused desktop files */
  kiba_home_dir = g_dir_open (g_strdup_printf ("%s/.kiba-dock/", g_get_home_dir ()), 0, &error);
  if (error != NULL)
  {
    fprintf (stderr, "Failed to open %s\n" "notifications: %s\n", g_strdup_printf ("%s/.kiba-dock/", g_get_home_dir ()), error->message);

    g_clear_error (&error);
  }
  else
  {
    while ((file_name = g_strdup (g_dir_read_name (kiba_home_dir))) != NULL)
    {
      gboolean new_desktop_file = TRUE;
      for (l = launchers; l != NULL; l = l->next)
      {
        char *path = g_strdup (l->data);
        char *file;

        file = gconf_client_get_string (dgconf->client, g_strdup_printf ("%s/file", path), &error);
        g_free (path);
        if (error == NULL && file != NULL && file_name != NULL)
        {
          if (g_str_has_suffix (file, file_name))
            new_desktop_file = FALSE;
        }
        else if (error != NULL)
          g_clear_error (&error);
      }

      if (new_desktop_file == TRUE &&
          g_file_test (g_strdup_printf ("%s/.kiba-dock/%s", g_get_home_dir (), file_name), G_FILE_TEST_IS_DIR) == FALSE)
      {
        char *key;

        key = g_strdup_printf (LAUNCHER_GCONF_PATH"/launchers/%s/file",
              g_strndup (file_name, strlen (file_name ) - 8));
        file_name = g_strdup_printf ("%s/.kiba-dock/%s", g_get_home_dir (), file_name);

        g_message ("found unused desktop file in homdir\n" "merging %s to gconf\n", file_name);

        gconf_client_set_string  (dgconf->client, key, file_name, &error);
        if (error != NULL) {
          fprintf (stderr, "Failed to merge unused desktop file %s\n" "notifications: %s\n",
                     file_name, error->message);
          g_clear_error (&error);
        }
        else
        {
          gconf_engine_associate_schema (dgconf->client->engine, key,
                                     "/schemas" LAUNCHER_GCONF_PATH "/launchers/file", &error);

          if (error != NULL) {
            fprintf (stderr, "failed to set schema for %s\n" "notifications: %s\n", key, error->message);
            g_clear_error (&error);
          }
        }
        g_free (key);
      }
    }
    g_free (file_name);
    g_dir_close (kiba_home_dir);
  }

  launcher->timeout_id = g_timeout_add (SECOND_INTERVAL, (GSourceFunc) kiba_launcher_time_handler, (gpointer) launcher);

  return;
}

char*
kiba_plugin_name (void)
{
  return "launcher";
}

void
kiba_plugin_paint (KibaDock *dock, KibaObject *object, cairo_t *cairo_context)
{
  KibaLauncherIcon *launcher_icon;
  double            r, g, b, offset;

  if (!dgconf->launcher_enable || dock->hiding)
    return;

  launcher_icon = kiba_get_launcher_icon (object);
  if (launcher_icon == NULL)
    return;

  /* fill launcher surface buffer */
  if (object->background_surface == NULL)
    kiba_update_surface_buffer (object, TRUE);

  cairo_save (cairo_context);

  cairo_set_operator (cairo_context, CAIRO_OPERATOR_SOURCE);

  if (object->rotate_angle != 0.0)
    offset = -object->normal_size / 2.0;
  else
    offset =  .0f;

  /* draw icon */
  cairo_set_source_surface (cairo_context, object->background_surface, offset, offset);
  cairo_paint_with_alpha (cairo_context, launcher_icon->blend_alpha);

  /* draw prelight */
  if (object->prelight || !launcher_icon->running)
  {
    if (object->prelight)
    {
      convert_color (dgconf->prelight_color, &r, &g, &b);
      cairo_set_source_rgba (cairo_context, r, g, b, dgconf->prelight_alpha);
    }
    else
    {
      double alpha = launcher_icon->blend_alpha - 1.0;
      if (alpha < 0.0) alpha *= -1.0;

      convert_color (dgconf->launcher_running_color, &r, &g, &b);
      cairo_set_source_rgba (cairo_context, r, g, b, alpha);
    }
    cairo_set_operator (cairo_context, CAIRO_OPERATOR_ATOP);
    cairo_paint (cairo_context);
  }

  if (launcher_icon->running && launcher_icon->blend_alpha <= dgconf->launcher_alpha)
    launcher_icon->blend_alpha += FADE_STEP;
  else if (!launcher_icon->running && launcher_icon->blend_alpha >= dgconf->launcher_inactive_alpha)
    launcher_icon->blend_alpha -= FADE_STEP;

/*
TODO foreground surface can be helpfull as notification that a app is running or something similar
  if (launcher->object->foreground_surface != NULL)
  {
    cairo_restore (cairo_context);
    cairo_set_source_surface (cairo_context, launcher->object->foreground_surface, offset_x, offset_y);
    cairo_set_operator (cairo_context, CAIRO_OPERATOR_SOURCE);
    cairo_paint_with_alpha (cairo_context, 0.5);
  }
*/

  cairo_restore (cairo_context);
}

void
kiba_plugin_timeout (KibaDock *dock, KibaObject *object, GError *error)
{
  KibaLauncherIcon *launcher_icon;

  if (dock == NULL || object == NULL || !g_strrstr (object->plugin->name, "launcher"))
    return;

  launcher_icon = kiba_get_launcher_icon (object);

  if (object->drawn_rotate_angle != 0.0 || object->drawn_rotate_angle != object->rotate_angle ||
     (launcher_icon->running && launcher_icon->blend_alpha < dgconf->launcher_alpha) ||
     (!launcher_icon->running && launcher_icon->blend_alpha > dgconf->launcher_inactive_alpha))
    object->rerender = TRUE;
}

void
kiba_plugin_notify (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
  KibaLauncher  *launcher;
  KibaDock      *dock;
  GError        *error = NULL;

  dock     = user_data;
  launcher = kiba_get_launcher (dock);

  if (g_str_has_suffix (entry->key, "/name") ||
      g_str_has_suffix (entry->key, "/exec") ||
      g_str_has_suffix (entry->key, "/active_icon") ||
      g_str_has_suffix (entry->key, "/inactive_icon") ||
      g_str_has_suffix (entry->key, "/launcher_position"))
    return;

  if (!g_str_has_suffix (entry->key, "/file"))
    set_option (dock, entry);

  if (g_str_has_suffix (entry->key, "/file"))
  {
    KibaLauncherIcon *launcher_icon;
    char *path, *last_slash;

    path = g_strdup (entry->key);
    last_slash = strrchr (path, '/');
    *last_slash = '\0';

    launcher_icon = kiba_launcher_new (dock, launcher, path, error);
    g_free (path);
    if (launcher_icon == NULL)
      return;

    launcher_icon->object->plugin = kiba_get_plugin_struct (dock, "launcher");

    kiba_merge_new_desktop_file (launcher_icon, client, error);

    GList *g_list;

    for (g_list = dock->objects; g_list != NULL; g_list = g_list->next)
    {
      KibaObject *object = g_list->data;

      if (strcmp (g_strrstr (gconf_value_get_string (gconf_entry_get_value (entry)), "/") + 1, entry->key) == 0)
      {
        g_free (launcher_icon);
        return;
      }
    }

    kiba_add_launcher (dock, launcher_icon, g_random_int_range (1, dock->geometry.width),
             g_random_int_range (1, dock->geometry.height), error);

    launcher->launcher_icons = g_list_append (launcher->launcher_icons, launcher_icon);

    return;
  }
  else if (strcmp (entry->key, LAUNCHER_GCONF_PATH "/icon_size") == 0)
    kiba_reload (dock);
  else if (strcmp (entry->key, LAUNCHER_GCONF_PATH "/launcher_alpha") == 0 ||
           strcmp (entry->key, LAUNCHER_GCONF_PATH "/launcher_inactive_alpha") == 0)
  {
    DO_FOR_ALL_OBJECTS
    (
      if (strcmp (object->plugin->name, "launcher") == 0)
        object->rerender = TRUE
    );
  }
  else if (strcmp (entry->key, LAUNCHER_GCONF_PATH "/launcher_enable") == 0)
  {
    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_delete_rope_spring (dock);
      kiba_delete_rope_strings (dock);
    }

    if (dgconf->launcher_enable == FALSE)
    {
      GList *launcher_objects = NULL;

      g_source_remove (launcher->timeout_id);

      DO_FOR_ALL_OBJECTS
      (
        if (strcmp (object->plugin->name, "launcher") == 0)
        {
          if (g_list_find (dock->plugins, object->plugin) != NULL)
            dock->plugins = g_list_remove (dock->plugins, object->plugin);

          kiba_object_free (dock, object);

          launcher_objects = g_list_append (launcher_objects, object);
        }
      );

      for (g_list = launcher_objects; g_list != NULL; g_list = g_list->next)
      {
        KibaObject *object = g_list->data;

        dock->objects = g_list_remove (dock->objects, object);
        object = NULL;
        dock->num_objects--;
      }

      g_list_free (g_list);

      kiba_reload_spacers (dock);

      kiba_update_object_positions (dock);
    }
    else
      kiba_plugin_init (dock, error);

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_add_rope_spring (dock, 1.0);
      kiba_add_rope_strings (dock);
    }

    kiba_layout (dock);
  }
  else
  {
    printf ("ignoring gconf event for %s\n", entry->key);
    return;
  }

  if (error != NULL)
    g_error_free (error);
}

void
kiba_plugin_menu (KibaDock *dock, KibaObject *object, GtkMenu *menu)
{
  KibaLauncherIcon *launcher;
  GtkWidget        *remove_item, *disable_item;
  Display          *display;

  display = XOpenDisplay (NULL);
  if (display == NULL){
    fprintf (stderr, "Unable to open display.\n");
    return;
  }

  launcher = kiba_get_launcher_icon (object);

  disable_item = gtk_menu_item_new_with_label ("Disable Launcher Plugin");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), disable_item);
  g_signal_connect (G_OBJECT (disable_item), "activate", G_CALLBACK (kiba_launcher_disable), NULL);
  gtk_widget_show (disable_item);

  remove_item = gtk_menu_item_new_with_label ("Remove Launcher");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), remove_item);
  g_signal_connect (G_OBJECT (remove_item), "activate", G_CALLBACK (kiba_launcher_remove), launcher);
  gtk_widget_show (remove_item);

  GList *winlist, *l;
  winlist = kiba_launcher_get_winlist (display, launcher);

  for (l = winlist; l != NULL; l = l->next)
  {
    GdkWindow *window = l->data;
    GtkWidget *window_entry, *change_state, *kill_item;
    GtkMenu   *winmenu;
    Window     xwindow;
    char      *text, *cutted_text;
    int        win_state;

    xwindow = gdk_x11_drawable_get_xid (window);

    if (0 == XFetchName (display, xwindow, &text))
      text = kiba_object_get_wmclass (display, xwindow);

    if (strlen (text) > 30)
      cutted_text = g_strdup_printf ("%s...", g_strndup (text, 27*sizeof(char)));
    else
      cutted_text = text;

    winmenu = (GtkMenu *) gtk_menu_new();

    window_entry = gtk_menu_item_new_with_label (cutted_text);
    g_free (cutted_text);
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), window_entry);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (window_entry), GTK_WIDGET(winmenu));
    gtk_widget_show (window_entry);

    kill_item = gtk_menu_item_new_with_label ("Close Programm");
    gtk_menu_shell_append (GTK_MENU_SHELL (winmenu), kill_item);
    g_signal_connect (G_OBJECT (kill_item), "activate", G_CALLBACK (kiba_object_kill_window), window);
    gtk_widget_show (kill_item);

    win_state = kiba_object_get_winstate (display, window);

    if (win_state == 3)
    {
      change_state = gtk_menu_item_new_with_label ("Restore Window");
      gtk_menu_shell_append (GTK_MENU_SHELL (winmenu), change_state);
      g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_object_show_window), window);
      gtk_widget_show (change_state);
    }
    else if (win_state == 1 && xwindow == kiba_object_get_active_window (display))
    {
      change_state = gtk_menu_item_new_with_label ("Iconify Window");
      gtk_menu_shell_append (GTK_MENU_SHELL (winmenu), change_state);
      g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_object_iconify_window),   window);
      gtk_widget_show (change_state);
    }
    else if (win_state == 1)
    {
      change_state = gtk_menu_item_new_with_label ("Focus Window");
      gtk_menu_shell_append (GTK_MENU_SHELL (winmenu), change_state);
      g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_object_focus_window), window);
      gtk_widget_show (change_state);
    }
  }

  XCloseDisplay (display);
}

void
kiba_plugin_release_event (KibaDock *dock, KibaObject *object)
{
  KibaLauncherIcon *launcher_icon;
  GdkWindow        *app_window;
  GSList           *l;
  Display          *display;

  display = XOpenDisplay (NULL);
  if (display == NULL){
    fprintf (stderr, "Unable to open display.\n");
    return;
  }

  launcher_icon = kiba_get_launcher_icon (object);
  app_window    = kiba_launcher_cmd_running (display, launcher_icon);

  for (l = dgconf->launcher_multiple_instances; l != NULL; l = l->next) \
  {
    gchar *exclude_string = l->data;

    if (g_strrstr (launcher_icon->exec_command, exclude_string))
      app_window = NULL;
  }

  if (app_window == NULL)
  {
    if (launcher_icon->exec_command == NULL)
      return;

    dock->exec_object = object;

    g_idle_add (kiba_launcher_exec, launcher_icon);
  }
  else
  {
    int      win_state, focus_revert_to;
    Window   focus_window;

    win_state = kiba_object_get_winstate (display, app_window);

    XGetInputFocus (display, &focus_window, &focus_revert_to);

    /* window is iconified */
    if (win_state == 3)
      kiba_object_show_window (GTK_WIDGET (dock), app_window);
    /* window has focus */
    else if (win_state == 1 && gdk_x11_drawable_get_xid (app_window) == focus_window)
      kiba_object_iconify_window (GTK_WIDGET (dock), app_window);
    /* window is below others */
    else if (win_state == 1)
      kiba_object_focus_window (GTK_WIDGET (dock), app_window);
    /* window isnt shown */
//     else if (win_state == 0)
//     {
//     }
  }
  XCloseDisplay (display);
}

static KibaLauncherIcon *
kiba_launcher_new (KibaDock *dock, KibaLauncher *launcher, char *gconf_path, GError *error)
{
  KibaLauncherIcon *launcher_icon;
  char             *desktop_file, *key, *new_desktop_entry, *system_path;

  key         = g_strdup_printf ("%s/file", gconf_path);
  system_path = gconf_client_get_string (dgconf->client, key, &error);
  if (error != NULL)
  {
    g_message ("Failed to merge .desktop file to home notifications: %s", error->message);
    g_clear_error (&error);
    return NULL;
  }

  if (system_path == NULL || strlen (system_path) < 2)
    return NULL;

  launcher_icon                      = g_new0 (KibaLauncherIcon, 1);
  launcher_icon->object              = g_new0 (KibaObject, 1);
  launcher_icon->object->key         = g_strdup_printf ("%s", gconf_path);

//   launcher_icon->object->position_notify        = TRUE;

  launcher_icon->object->normal_size = dgconf->icon_size;

  kiba_delete_rope_spring (dock);

  new_desktop_entry = merge_desktop_to_home (dock, system_path, launcher_icon->object->key);
  if (new_desktop_entry == NULL)
  {
    g_free (key);
    g_free (launcher_icon->object);
    g_free (launcher_icon);
    return NULL;
  }

  desktop_file = gconf_client_get_string (dgconf->client, key, &error);
  if (error != NULL)
  {
    g_message ("error getting desktop file path for launcher: %s\n",
                error->message);
    g_clear_error (&error);
    g_free (key);
    g_free (launcher_icon->object);
    g_free (launcher_icon);
    return NULL;
  }

  parse_desktop_file (launcher_icon, desktop_file, &error);
  free (desktop_file);
  if (error != NULL)
  {
    g_message ("error parsing desktop file for %s: %s\n",
                launcher_icon->object->key, error->message);
    g_clear_error (&error);
    g_free (key);
    g_free (launcher_icon->object);
    g_free (launcher_icon);
    return NULL;
  }

  #ifdef HAVE_SVG
  if (!kiba_load_svg (launcher_icon->object, launcher_icon->active_icon, &launcher_icon->object->bg_svg_handle, &error))
  #endif
    kiba_load_png (dock, launcher_icon->object, launcher_icon->active_icon, &launcher_icon->object->bg_pixbuf, &error);

  g_free (key);

  return launcher_icon;
}

static void
kiba_add_launcher (KibaDock *dock, KibaLauncherIcon *launcher, int x, int y, GError *error)
{
  AddSpacerClosure closure;

  closure.dock           = dock;
  closure.object         = akamaru_model_add_object (&dock->model, x, y, 12, NULL);
  closure.spacer_size    = launcher->object->size + dgconf->object_space;
  closure.object->radius = closure.spacer_size / 2.0;

  launcher->object->spring =
    akamaru_model_add_spring (&dock->model, dock->anchor->object, closure.object, 0);

  akamaru_model_for_each_object (&dock->model, add_spacer, &closure);
  dock->objects = g_list_append (dock->objects, launcher->object);
  dock->num_objects++;

  launcher->object->object           = closure.object;
  launcher->object->object->friction = dock->model.friction;
  launcher->object->win_position.x   = launcher->object->object->position.x;
  launcher->object->win_position.y   = launcher->object->object->position.y;
  launcher->object->rotate_angle     = 0.0;
  launcher->object->scale_factor     = 1.0;
  launcher->blend_alpha              = dgconf->launcher_inactive_alpha;
  launcher->object->normal_size      = dgconf->icon_size;
  launcher->object->name             = launcher->name;
  launcher->object->exec             = FALSE;

//   kiba_object_update_position (dock, launcher->object, reload);

  if (GTK_WIDGET_REALIZED (GTK_WIDGET (dock)))
    kiba_object_create_window (dock, launcher->object);

  if (GTK_WIDGET_MAPPED (GTK_WIDGET (dock)))
    gdk_window_show (launcher->object->window);

  launcher->notify_id = gconf_client_notify_add
   (dgconf->client, g_strdup (launcher->object->key),
    launcher_data_changed_callback, launcher, NULL, &error);
  if (error != NULL)
  {
    fprintf (stderr, "Failed to add gconf notify for %s\n" "notifications: %s\n",
             launcher->name, error->message);
    g_clear_error (&error);
  }

  kiba_layout (dock);
}

static gboolean
kiba_launcher_remove (GtkWidget* widget, gpointer data)
{
  KibaLauncher        *launcher;
  KibaLauncherIcon    *launcher_icon = data;
  GConfUnsetFlags      flags;
  char                *launcher_path;

  launcher = launcher_icon->object->plugin->plugin_struct;
  gconf_client_notify_remove (lgconf->client, launcher_icon->notify_id);

  GSList *entries, *l;
  entries = gconf_client_all_entries (lgconf->client, launcher_icon->object->key, NULL);
  for (l = entries; l != NULL; l = l->next)
  {
    GConfEntry *entry = l->data;

    gconf_client_unset (lgconf->client, entry->key, NULL);
  }

  /* remove gconf entry and .desktop file */
//TODO this doesnt unset keys wich are associated with an schema
  gconf_client_recursive_unset (lgconf->client, launcher_icon->object->key, flags, NULL);
  launcher_path = g_strdup_printf ("%s/.kiba-dock/%s", g_get_home_dir (), launcher_icon->desktop_key);
  if (g_file_test (launcher_path, (G_FILE_TEST_EXISTS)))
    g_remove (launcher_path);
  g_free (launcher_path);

  akamaru_model_delete_spring (&launcher->dock->model, launcher_icon->object->spring);
  launcher_icon->object->spring = NULL;

  /* rearrange the rope */
  if (strcmp (lgconf->object_model, "rope") == 0)
    kiba_update_rope_positions (ldock, launcher_icon->object);

  if (strcmp (lgconf->object_model, "wobbly") == 0 ||
      strcmp (lgconf->object_model, "spring_for_all") == 0)
    kiba_delete_wobbly_spring (launcher->dock);

  kiba_object_free (ldock, launcher_icon->object);
  launcher->dock->objects = g_list_remove (launcher->dock->objects, launcher_icon->object);
  launcher_icon->object = NULL;

  kiba_reload_spacers (launcher->dock);

  if (strcmp (lgconf->object_model, "rope") == 0)
    kiba_add_rope_spring (launcher->dock, 1.0);

  if (strcmp (lgconf->object_model, "wobbly") == 0 ||
      strcmp (lgconf->object_model, "spring_for_all") == 0)
    kiba_add_wobbly_spring (launcher->dock);

  kiba_layout (launcher->dock);

  if (lgconf->text_enable)
  {
    launcher->dock->text->hidden = TRUE;
    gdk_window_hide (launcher->dock->text->window);
    launcher->dock->text->requests = 0;
  }

//   kiba_update_object_positions (launcher->dock);

  g_free (launcher_icon);

  return TRUE;
}

static void
kiba_launcher_disable (void)
{
  GConfClient *gconf_client;
  GError *error = NULL;

  gconf_client = gconf_client_get_default ();

  if (gconf_client_set_bool (gconf_client, LAUNCHER_GCONF_PATH "/launcher_enable", FALSE, &error) == FALSE)
    fprintf (stderr, "failed to disable the taskbar\n");

  if (error != NULL)
  {
    fprintf (stderr, "%s\n", error->message);
    g_error_free (error);
  }
}

static KibaLauncher *
kiba_get_launcher (KibaDock *dock)
{
  GList *l;

  for (l = dock->plugins; l != NULL; l = l->next) \
  {
    KibaPlugin *plugin = l->data;

    if (g_strrstr (plugin->name, "launcher"))
      return plugin->plugin_struct;
  }

  return NULL;
}

static KibaLauncherIcon *
kiba_get_launcher_icon (KibaObject *object)
{
  KibaLauncher *launcher;
  GList        *l;

  launcher = object->plugin->plugin_struct;

  for (l = launcher->launcher_icons; l != NULL; l = l->next) \
  {
    KibaLauncherIcon *launcher_icon = l->data;

    if (launcher_icon->object == object)
      return launcher_icon;
  }

  return NULL;
}

static gboolean
kiba_replace_desktop_entry (KibaLauncherIcon *launcher, char *desktop_entry, char *new_value, GError *error)
{
  FILE        *desktop_file;
  gsize        file_length;
  gchar       *puffer, *begin, *end, *new_entry, *desktop_path;

  desktop_path = g_strdup_printf ("%s/.kiba-dock/%s", g_get_home_dir (), launcher->desktop_key);

  /* open desktop file */
  if ((desktop_file = fopen (desktop_path, "r")) == NULL)
  {
    g_free (desktop_path);
    return FALSE;
  }

  /* obtain file size */
  fseek (desktop_file , 0 , SEEK_END);
  file_length = ftell (desktop_file);
  rewind (desktop_file);
  fclose (desktop_file);

  /* allocate memory to contain the whole file */
  puffer = (char*) malloc (file_length);
  if (puffer == NULL)
  {
    g_free (desktop_path);
    return FALSE;
  }

  /* load file into puffer */
  g_file_get_contents (desktop_path, &puffer, &file_length, &error);
  if (error != NULL)
  {
    g_message ("cant load file %s into buffer notifications: %s", desktop_path, error->message);
    g_clear_error (&error);
    g_free (desktop_path);
    g_free (puffer);
    return FALSE;
  }

  /* split desktop file */
  end = g_strstr_len (puffer, strlen(puffer), desktop_entry);
  register int count = 1;
  while (end[count] != '\n')
    count++;
  strcpy (end, end + count);

  begin = (char*) calloc (sizeof (char*), strlen (puffer) - strlen (end));
  strncpy (begin, puffer, strlen (puffer) - strlen (end));

  /* write new entry */
  new_entry = g_strdup_printf ("%s%s%s%s", begin, desktop_entry, new_value, end);
  g_file_set_contents (desktop_path, new_entry, strlen (new_entry), &error);
  if (error != NULL)
  {
    g_message ("error writing the buffer to file %s notifications: %s", desktop_path, error->message);
    g_clear_error (&error);
  }

  g_free (new_entry);
  g_free (desktop_path);
  g_free (puffer);
  g_free (begin);

  return TRUE;
}

static void
launcher_data_changed_callback (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
  KibaLauncher     *launcher;
  KibaLauncherIcon *launcher_icon = user_data;
  GError           *error         = NULL;

  if (launcher_icon == NULL)
//  || !launcher_icon->object->position_notify)
    return;

  launcher = launcher_icon->object->plugin->plugin_struct;

  if (g_str_has_suffix (entry->key, "/name"))
  {
    launcher_icon->name = gconf_client_get_string (client, entry->key, &error);
    if (error != NULL)
    {
      g_message ("Failed get launcher name for %s notifications: %s",
                launcher_icon->object->key, error->message);
      g_error_free (error);
      return;
    }
    if (launcher_icon->name == NULL)
      return;
    kiba_replace_desktop_entry (launcher_icon, "\nName=", launcher_icon->name, error);

    /* update pango for launcher */
    kiba_text_update_pango_for_object (launcher_icon->object, ldock->gconf, ldock->text);
  }
  else if (g_str_has_suffix (entry->key, "/exec"))
  {
    launcher_icon->exec_command = gconf_client_get_string (client, entry->key, &error);
    if (error != NULL)
    {
      g_message ("Failed get launcher exec for %s notifications: %s",
                launcher_icon->object->key, error->message);
      g_error_free (error);
      return;
    }
    if (launcher_icon->exec_command == NULL)
      return;
    kiba_replace_desktop_entry (launcher_icon, "\nExec=", launcher_icon->exec_command, error);
  }
//   else if (g_str_has_suffix (entry->key, "/launcher_position"))
//     kiba_object_position_changed (launcher->dock, launcher_icon->object, entry);
  else if (g_str_has_suffix (entry->key, "/active_icon") && launcher_icon->running)
    launcher_icon->active_icon = kiba_launcher_load_new_icon (ldock, launcher_icon, launcher_icon->active_icon, entry->key, error);
  else if (g_str_has_suffix (entry->key, "/inactive_icon") && !launcher_icon->running)
    launcher_icon->inactive_icon = kiba_launcher_load_new_icon (ldock, launcher_icon, launcher_icon->inactive_icon, entry->key, error);

  if (error != NULL)
    g_error_free (error);
}

static char*
kiba_launcher_load_new_icon (KibaDock *dock, KibaLauncherIcon *launcher_icon, char* icon, char* key, GError *error)
{
  char *icon_buf = g_strdup (icon);

  icon = gconf_client_get_string (dock->gconf->client, key, &error);

  if (icon == NULL)
  {
    if (error != NULL)
    {
      g_message ("Failed get launcher icon for %s notifications: %s",
                 launcher_icon->object->key, error->message);
      g_error_free (error);
    }

    icon = g_strdup (icon_buf);
    g_free (icon_buf);

    #ifdef HAVE_SVG
    if (!kiba_load_svg (launcher_icon->object, icon, &launcher_icon->object->bg_svg_handle, &error))
    #endif
      kiba_load_png (dock, launcher_icon->object, icon, &launcher_icon->object->bg_pixbuf, &error);

    return icon;
  }

  #ifdef HAVE_SVG
  if (!kiba_load_svg (launcher_icon->object, icon, &launcher_icon->object->bg_svg_handle, &error))
  #endif
    kiba_load_png (dock, launcher_icon->object, icon, &launcher_icon->object->bg_pixbuf, &error);

  if (launcher_icon->object->bg_pixbuf != NULL || launcher_icon->object->bg_svg_handle != NULL)
  {
    /* update surface buffer */
    cairo_t *cr = NULL;
    cr = dock_cairo_create (launcher_icon->object);
    if (cr == NULL)
    {
      g_free (icon_buf);
      return;
    }
    fill_surface_buffer (cr, launcher_icon->object, TRUE);
    cairo_destroy (cr);

    launcher_icon->object->rerender = TRUE;
    kiba_replace_desktop_entry (launcher_icon, "\nIcon=", icon, error);
  }
  else
  {
    g_message ("cant load icon %s\n", icon);
    icon = g_strdup (icon_buf);

    #ifdef HAVE_SVG
    if (!kiba_load_svg (launcher_icon->object, icon, &launcher_icon->object->bg_svg_handle, &error))
    #endif
      kiba_load_png (dock, launcher_icon->object, icon, &launcher_icon->object->bg_pixbuf, &error);

    g_free (icon_buf);
  }

  return icon_buf;
}

static void
kiba_merge_new_desktop_file (KibaLauncherIcon *launcher_icon, GConfClient *gconf_client, GError *error)
{
  KibaLauncher *launcher;
  gchar        *key;

  launcher = launcher_icon->object->plugin->plugin_struct;

  key = g_strdup_printf ("%s/name", launcher_icon->object->key);
  gconf_client_set_string (gconf_client, key, launcher_icon->name, &error);
  if (error != NULL)
  {
    g_message ("Failed to set name Option for launcher notifications: %s", error->message);
    g_clear_error (&error);
  }
  else
  {
    gconf_engine_associate_schema (lgconf->client->engine, key,
                               "/schemas" LAUNCHER_GCONF_PATH "/launchers/name", &error);

    if (error != NULL) {
      printf ("failed to set schema for %s: %s\n", key, error->message);
      g_clear_error (&error);
    }
  }

  key = g_strdup_printf ("%s/exec", launcher_icon->object->key);
  gconf_client_set_string (gconf_client, key, launcher_icon->exec_command, &error);
  if (error != NULL)
  {
    g_message ("Failed to set exec Option for launcher notifications: %s", error->message);
    g_clear_error (&error);
  }
  else
  {
    gconf_engine_associate_schema (lgconf->client->engine, key,
                               "/schemas" LAUNCHER_GCONF_PATH "/launchers/exec", &error);

    if (error != NULL) {
      printf ("failed to set schema for %s: %s\n", key, error->message);
      g_clear_error (&error);
    }
  }

  key = g_strdup_printf ("%s/active_icon", launcher_icon->object->key);
  gconf_client_set_string (gconf_client, key, launcher_icon->active_icon, &error);
  if (error != NULL)
  {
    g_message ("Failed to set icon Option for launcher notifications: %s", error->message);
    g_clear_error (&error);
  }
  else
  {
    gconf_engine_associate_schema (lgconf->client->engine, key,
                               "/schemas" LAUNCHER_GCONF_PATH "/launchers/active_icon", &error);

    if (error != NULL) {
      printf ("failed to set schema for %s: %s\n", key, error->message);
      g_clear_error (&error);
    }
  }

  if (launcher_icon->inactive_icon == NULL)
    launcher_icon->inactive_icon = launcher_icon->active_icon;

  key = g_strdup_printf ("%s/inactive_icon", launcher_icon->object->key);
  gconf_client_set_string (gconf_client, key, launcher_icon->inactive_icon, &error);
  if (error != NULL)
  {
    g_message ("Failed to set icon Option for launcher notifications: %s", error->message);
    g_clear_error (&error);
  }
  else
  {
    gconf_engine_associate_schema (lgconf->client->engine, key,
                               "/schemas" LAUNCHER_GCONF_PATH "/launchers/active_icon", &error);

    if (error != NULL) {
      printf ("failed to set schema for %s: %s\n", key, error->message);
      g_clear_error (&error);
    }
  }

  free (key);
}

static void
parse_desktop_file (KibaLauncherIcon *launcher_icon, const char *uri, GError **error)
{
  static const char file_uri_prefix[] = "file://";
  static const char desktop_file_suffix[] = ".desktop";
  static const char desktop_entry_group[] = "Desktop Entry";
  static const struct { const char *name; size_t offset; } fields[] = {
    { "Icon",    G_STRUCT_OFFSET (KibaLauncherIcon, active_icon) },
    { "Exec",    G_STRUCT_OFFSET (KibaLauncherIcon, exec_command) },
    { "Name",    G_STRUCT_OFFSET (KibaLauncherIcon, name) },
  };

  GKeyFile   *file;
  const char *filename;
  char       *s;
  int         i;

  if (!g_str_has_suffix (uri, desktop_file_suffix)) {
    g_set_error (error, KIBA_ERROR, 0, "%s is not a .desktop file", uri);
    return;
  }
  if (g_str_has_prefix (uri, file_uri_prefix))
    filename = uri + strlen (file_uri_prefix);
  else
    filename = uri;

  file = g_key_file_new ();
  if (!g_key_file_load_from_file (file, filename, 0, error))
    return;

  launcher_icon->desktop_key =  g_strndup (g_strrstr (filename, "/") + 1,
          strlen (g_strrstr (filename, "/") + 1));

  for (i = 0; i < G_N_ELEMENTS (fields); i++) {
    s = g_key_file_get_value (file, desktop_entry_group, fields[i].name, error);
    if (s == NULL)
      return;

    G_STRUCT_MEMBER (char *, launcher_icon, fields[i].offset) = s;
  }
}

static gboolean
kiba_launcher_exec (gpointer data)
{
  KibaLauncherIcon *launcher_icon = data;
  int           argc;
  char        **argv, **p, **q;
  GError       *error = NULL;

  /* FIXME: Support drag and drop on objects with %U and %F */
  if (!g_shell_parse_argv (launcher_icon->exec_command, &argc, &argv, &error))
    {
      g_message ("failed to parse '%s': %s\n", launcher_icon->exec_command, error->message);
      g_error_free (error);
      return FALSE;
    }

  for (p = argv, q = argv; *p != NULL; p++)
    {
      if (*p[0] == '%' && strlen (*p) == 2)
	switch (*p[1])
	  {
	  case 'U':
	  case 'u':
	  case 'F':
	  case 'f':
	  default:
	    g_free (*p);
	    break;
	  }
      else
	*q++ = *p;
    }
  *q++ = NULL;

  g_spawn_async (NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
                 NULL, NULL, NULL, &error);
  if (error != NULL)
    g_message ("failed to launch %s: %s\n", argv[0], error->message);
  g_strfreev (argv);

  return FALSE;
}

GdkWindow *
kiba_launcher_cmd_running (Display *display, KibaLauncherIcon *launcher_icon)
{
  GdkScreen        *screen;
  GList            *root_children, *l;

  screen     = gdk_screen_get_default ();

  root_children = gdk_window_get_children (gdk_get_default_root_window());

  if (root_children == NULL)
    return NULL;

  for (l = root_children; l != NULL; l = l->next)
  {
    GdkWindow  *window = l->data;
    Window      xwindow;

    xwindow = gdk_x11_drawable_get_xid (window);

    XTextProperty text_property;
    Atom cmd_atom = XInternAtom (display, "WM_CLASS", TRUE);
    if (XGetTextProperty (display, xwindow, &text_property, cmd_atom))
    {
      if (g_strrstr (launcher_icon->exec_command, text_property.value))
        return window;
    }
  }

  return NULL;
}

GList *
kiba_launcher_get_winlist (Display *display, KibaLauncherIcon *launcher_icon)
{
  GdkScreen        *screen;
  GList            *root_children, *app_windows = NULL, *l;

  screen        = gdk_screen_get_default ();
  root_children = gdk_window_get_children (gdk_get_default_root_window());

  if (root_children == NULL)
    return NULL;

  for (l = root_children; l != NULL; l = l->next)
  {
    GdkWindow  *window = l->data;
    Window      xwindow;

    xwindow = gdk_x11_drawable_get_xid (window);

    XTextProperty text_property;
    Atom cmd_atom = XInternAtom (display, "WM_CLASS", TRUE);
    if (XGetTextProperty (display, xwindow, &text_property, cmd_atom))
    {
      if (g_strrstr (launcher_icon->exec_command, text_property.value))
        app_windows = g_list_append (app_windows, window);
    }
  }

  return app_windows;
}

static gboolean
kiba_launcher_time_handler (KibaLauncher *launcher)
{
  GList   *launchers;
  GError  *error = NULL;

  if (launcher == NULL)
    return FALSE;

  /* check if the app wich the launcher would start is running */
  for (launchers = ldock->objects; launchers != NULL; launchers = launchers->next)
  {
    KibaObject *object = launchers->data;

    if (g_strrstr (object->plugin->name, "launcher"))
    {
      KibaLauncherIcon *launcher_icon;
      Display          *display;
      GdkWindow        *app_window;

      launcher_icon = kiba_get_launcher_icon (object);
      if (launcher_icon == NULL)
        return TRUE;

      display = XOpenDisplay (NULL);
      if (display == NULL){
          fprintf (stderr, "Unable to open display.\n");
          return TRUE;
      }

      app_window = kiba_launcher_cmd_running (display, launcher_icon);

      /* bounce when urgency flag is set */
      if (app_window != NULL && kiba_object_window_urgency (display, app_window))
      {
        if (ldock->d->orientation == GTK_ORIENTATION_VERTICAL)
          object->object->previous_position.x += (ldock->d->halignment - 1) * 1.0;
        else
          object->object->previous_position.y += (ldock->d->valignment - 1) * 1.0;
      }

      XCloseDisplay (display);

      /* set the running state */
      if (app_window != NULL && !launcher_icon->running)
      {
        launcher_icon->running = TRUE;

        /* load active icon */
        if (launcher_icon->active_icon != NULL && launcher_icon->inactive_icon != NULL &&
            !g_strrstr (launcher_icon->active_icon, launcher_icon->inactive_icon))
          launcher_icon->active_icon = kiba_launcher_load_new_icon (ldock, launcher_icon, launcher_icon->active_icon, g_strdup_printf ("%s/active_icon", launcher_icon->object->key), error);

        launcher_icon->object->rerender = TRUE;
      }
      if (app_window == NULL && launcher_icon->running)
      {
        launcher_icon->running = FALSE;

        /* load inactive icon */
        if (launcher_icon->active_icon != NULL && launcher_icon->inactive_icon != NULL &&
            !g_strrstr (launcher_icon->active_icon, launcher_icon->inactive_icon))
          launcher_icon->inactive_icon = kiba_launcher_load_new_icon (ldock, launcher_icon, launcher_icon->inactive_icon, g_strdup_printf ("%s/inactive_icon", launcher_icon->object->key), error);

        launcher_icon->object->rerender = TRUE;
      }
    }
  }

  return TRUE;
}
