/*                                           -*- mode: c; c-basic-offset: 2 -*-
 *
 * This is kiba, a fun GNOME dock
 */

#include <time.h>
#include <math.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>
#include <cairo.h>
#include <glibtop.h>
#include <glibtop/mem.h>

#include <X11/Xlib.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#include <akamaru.h>

#include <kiba-dock.h>
#include <kiba-akamaru.h>
#include <kiba-render.h>
#include <kiba-gconf.h>
#include <kiba-object.h>

#define SECOND_INTERVAL  1000
#define MEMORY_GCONF_PATH KIBA_GCONF_PATH"/plugins/memory"

typedef struct _KibaMemory   KibaMemory;

struct _KibaMemory {
  KibaDock        *dock;
  KibaObject      *object;

  unsigned long    free_memory;
  double           free_memory_percent;

  gint             gconf_notify_id;
  gint             timeout_id;
};

void
kiba_plugin_init              (KibaDock *dock, GError *error);
char*
kiba_plugin_name              (void);
void
kiba_plugin_paint             (KibaDock *dock, KibaObject *object, cairo_t *pixmap_context);
void
kiba_plugin_timeout           (KibaDock *dock, KibaObject *object, GError *error);
void
kiba_plugin_notify            (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data);
void
kiba_plugin_menu              (KibaDock *dock, KibaObject *object, GtkMenu *menu);
void
kiba_plugin_release_event     (KibaDock *dock, KibaObject *object);

static KibaMemory*
kiba_memory_new               (KibaDock *dock);
static gboolean
kiba_add_memory               (KibaDock *dock, KibaMemory *memory, gint x, gint y, GError *error);
static cairo_surface_t*
draw_memory_background        (KibaMemory *memory, cairo_t *memory_context);
static gboolean
kiba_memory_time_handler      (KibaDock *dock);
static KibaMemory *
kiba_get_memory               (KibaDock *dock);
static void
kiba_memory_disable           (void);

void
kiba_plugin_init (KibaDock *dock, GError *error)
{
  KibaPlugin *plugin;
  KibaMemory *memory;

  if (!dgconf->memory_enable)
    return;

  glibtop_init();

  plugin = g_new0 (KibaPlugin, 1);
  memory = kiba_memory_new (dock);

  if (memory == NULL || plugin == NULL)
    return;

  plugin->name           = g_strdup ("memory");
  plugin->plugin_struct  = memory;
  memory->object->plugin = plugin;

  kiba_add_memory (dock, memory, g_random_int_range (1, dock->geometry.width),
                  g_random_int_range (1, dock->geometry.height), error);

  memory->timeout_id = g_timeout_add (SECOND_INTERVAL, (GSourceFunc) kiba_memory_time_handler, (gpointer) dock);

  dock->plugins = g_list_append (dock->plugins, plugin);

  return;
}

char*
kiba_plugin_name (void)
{
  return "memory";
}

void
kiba_plugin_paint (KibaDock *dock, KibaObject *object, cairo_t *pixmap_context)
{
  KibaPlugin      *plugin;
  KibaMemory      *memory;
  cairo_pattern_t *pattern;
  double           r, g, b, offset;

  if (!dgconf->memory_enable || dock->hiding)
    return;

  plugin = kiba_get_plugin_struct (dock, "memory");
  if (plugin == NULL)
    return;

  memory = (KibaMemory*) plugin->plugin_struct;
  if (memory == NULL)
    return;

  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_SOURCE);

  /* translate/rotate context for different positions */
  if (dock->d->anchor.y == 0)
  {
    cairo_translate (pixmap_context, object->normal_size, object->normal_size);
    cairo_rotate (pixmap_context, M_PI);
  }
  else if (dock->d->anchor.x == 0)
  {
    cairo_translate (pixmap_context, object->normal_size,  .0);
    cairo_rotate (pixmap_context, M_PI/2);
  }
  else if (dock->d->anchor.x == 4)
  {
    cairo_translate (pixmap_context, .0,  object->normal_size);
    cairo_rotate (pixmap_context, 3*M_PI/2);
  }

  if (object->rotate_angle != 0.0)
    offset = -object->normal_size / 2.0;
  else
    offset =  .0f;

  cairo_set_source_surface (pixmap_context, object->background_surface, offset, offset);
  cairo_paint_with_alpha (pixmap_context, dgconf->memory_alpha);

  cairo_restore (pixmap_context);

  cairo_set_operator (pixmap_context, CAIRO_OPERATOR_ATOP);

  convert_color (dgconf->memory_progress_color, &r, &g, &b);
  cairo_set_source_rgba (pixmap_context, r, g, b, dgconf->memory_progress_alpha);
  cairo_set_line_width (pixmap_context, 10.0);

  cairo_move_to (pixmap_context, object->normal_size * memory->free_memory_percent, 0.0);
  cairo_rel_line_to (pixmap_context, 0.0, object->size);
  cairo_rel_line_to (pixmap_context, -object->size, 0.0);
  cairo_rel_line_to (pixmap_context, 0.0, -object->size);
  cairo_rel_line_to (pixmap_context, object->size, 0.0);
  cairo_close_path (pixmap_context);

  if (g_str_has_suffix (dgconf->memory_progress_bg, ".png"))
  {
    pattern = kiba_get_png_pattern (dock, pixmap_context, dgconf->memory_progress_bg, TRUE);

    cairo_set_source (pixmap_context, pattern);
  }

  cairo_fill_preserve (pixmap_context);

  if (object->prelight)
  {
    convert_color (dgconf->prelight_color, &r, &g, &b);
    cairo_set_source_rgba (pixmap_context, r, g, b, dgconf->prelight_alpha);
    cairo_set_operator (pixmap_context, CAIRO_OPERATOR_ATOP);
    cairo_paint (pixmap_context);
  }

  cairo_restore (pixmap_context);
}

void
kiba_plugin_timeout (KibaDock *dock, KibaObject *object, GError *error)
{
  KibaMemory *memory;

  if (!dgconf->memory_enable || object->window == NULL)
    return;

  memory = kiba_get_memory (dock);

  if (memory->object->bgsurface_needs_update == TRUE ||
      memory->object->fgsurface_needs_update == TRUE)
  {
    cairo_t*         memory_context = NULL;

    memory_context = dock_cairo_create (memory->object);
    cairo_set_operator (memory_context, CAIRO_OPERATOR_SOURCE);

    if (dgconf->memory_face != NULL &&
       (strstr (dgconf->memory_face, "svg") != NULL || strstr (dgconf->memory_face, "png") != NULL))
    {
      #ifdef HAVE_SVG
      if (!kiba_load_svg (memory->object, dgconf->memory_face, &memory->object->bg_svg_handle, &error))
      #endif
        kiba_load_png (dock, memory->object, dgconf->memory_face, &memory->object->bg_pixbuf, &error);

      kiba_update_surface_buffer (object, object->bgsurface_needs_update);
    }
    else if (object->bgsurface_needs_update == TRUE)
      object->background_surface = draw_memory_background (memory, memory_context);

    memory->object->bgsurface_needs_update = FALSE;
    memory->object->fgsurface_needs_update = FALSE;

    memory->object->rerender = TRUE;

    cairo_destroy (memory_context);
  }

  if (memory->object->drawn_rotate_angle != 0.0 ||
      memory->object->drawn_rotate_angle != memory->object->rotate_angle)
    memory->object->rerender = TRUE;
}

void
kiba_plugin_notify (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
  KibaDock   *dock;
  KibaMemory *memory;
  GError     *error;

  dock   = user_data;
  memory = kiba_get_memory (dock);

  set_option (dock, entry);

  if (strcmp (entry->key, MEMORY_GCONF_PATH "/memory_object_size") == 0)
  {
    kiba_reload (dock);
/*
    dock->clock->object->normal_size   = (double) gconf_value_get_float (entry->value);
    dock->clock->object->size          = (double) gconf_value_get_float (entry->value);

    dock->clock->object->bgsurface_needs_update = TRUE;
    dock->clock->object->fgsurface_needs_update = TRUE;

    gdk_window_resize (dock->clock->object->window, dock->clock->object->size, dock->clock->object->size);

    kiba_reload_spacers (dock);

    kiba_layout (dock);
*/
  }
  else if (strcmp (entry->key, MEMORY_GCONF_PATH "/memory_enable") == 0)
  {
    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_delete_rope_spring (dock);
      kiba_delete_rope_strings (dock);
    }

    if (dgconf->memory_enable == FALSE && memory != NULL)
    {
      dock->plugins = g_list_remove (dock->plugins, memory->object->plugin);

      g_source_remove (memory->timeout_id);

      kiba_object_free (dock, memory->object);
      dock->objects = g_list_remove (dock->objects, memory->object);
      memory->object = NULL;
      dock->num_objects--;

      memory = NULL;

      kiba_reload_spacers (dock);

      kiba_update_object_positions (dock);
    }
    else
      kiba_plugin_init (dock, error);

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_add_rope_spring (dock, 1.0);
      kiba_add_rope_strings (dock);
    }

    kiba_layout (dock);
  }
  else if (memory != NULL &&
          (strcmp (entry->key, MEMORY_GCONF_PATH "/memory_face") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_border_width") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_color") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_border_color") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_border_alpha") == 0))
    memory->object->bgsurface_needs_update = TRUE;
  else if (memory != NULL &&
          (strcmp (entry->key, MEMORY_GCONF_PATH "/memory_alpha") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_progress_bg") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_progress_alpha") == 0 ||
           strcmp (entry->key, MEMORY_GCONF_PATH "/memory_progress_color") == 0))
    memory->object->rerender = TRUE;
  else if (g_str_has_suffix (entry->key, "/memory_position"))
    kiba_object_position_changed (dock, memory->object, entry);
}

void
kiba_plugin_menu (KibaDock *dock, KibaObject *object, GtkMenu *menu)
{
  GtkWidget  *disable_item;

  disable_item = gtk_menu_item_new_with_label ("Disable Memory");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), disable_item);
  g_signal_connect (G_OBJECT (disable_item), "activate", G_CALLBACK (kiba_memory_disable), NULL);
  gtk_widget_show (disable_item);
}

void
kiba_plugin_release_event (KibaDock *dock, KibaObject *object)
{

}

static KibaMemory *
kiba_memory_new (KibaDock *dock)
{
  KibaMemory *memory;

  memory                           = g_new0 (KibaMemory, 1);
  memory->object                   = g_new0 (KibaObject, 1);
//   memory->object->position_notify  = TRUE;

  return memory;
}

static gboolean
kiba_add_memory (KibaDock *dock, KibaMemory *memory, gint x, gint y, GError *error)
{
  AddSpacerClosure  closure;

  closure.dock           = dock;
  closure.object         = akamaru_model_add_object (&dock->model, x, y, 12, NULL);
  closure.spacer_size    = (double) dgconf->memory_object_size + dgconf->object_space;
  closure.object->radius = closure.spacer_size / 2.0;

  memory->object->spring =
    akamaru_model_add_spring (&dock->model, dock->anchor->object, closure.object, 0);

  dock->objects = g_list_append (dock->objects, memory->object);
  dock->num_objects++;

  akamaru_model_for_each_object (&dock->model, add_spacer, &closure);

  memory->object->object           = closure.object;
  memory->dock                     = dock;
  memory->object->object->friction = dock->model.friction;
  memory->object->win_position.x   = memory->object->object->position.x;
  memory->object->win_position.y   = memory->object->object->position.y;
  memory->object->normal_size      = dgconf->memory_object_size;
  memory->object->scale_factor     = 1.0;
  memory->object->key              = g_strdup (MEMORY_GCONF_PATH);
  memory->object->name             = g_strdup ("memory");

//   kiba_object_update_position (dock, memory->object, reload);

  if (GTK_WIDGET_REALIZED (GTK_WIDGET (dock)))
    kiba_object_create_window (dock, memory->object);

  if (GTK_WIDGET_MAPPED (GTK_WIDGET (dock)))
    gdk_window_show (memory->object->window);

  if (memory->dgconf->memory_face != NULL)
  {
    #ifdef HAVE_SVG
    if (!kiba_load_svg (memory->object, dgconf->memory_face, &memory->object->bg_svg_handle, &error))
    #endif
      kiba_load_png (dock, memory->object, dgconf->memory_face, &memory->object->bg_pixbuf, &error);
  }

  kiba_layout (dock);
}

static KibaMemory *
kiba_get_memory (KibaDock *dock)
{
  DO_FOR_ALL_OBJECTS
  (
    if (g_strrstr (object->plugin->name, "memory"))
      return object->plugin->plugin_struct;
  )

  return NULL;
}

/* draws fallback background when no svg or png is defined */
static cairo_surface_t*
draw_memory_background (KibaMemory *memory, cairo_t *memory_context)
{
  cairo_t*         cr = NULL;
  cairo_surface_t* new_surface     = NULL;
  gint     x0, y0, x1, y1;
  double   r, g, b;

  memory->object->buffer_size = memory->object->size;

  /* border points */
  x0 = memory->dgconf->memory_border_width / 2.0;
  x1 = memory->object->size - memory->dgconf->memory_border_width / 2.0;
  y0 = (memory->object->size - memory->dgconf->memory_border_width) * 3.0/4.0;
  y1 = memory->dgconf->memory_border_width / 2.0;

  cairo_surface_destroy (memory->object->background_surface);

  new_surface = cairo_surface_create_similar (cairo_get_target (memory_context), CAIRO_CONTENT_COLOR_ALPHA, memory->object->size, memory->object->size);
  if (cairo_surface_status (new_surface) != CAIRO_STATUS_SUCCESS)
    return NULL;

  cr = cairo_create (new_surface);
  if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
    return NULL;

  cairo_translate (cr, 0.0, (memory->object->size - memory->dgconf->memory_border_width) * 1.0/4.0);

  /* clear context */
  cairo_set_source_rgba (cr, 1.0f, 1.0f, 1.0f, 0.0f);
  cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
  cairo_paint (cr);

  kiba_paint_shape (memory->dock, cr, "rounded", "rounded", "rounded", "rounded",
                    FALSE, TRUE, x0, y0, x1, y1, 14.5, 0.5);

  convert_color (memory->dgconf->memory_color, &r, &g, &b);
  cairo_set_source_rgba (cr, r, g, b, memory->dgconf->memory_alpha);
  cairo_fill_preserve (cr);

  convert_color (memory->dgconf->memory_border_color, &r, &g, &b);
  cairo_set_source_rgba (cr, r, g, b, memory->dgconf->memory_border_alpha);
  cairo_set_line_width (cr, memory->dgconf->memory_border_width);

  cairo_stroke (cr);

  cairo_destroy (cr);

  return new_surface;
}

static void
kiba_memory_disable (void)
{
  GConfClient *gconf_client;
  GError *error = NULL;

  gconf_client = gconf_client_get_default ();

  if (gconf_client_set_bool (gconf_client, MEMORY_GCONF_PATH "/memory_enable", FALSE, &error) == FALSE)
    fprintf (stderr, "failed to disable the memory plugin\n");

  if (error != NULL)
  {
    fprintf (stderr, "%s\n", error->message);
    g_error_free (error);
  }
}

static gboolean
kiba_memory_time_handler (KibaDock *dock)
{
  KibaMemory  *memory;
  glibtop_mem  ram_memory;

  memory = kiba_get_memory (dock);
  if (dock->num_objects == 0 || memory == NULL)
    return FALSE;

/*
  glibtop_cpu cpu;
  glibtop_get_cpu (&cpu);
printf("CPU TYPE INFORMATIONS \n\n"
"Cpu Total : %ld \n"
"Cpu User : %ld \n"
"Cpu Nice : %ld \n"
"Cpu Sys : %ld \n"
"Cpu Idle : %ld \n"
"Cpu Frequences : %ld \n",
(unsigned long)cpu.total,
(unsigned long)cpu.user,
(unsigned long)cpu.nice,
(unsigned long)cpu.sys,
(unsigned long)cpu.idle,
(unsigned long)cpu.frequency);
*/

  if (!dock->hiding)
  {
    glibtop_get_mem (&ram_memory);

    if ((unsigned long) (ram_memory.user/(1024*1024))
        != memory->free_memory)
    {
      memory->free_memory = (unsigned long) ram_memory.user /(1024*1024);
      memory->free_memory_percent = ((double) memory->free_memory * 1024.0 * 1024.0) / (double) ram_memory.total;

      memory->object->name = g_strdup_printf ("%ld MB of %ld MB used",
        memory->free_memory, (unsigned long) ram_memory.total/(1024*1024));

      kiba_text_update_pango_for_object (memory->object, dock->gconf, dock->text);

      memory->object->rerender = TRUE;
    }
  }

  return TRUE;
}
