/*
 * Taskbar Plugin for Kiba-Dock
 * written by Daniel Bessler @ 2007
 */

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkwindow.h>
#include <glib/gstdio.h>
#include <gconf/gconf-client.h>
#include <cairo.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

#include "../config.h"
#undef HAVE_GLITZ

#ifdef HAVE_SVG
  #include <librsvg/rsvg.h>
  #include <librsvg/rsvg-cairo.h>
#endif

#ifdef HAVE_GLITZ
  #include <gdk/gdkx.h>
  #include <glitz-glx.h>
  #include <cairo-glitz.h>
#endif

#include <akamaru.h>

#include <kiba-dock.h>
#include <kiba-akamaru.h>
#include <kiba-render.h>
#include <kiba-gconf.h>
#include <kiba-object.h>

#define TIMEOUT_INTERVAL  1000
#define TASKBAR_GCONF_PATH KIBA_GCONF_PATH"/plugins/taskbar"

typedef struct _KibaTaskbar      KibaTaskbar;
typedef struct _KibaTaskbarIcon  KibaTaskbarIcon;

struct _KibaTaskbarIcon {
  KibaObject  *object;
  GdkWindow   *app_window;
  GdkPixbuf   *icon_pixbuf;

  double       blend_alpha;
  gboolean     iconify;       /* stores if app_window is iconified */
  gboolean     pixbuf_loaded;
};

struct _KibaTaskbar {
  KibaDock    *dock;
  GList       *taskbar_icons; /* list of all KibaTaskbarIcon's */

  gint         timeout_id;
};

void
kiba_plugin_init                (KibaDock *dock, GError *error);
char*
kiba_plugin_name                (void);
void
kiba_plugin_paint               (KibaDock *dock, KibaObject *object, cairo_t *cairo_context);
void
kiba_plugin_timeout             (KibaDock *dock, KibaObject *object, GError *error);
void
kiba_plugin_notify              (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data);
void
kiba_plugin_menu                (KibaDock *dock, KibaObject *object, GtkMenu *menu);
void
kiba_plugin_release_event       (KibaDock *dock, KibaObject *object);

static KibaTaskbarIcon*
kiba_taskbar_icon_new            (KibaDock *dock);
static void
kiba_taskbar_add_icon            (KibaDock *dock, KibaTaskbar *taskbar, KibaTaskbarIcon *taskbar_icon, int x, int y, GError *error);
static gboolean
kiba_taskbat_disable             (GtkWidget *widget, gpointer data);
static KibaTaskbarIcon *
kiba_taskbar_get_icon            (KibaObject *object);
static void
kiba_taskbar_disable             (void);
static gboolean
kiba_taskbar_time_handler        (KibaDock *dock);
static void
kiba_taskbar_get_window_pixbuf   (Display *display, KibaDock *dock, KibaTaskbarIcon *taskbar_icon, gint width, gint height, GError *error);
static KibaTaskbar*
kiba_taskbar_get                 (KibaDock *dock);
static void
kiba_taskbar_icon_update_surface (KibaDock *dock, KibaObject *object, gboolean background);
static gboolean
kiba_taskbar_skip_app            (Display *display, KibaDock *dock, GdkWindow *window);
static char*
kiba_taskbar_get_wintype         (Display *display, GdkWindow *window);
static gboolean
kiba_taskbar_skip_hint           (Display *display, Window window);
static gboolean
kiba_taskbar_skip_hints          (Display *display, Window window);
static void
kiba_taskbar_reload_mindow       (GtkWidget* widget, gpointer data);

/* add a listicon for every window in the screen window stack wich has the type hint Normal */
void
kiba_plugin_init (KibaDock *dock, GError *error)
{
  KibaPlugin  *plugin;
  KibaTaskbar *taskbar;
  GdkScreen   *screen;
  GList       *root_stack, *l;
  Display     *display;

  if (!dgconf->taskbar_enable)
    return;

  plugin  = g_new0 (KibaPlugin, 1);
  taskbar = g_new0 (KibaTaskbar, 1);
  if (plugin == NULL || taskbar == NULL)
    return;

  plugin->name           = g_strdup ("taskbar");
  plugin->plugin_struct  = taskbar;

  taskbar->dock = dock;

  screen     = gdk_screen_get_default ();
  root_stack = gdk_screen_get_window_stack (screen);

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  for (l = root_stack; l != NULL; l = l->next)
  {
    GdkWindow  *window = l->data;
    gchar      *state;

    if (window == NULL || kiba_taskbar_skip_app (display, dock, window))
      continue;

    /* skip apps with the _NET_WM_SKIP_TASKBAR hint */
    if (kiba_taskbar_skip_hints (display, gdk_x11_drawable_get_xid (window)))
      continue;

    state = kiba_taskbar_get_wintype (display, window);

    /* adding newly created window */
    if (state != NULL && g_strrstr (state, "_NET_WM_WINDOW_TYPE_NORMAL"))
    {
      KibaTaskbarIcon *taskbar_icon;

      taskbar_icon = kiba_taskbar_icon_new (dock);

      if (taskbar_icon != NULL)
      {
        taskbar_icon->object->plugin = plugin;
        taskbar_icon->app_window     = window;

        kiba_taskbar_add_icon (dock, taskbar, taskbar_icon, g_random_int_range (1, dock->geometry.width),
                               g_random_int_range (1, dock->geometry.height), error);
      }
    }
  }
  g_list_free(root_stack);

  XCloseDisplay (display);

  dock->plugins = g_list_append (dock->plugins, plugin);

  taskbar->timeout_id =
    g_timeout_add (TIMEOUT_INTERVAL, (GSourceFunc) kiba_taskbar_time_handler, (gpointer) dock);

  return;
}

char*
kiba_plugin_name (void)
{
  return "taskbar";
}

void
kiba_plugin_paint (KibaDock *dock, KibaObject *object, cairo_t *cairo_context)
{
  KibaTaskbarIcon *taskbar_icon;
  double           r, g, b, offset;

  if (!dgconf->taskbar_enable || dock->hiding)
    return;

  taskbar_icon = kiba_taskbar_get_icon (object);
  if (taskbar_icon == NULL)
    return;

  cairo_save (cairo_context);

  cairo_set_operator (cairo_context, CAIRO_OPERATOR_SOURCE);

  /* translate context for objects with width != height */
  if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
    cairo_translate (cairo_context,
        dock->d->valignment/2*(object->width < object->height ? (object->height - object->width) : 0.0),
        (object->width > object->height ? ((object->width - object->height) / 2.0) : 0.0));
  else
    cairo_translate (cairo_context,
        (object->width < object->height ? ((object->height - object->width) / 2.0) : 0.0),
         dock->d->valignment/2*(object->width > object->height ? (object->width - object->height) : 0.0));

  if (object->rotate_angle != 0.0)
    offset = -object->normal_size / 2.0;
  else
    offset =  .0f;

  /* draw bg surface */
  if (object->background_surface != NULL)
  {
    cairo_set_operator (cairo_context, CAIRO_OPERATOR_SOURCE);
    cairo_set_source_surface (cairo_context, object->background_surface, offset, offset);
    cairo_paint (cairo_context);
  }

  /* draw fg surface (little rectangle at the top) */
  if (object->foreground_surface != NULL && taskbar_icon->iconify)
  {
    cairo_set_operator (cairo_context, CAIRO_OPERATOR_ATOP);
    cairo_set_source_surface (cairo_context, object->foreground_surface, offset, offset);
    cairo_paint (cairo_context);
  }

  /* draw prelight */
  if (object->prelight || taskbar_icon->blend_alpha != 0.0 ||
     (taskbar_icon->iconify && dgconf->taskbar_inactive_icon_blend))
  {
    if (object->prelight)
    {
      convert_color (dgconf->prelight_color, &r, &g, &b);
      cairo_set_source_rgba (cairo_context, r, g, b, dgconf->prelight_alpha);

      taskbar_icon->blend_alpha = 0.0;
    }
    else
    {
      convert_color (dgconf->taskbar_inactive_blend_color, &r, &g, &b);
      cairo_set_source_rgba (cairo_context, r, g, b, taskbar_icon->blend_alpha);

      if (taskbar_icon->iconify)
        taskbar_icon->blend_alpha += 0.005;
      else
        taskbar_icon->blend_alpha -= 0.005;
    }
    cairo_set_operator (cairo_context, CAIRO_OPERATOR_ATOP);
    cairo_paint (cairo_context);
  }

  cairo_restore (cairo_context);
}

void
kiba_plugin_timeout (KibaDock *dock, KibaObject *object, GError *error)
{
  KibaTaskbarIcon *icon;
  GList *l;

  if (!dgconf->launcher_enable || object->window == NULL)
    return;

  icon = kiba_taskbar_get_icon (object);

  if (object->bgsurface_needs_update)
    kiba_taskbar_icon_update_surface (dock, object, TRUE);

  if (object->fgsurface_needs_update)
    kiba_taskbar_icon_update_surface (dock, object, FALSE);

  if (object->drawn_rotate_angle != 0.0 || object->drawn_rotate_angle != object->rotate_angle ||
     (icon->iconify && icon->blend_alpha < dgconf->taskbar_inactive_blend_alpha) ||
     (!icon->iconify && icon->blend_alpha > 0.0))
    object->rerender = TRUE;
}

void
kiba_plugin_notify (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
  KibaDock *dock;
  GError   *error;

  dock = user_data;

  set_option (dock, entry);

  if (strcmp (entry->key, TASKBAR_GCONF_PATH "/taskbar_icon_size") == 0)
    kiba_reload (dock);
  else if (strcmp (entry->key, TASKBAR_GCONF_PATH "/taskbar_enable") == 0)
  {
    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_delete_rope_spring (dock);
      kiba_delete_rope_strings (dock);
    }

    if (dgconf->taskbar_enable == FALSE)
    {
      KibaTaskbar *taskbar;
      GList       *taskbar_icons = NULL, *l;

      taskbar = kiba_taskbar_get (dock);

      DO_FOR_ALL_OBJECTS
      (
        if (g_strrstr (object->plugin->name, "taskbar"))
          taskbar_icons = g_list_append (taskbar_icons, kiba_taskbar_get_icon (object));
      }

      for (l = taskbar_icons; l != NULL; l = l->next)
      {
        KibaTaskbarIcon *taskbar_icon = l->data;

        taskbar->taskbar_icons = g_list_remove (taskbar->taskbar_icons, taskbar_icon);

        if (g_list_find (dock->plugins, taskbar_icon->object->plugin) != NULL)
          dock->plugins = g_list_remove (dock->plugins, taskbar_icon->object->plugin);

        kiba_object_free (dock, taskbar_icon->object);
        dock->objects = g_list_remove (dock->objects, taskbar_icon->object);
        taskbar_icon->object = NULL;
        dock->num_objects--;

        g_free (taskbar_icon);
      )

      g_source_remove (taskbar->timeout_id);

      g_free (taskbar);

      kiba_reload_spacers (dock);

      kiba_update_object_positions (dock);
    }
    else
      kiba_plugin_init (dock, error);

    if (strcmp (dgconf->object_model, "rope") == 0)
    {
      kiba_add_rope_spring (dock, 1.0);
      kiba_add_rope_strings (dock);
    }

    kiba_layout (dock);
  }
  else if (g_strrstr (entry->key, TASKBAR_GCONF_PATH "/taskbar_border_"))
  {
    DO_FOR_ALL_OBJECTS
    (
      if (g_strrstr (object->plugin->name, "taskbar"))
      {
        object->bgsurface_needs_update = TRUE;
        object->fgsurface_needs_update = TRUE;
      }
    )
  }
  else if (g_strrstr (entry->key, TASKBAR_GCONF_PATH "/taskbar_icon_") ||
           g_strrstr (entry->key, TASKBAR_GCONF_PATH "/taskbar_inactive_icon"))
  {
    DO_FOR_ALL_OBJECTS
    (
      if (strcmp (object->plugin->name, "taskbar") == 0)
         object->bgsurface_needs_update = TRUE;
    );
  }
  else if (g_strrstr (entry->key, TASKBAR_GCONF_PATH "/taskbar_inactive_blend"))
  {
    DO_FOR_ALL_OBJECTS
    (
      if (strcmp (object->plugin->name, "taskbar") == 0)
         object->rerender = TRUE;
    );
  }
  else if (strcmp (entry->key, TASKBAR_GCONF_PATH "/taskbar_skip_app") != 0)
  {
    fprintf (stderr, "ignoring gconf event for %s\n", entry->key);
    return;
  }
}

void
kiba_plugin_menu (KibaDock *dock, KibaObject *object, GtkMenu *menu)
{
  GtkWidget        *disable_item;
  KibaTaskbarIcon  *taskbar_icon = NULL;
  GtkWidget        *sparator, *change_state, *kill_item, *reload_item;
  Display          *display;
  int               win_state;

  if (!g_strrstr (object->plugin->name, "taskbar"))
    return;

  taskbar_icon = kiba_taskbar_get_icon (object);
  g_return_if_fail (taskbar_icon != NULL);

  disable_item = gtk_menu_item_new_with_label ("Disable Taskbar");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), disable_item);
  g_signal_connect (G_OBJECT (disable_item), "activate", G_CALLBACK (kiba_taskbar_disable), NULL);
  gtk_widget_show (disable_item);

  kill_item = gtk_menu_item_new_with_label ("Close Programm");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), kill_item);
  g_signal_connect (G_OBJECT (kill_item), "activate", G_CALLBACK (kiba_object_kill_window), taskbar_icon->app_window);
  gtk_widget_show (kill_item);

  reload_item = gtk_menu_item_new_with_label ("Reload Mindow");
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), reload_item);
  g_signal_connect (G_OBJECT (reload_item), "activate", G_CALLBACK (kiba_taskbar_reload_mindow), taskbar_icon);
  gtk_widget_show (reload_item);

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  win_state = kiba_object_get_winstate (display, taskbar_icon->app_window);

  if (win_state == 3)
  {
    change_state = gtk_menu_item_new_with_label ("Restore Window");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), change_state);
    g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_object_show_window), taskbar_icon->app_window);
    gtk_widget_show (change_state);
  }
  else if (win_state == 1 && gdk_x11_drawable_get_xid (taskbar_icon->app_window) == kiba_object_get_active_window (display))
  {
    change_state = gtk_menu_item_new_with_label ("Iconify Window");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), change_state);
    g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_object_iconify_window), taskbar_icon->app_window);
    gtk_widget_show (change_state);
  }
  else if (win_state == 1)
  {
    change_state = gtk_menu_item_new_with_label ("Focus Window");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), change_state);
    g_signal_connect (G_OBJECT (change_state), "activate", G_CALLBACK (kiba_object_focus_window), taskbar_icon->app_window);
    gtk_widget_show (change_state);
  }

  XCloseDisplay (display);
}

void
kiba_plugin_release_event (KibaDock *dock, KibaObject *object)
{
  KibaTaskbarIcon  *taskbar_icon = NULL;
  int               win_state, focus_revert_to;
  Display          *display;
  Window            focus_window;

  if (!g_strrstr (object->plugin->name, "taskbar"))
    return;

  taskbar_icon = kiba_taskbar_get_icon (object);
  g_return_if_fail (taskbar_icon != NULL);

  dock->exec_object = object;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  win_state = kiba_object_get_winstate (display, taskbar_icon->app_window);

  XGetInputFocus (display, &focus_window, &focus_revert_to);

  /* show the window and focus it */
  if (win_state == 3)
    kiba_object_show_window (GTK_WIDGET (dock), taskbar_icon->app_window);
  /* iconify the window */
  else if (win_state == 1 && gdk_x11_drawable_get_xid (taskbar_icon->app_window) == focus_window)
    kiba_object_iconify_window (GTK_WIDGET (dock), taskbar_icon->app_window);
  /* focus window */
  else if (win_state == 1)
    kiba_object_focus_window (GTK_WIDGET (dock), taskbar_icon->app_window);

  XCloseDisplay (display);
}

static KibaTaskbarIcon *
kiba_taskbar_icon_new (KibaDock *dock)
{
  KibaTaskbarIcon *taskbar_icon;
  GError          *error = NULL;

  taskbar_icon                          = g_new0 (KibaTaskbarIcon, 1);
  taskbar_icon->object                  = g_new0 (KibaObject, 1);
//   taskbar_icon->object->position_notify = TRUE;

  return taskbar_icon;
}

static void
kiba_taskbar_add_icon (KibaDock *dock, KibaTaskbar *taskbar, KibaTaskbarIcon *taskbar_icon, int x, int y, GError *error)
{
  AddSpacerClosure closure;
  Window           window;
  Display         *display;
  char            *window_name;
  int              app_x, app_y, width, height, depth,
                   win_state, window_workspace_number, active_workspace_number;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return;
  }

  window = gdk_x11_drawable_get_xid (taskbar_icon->app_window);

  closure.dock           = dock;
  closure.object         = akamaru_model_add_object (&dock->model, x, y, 12, NULL);
  closure.spacer_size    = (double) dgconf->taskbar_icon_size + dgconf->object_space;
  closure.object->radius = closure.spacer_size / 2.0;

  taskbar_icon->object->spring =
    akamaru_model_add_spring (&dock->model, dock->anchor->object, closure.object, 0);
  dock->objects         = g_list_append (dock->objects, taskbar_icon->object);
  dock->num_objects++;

  akamaru_model_for_each_object (&dock->model, add_spacer, &closure);

  taskbar_icon->object->object           = closure.object;
  taskbar_icon->object->object->friction = dock->model.friction;
  taskbar_icon->object->win_position.x   = taskbar_icon->object->object->position.x;
  taskbar_icon->object->win_position.y   = taskbar_icon->object->object->position.y;
  taskbar_icon->object->normal_size      = dgconf->taskbar_icon_size;
  taskbar_icon->object->scale_factor     = 1.0;
  taskbar_icon->blend_alpha              = 0.0;
  taskbar_icon->object->key              = g_strdup (TASKBAR_GCONF_PATH);

  if (0 != XFetchName (display, window, &window_name))
    taskbar_icon->object->name = window_name;
  else
    taskbar_icon->object->name = kiba_object_get_wmclass (display, window);

  gdk_window_get_geometry (taskbar_icon->app_window, &app_x, &app_y, &width, &height, &depth);

//   kiba_object_update_position (dock, taskbar_icon->object, reload);

  if (GTK_WIDGET_REALIZED (GTK_WIDGET (dock)))
    kiba_object_create_window (dock, taskbar_icon->object);

  if (GTK_WIDGET_MAPPED (GTK_WIDGET (dock)))
    gdk_window_show (taskbar_icon->object->window);

  win_state = kiba_object_get_winstate (display, taskbar_icon->app_window);

  /* check if window is on active workspace */
  window_workspace_number = kiba_get_win_workspace_number (display, window);
  active_workspace_number = kiba_get_active_workspace_number (display);

  if (win_state != 3 && win_state != 0 && window_workspace_number == active_workspace_number)
    taskbar_icon->pixbuf_loaded = TRUE;

  kiba_taskbar_get_window_pixbuf (display, dock, taskbar_icon, width, height, error);

  XCloseDisplay (display);

  taskbar_icon->object->buffer_width  = width;
  taskbar_icon->object->buffer_height = height;
  if (width > height)
  {
    taskbar_icon->object->width = dgconf->taskbar_icon_size;
    taskbar_icon->object->height = dgconf->taskbar_icon_size * height / width;
  }
  else
  {
    taskbar_icon->object->width = dgconf->taskbar_icon_size * width / height;
    taskbar_icon->object->height = dgconf->taskbar_icon_size;
  }

//TODO remove object->size
taskbar_icon->object->size     = dgconf->taskbar_icon_size;
taskbar_icon->object->buffer_size = (double) taskbar_icon->object->size;

  taskbar->taskbar_icons = g_list_append (taskbar->taskbar_icons, taskbar_icon);

  if (taskbar_icon->object->bg_pixbuf != NULL)
    kiba_taskbar_icon_update_surface (dock, taskbar_icon->object, TRUE);

  kiba_taskbar_icon_update_surface (dock, taskbar_icon->object, FALSE);

  kiba_layout (dock);
}

static void
kiba_taskbar_get_window_pixbuf (Display *display, KibaDock *dock, KibaTaskbarIcon *taskbar_icon, gint width, gint height, GError *error)
{
  GdkScreen      *screen;
  GdkColormap    *colormap;

  screen   = gdk_screen_get_default ();
  colormap = gdk_drawable_get_colormap (taskbar_icon->app_window);

//     taskbar_icon->pixbuf_loaded = FALSE;

  Window    xwindow;
  char     *icon_path = NULL;

  xwindow  = gdk_x11_drawable_get_xid (taskbar_icon->app_window);

  icon_path = kiba_object_get_wmclass (display, xwindow);

  if (icon_path != NULL)
  {
    kiba_load_png (dock, taskbar_icon->object, icon_path, &taskbar_icon->icon_pixbuf, &error);

    g_free (icon_path);
  }

  /* load fallback icon */
  if (taskbar_icon->icon_pixbuf == NULL && dgconf->taskbar_fallback_icon != NULL)
   kiba_load_png (dock, taskbar_icon->object, dgconf->taskbar_fallback_icon, &taskbar_icon->icon_pixbuf, &error);

  /* load pixbuf from window, i know this is ressource hungry, a better way is needed */
  taskbar_icon->object->bg_pixbuf =
    gdk_pixbuf_get_from_drawable (NULL, taskbar_icon->app_window, colormap, 0, 0, 0, 0, width, height);

/*
   int win_state = kiba_object_get_winstate (display, taskbar_icon->app_window);

   if (win_state != 3 && win_state != 0 && !taskbar_icon->pixbuf_loaded)
   {
     kiba_object_iconify_window (GTK_WIDGET (dock), taskbar_icon->app_window);
     taskbar_icon->pixbuf_loaded = TRUE;
   }
*/

/*
  Atom actual;
  int result, format, i;
  unsigned long n, left;
  unsigned long *data;
  Atom icon_atom = XInternAtom (display, "_NET_WM_ICON", TRUE);

  result = XGetWindowProperty (display, xwindow, icon_atom, 0L, 65536L, FALSE, XA_CARDINAL, &actual,
                               &format, &n, &left, (unsigned char **)&data);

  if (result == Success && n)
  {
    int iw, ih, j;

    for (i = 0; i + 2 < n; i += iw * ih + 2)
    {
      iw = data[i];
      ih = data[i + 1];

      if (iw * ih + 2 > n - i)
        break;

      if (iw && ih)
      {
        printf ("got icon for %s\n", taskbar_icon->object->name);
      }
    }
  XFree(data);
  }
*/
}

static void
kiba_taskbar_icon_update_surface (KibaDock *dock, KibaObject *object, gboolean background)
{
  KibaTaskbarIcon *taskbar_icon;
  cairo_surface_t* pNewSurface;
  cairo_t *drawing_context, *cr = NULL;
  gint     x0, y0, x1, y1;
  double   r, g, b;

  cr = dock_cairo_create (object);
  if (cr == NULL)
    return;

  pNewSurface = cairo_surface_create_similar (cairo_get_target (cr), CAIRO_CONTENT_COLOR_ALPHA, object->size, object->size);

  if (cairo_surface_status (pNewSurface) != CAIRO_STATUS_SUCCESS)
    return;

  drawing_context = cairo_create (pNewSurface);
  if (cairo_status (drawing_context) != CAIRO_STATUS_SUCCESS)
    return;

  taskbar_icon = kiba_taskbar_get_icon (object);

  cairo_set_source_rgba (drawing_context, 1.0f, 1.0f, 1.0f, 0.0f);
  cairo_set_operator (drawing_context, CAIRO_OPERATOR_OVER);
  cairo_paint (drawing_context);

  cairo_set_operator (drawing_context, CAIRO_OPERATOR_SOURCE);

  /* draw stuff */
  if (background)
  {
    #define border_width dgconf->taskbar_border_width
    x0 = border_width;
    x1 = object->width - border_width;
    y0 = object->height - border_width;
    y1 = border_width;

    cairo_save (drawing_context);
      kiba_paint_shape (dock, drawing_context, "rounded", "rounded", "rounded", "rounded",
                        FALSE, TRUE, x0, y0, x1, y1, dgconf->taskbar_border_radius, 0.5);
      cairo_set_source_rgba (drawing_context, 1.0, 1.0, 1.0, dgconf->taskbar_icon_alpha);
      cairo_fill_preserve (drawing_context);
      cairo_stroke (drawing_context);

      cairo_set_operator (drawing_context, CAIRO_OPERATOR_ATOP);

      if (object->width > object->height)
        cairo_scale (drawing_context, object->width / object->buffer_width,
                                      object->width / object->buffer_width);
      else
        cairo_scale (drawing_context, object->height / object->buffer_height,
                                      object->height / object->buffer_height);
      gdk_cairo_set_source_pixbuf (drawing_context, object->bg_pixbuf, 0, 0);
      cairo_paint (drawing_context);
    cairo_restore (drawing_context);

    kiba_paint_shape (dock, drawing_context, "rounded", "rounded", "rounded", "rounded",
                      FALSE, TRUE, x0 - border_width / 2.0, y0 + border_width / 2.0, x1 + border_width / 2.0, y1 - border_width / 2.0, dgconf->taskbar_border_radius, 0.5);
    convert_color (dgconf->taskbar_border_color, &r, &g, &b);
    cairo_set_source_rgba (drawing_context, r, g, b, dgconf->taskbar_border_alpha);
    cairo_set_line_width (drawing_context, border_width);
    cairo_stroke (drawing_context);

    /* load icon on surface */
    KibaTaskbarIcon *icon = kiba_taskbar_get_icon (object);
    if (dgconf->taskbar_inactive_icon_blend && icon != NULL && icon->icon_pixbuf != NULL)
    {
      cairo_save (drawing_context);
        cairo_set_operator (drawing_context, CAIRO_OPERATOR_ATOP);

        double offset = dgconf->taskbar_icon_scale_factor *
            (object->height > object->width ? object->height : object->width);

        if (g_strrstr (dgconf->taskbar_icon_edge, "top_right"))
          cairo_translate (drawing_context, object->width - offset, .0);
        if (g_strrstr (dgconf->taskbar_icon_edge, "bottom_right"))
          cairo_translate (drawing_context, object->width  - offset, object->height - offset);
        if (g_strrstr (dgconf->taskbar_icon_edge, "bottom_left"))
          cairo_translate (drawing_context, 0.0, object->height - offset);

        cairo_scale (drawing_context, dgconf->taskbar_icon_scale_factor, dgconf->taskbar_icon_scale_factor);
        gdk_cairo_set_source_pixbuf (drawing_context, icon->icon_pixbuf, 0, 0);
        cairo_paint (drawing_context);
      cairo_restore (drawing_context);
    }

    #undef border_width
  }
  else /* paint foreground */
  {
    #define radius 2.0
    convert_color (dgconf->taskbar_border_color, &r, &g, &b);
    cairo_set_source_rgba (drawing_context, r, g, b, dgconf->taskbar_border_alpha);
    cairo_set_line_width (drawing_context, 1.0);

    cairo_move_to (drawing_context, object->size / 2.0 - 10.0, 0.0);
    cairo_rel_line_to (drawing_context, 10.0 - radius, 10.0 - radius);
    cairo_rel_line_to (drawing_context, 4.0, 0.0);
    cairo_rel_line_to (drawing_context, 10.0 - radius, -10.0 + radius);
    cairo_rel_line_to (drawing_context, -10.0, 0.0);
    cairo_set_line_join (drawing_context, CAIRO_LINE_JOIN_ROUND);

    cairo_fill_preserve (drawing_context);

    cairo_stroke (drawing_context);

    #undef radius
  }

  cairo_destroy (drawing_context);

  if (background)
  {
    object->bgsurface_needs_update = FALSE;

    if (object->background_surface)
    {
      cairo_surface_destroy (object->background_surface);
      object->background_surface = NULL;
    }

    object->background_surface = pNewSurface;
  }
  else
  {
    object->fgsurface_needs_update = FALSE;

    if (object->foreground_surface)
    {
      cairo_surface_destroy (object->foreground_surface);
      object->foreground_surface = NULL;
    }

    object->foreground_surface = pNewSurface;
  }

  object->rerender = TRUE;

  cairo_destroy (cr);
}

static gboolean
kiba_taskbar_skip_app (Display *display, KibaDock *dock, GdkWindow *window)
{
  GSList            *skip_list;
  Window             xwindow;
  XTextProperty      text_property;
  Atom               cmd_atom;

  xwindow = gdk_x11_drawable_get_xid (window);

  /* skip apps in exclude list */
  cmd_atom = XInternAtom (display, "WM_CLASS", TRUE);
  if (XGetTextProperty (display, xwindow, &text_property, cmd_atom))
  {
    for (skip_list = dgconf->taskbar_skip_app; skip_list != NULL; skip_list = skip_list->next)
    {
      char *skip_app = skip_list->data;

      if (g_strrstr (text_property.value, skip_app))
        return TRUE;
    }
  }

  return FALSE;
}

static char*
kiba_taskbar_get_wintype (Display *display, GdkWindow *window)
{
  Window              xwindow;
  Atom                type_return, state_atom, *type_atom;
  int                 format_return;
  unsigned long int   nitems_return;
  unsigned long int   bytes_after_return;
  unsigned char      *state;
  char               *state_string = NULL;

  xwindow    = gdk_x11_drawable_get_xid (window);
  state_atom = XInternAtom (display, "_NET_WM_WINDOW_TYPE", TRUE);

  gdk_error_trap_push();
  if (XGetWindowProperty (display, xwindow, state_atom, 0, G_MAXLONG, False, AnyPropertyType,
       &type_return, &format_return, &nitems_return, &bytes_after_return, &state) == Success)
  {
    if (state != NULL && state != None)
    {
      type_atom = (Atom *)state;
      if (type_atom)
        state_string = XGetAtomName (display, *type_atom);
    }
  }

  return state_string;
}

static gboolean
kiba_taskbar_skip_hints (Display *display, Window window)
{
  Atom type, wm_state_atom, skip_atom;
  gint format;
  gulong nitems;
  gulong bytes_after;
  GdkAtom *atoms = NULL;
  gulong i;

  wm_state_atom = XInternAtom (display, "_NET_WM_STATE", FALSE);

  XGetWindowProperty (display, window, wm_state_atom, 0, G_MAXLONG, False, XA_ATOM,
            &type, &format, &nitems, &bytes_after, (guchar **)&atoms);
  if (type != None)
  {
      skip_atom = XInternAtom (display, "_NET_WM_STATE_SKIP_TASKBAR", FALSE);

      i = 0;
      while (i < nitems)
      {
        if (g_strrstr (XGetAtomName (display, skip_atom), XGetAtomName (display, (Atom) atoms[i])))
          return TRUE;

        ++i;
      }

    XFree (atoms);
  }


  return FALSE;
}

static void
kiba_taskbar_disable (void)
{
  GConfClient *gconf_client;
  GError *error = NULL;

  gconf_client = gconf_client_get_default ();

  if (gconf_client_set_bool (gconf_client, TASKBAR_GCONF_PATH "/taskbar_enable", FALSE, &error) == FALSE)
    fprintf (stderr, "failed to disable the taskbar\n");

  if (error != NULL)
  {
    fprintf (stderr, "%s\n", error->message);
    g_error_free (error);
  }
}

static void
kiba_taskbar_reload_mindow (GtkWidget* widget, gpointer data)
{
  KibaTaskbarIcon *taskbar_icon = (KibaTaskbarIcon*) data;

  if (taskbar_icon == NULL)
    return;

  taskbar_icon->pixbuf_loaded = FALSE;
}

static KibaTaskbarIcon *
kiba_taskbar_get_icon (KibaObject *object)
{
  KibaTaskbar *taskbar;
  GList       *l;

  taskbar = object->plugin->plugin_struct;

  for (l = taskbar->taskbar_icons; l != NULL; l = l->next)
  {
    KibaTaskbarIcon *taskbar_icon = l->data;

    if (taskbar_icon->object == object)
      return taskbar_icon;
  }

  return NULL;
}

static KibaTaskbar*
kiba_taskbar_get (KibaDock *dock)
{
  GList *l;

  for (l = dock->plugins; l != NULL; l = l->next)
  {
    KibaPlugin *plugin = l->data;

    if (g_strrstr (plugin->name, "taskbar"))
      return plugin->plugin_struct;
  }

  return NULL;
}

static gboolean
kiba_taskbar_time_handler (KibaDock *dock)
{
  GdkScreen          *screen;
  GList              *root_stack, *taskbar_windows = NULL, *l;
  GError             *error = NULL;
  Display            *display;

  if (!dgconf->taskbar_enable)
    return TRUE;

  display = XOpenDisplay (NULL);
  if (display == NULL){
      fprintf (stderr, "Unable to open display.\n");
      return TRUE;
  }

  screen     = gdk_screen_get_default ();
  root_stack = gdk_screen_get_window_stack (screen);

  if (root_stack == NULL)
    return TRUE;

  /* removing icons for not running programms */
  for (l = dock->objects; l != NULL; l = l->next)
  {
    KibaObject *object = l->data;

    if (g_strrstr (object->plugin->name, "taskbar"))
    {
      KibaTaskbarIcon *taskbar_icon = kiba_taskbar_get_icon (object);
      gboolean         exclude_win;

      exclude_win = kiba_taskbar_skip_app (display, dock, taskbar_icon->app_window);

      if (taskbar_icon != NULL && (g_list_find (root_stack, taskbar_icon->app_window) == NULL ||
          exclude_win))
      {
        g_object_unref (taskbar_icon->app_window);
        taskbar_icon->app_window = NULL;

        kiba_object_free (dock, taskbar_icon->object);
        dock->objects = g_list_remove (dock->objects, taskbar_icon->object);
        taskbar_icon->object = NULL;

        kiba_reload_spacers (dock);
        kiba_layout (dock);
      }
    }
  }

  /* fill taskbar_windows list */
  for (l = dock->objects; l != NULL; l = l->next)
  {
    KibaObject *object = l->data;

    if (g_strrstr (object->plugin->name, "taskbar"))
    {
      KibaTaskbarIcon *taskbar_icon = kiba_taskbar_get_icon (object);

      if (taskbar_icon->app_window != NULL)
      {
        int win_state, window_workspace_number, active_workspace_number;

        taskbar_windows = g_list_append (taskbar_windows, taskbar_icon->app_window);

        /* bounce when urgency flag is set */
        if (taskbar_icon->app_window != NULL && kiba_object_window_urgency (display, taskbar_icon->app_window))
        {
          if (dock->d->orientation == GTK_ORIENTATION_VERTICAL)
            taskbar_icon->object->object->previous_position.x += (dock->d->halignment - 1) * 1.0;
          else
            taskbar_icon->object->object->previous_position.y += (dock->d->valignment - 1) * 1.0;
        }

        win_state               = kiba_object_get_winstate (display, taskbar_icon->app_window);
        window_workspace_number = kiba_get_win_workspace_number (display, gdk_x11_drawable_get_xid (taskbar_icon->app_window));
        active_workspace_number = kiba_get_active_workspace_number (display);

        if (!taskbar_icon->pixbuf_loaded && win_state != 3 && win_state != 0 &&
            window_workspace_number == active_workspace_number)
        {
          gint app_x, app_y, width, height, depth;

          taskbar_icon->pixbuf_loaded = TRUE;

          gdk_window_get_geometry (taskbar_icon->app_window, &app_x, &app_y, &width, &height, &depth);

          kiba_taskbar_get_window_pixbuf  (display, dock, taskbar_icon, width, height, error);

          taskbar_icon->object->bgsurface_needs_update = TRUE;
        }

        /* repaint with foreground surface */
        if (win_state == 3 && !taskbar_icon->iconify)
        {
          taskbar_icon->iconify          = TRUE;
          taskbar_icon->object->rerender = TRUE;
        }
        /* repaint without foreground surface */
        else if (win_state != 3 && taskbar_icon->iconify)
        {
          taskbar_icon->iconify          = FALSE;
          taskbar_icon->object->rerender = TRUE;
        }
      }
    }
  }

  /* check for new windows in the window stack */
  for (l = root_stack; l != NULL; l = l->next)
  {
    GdkWindow          *window = l->data;
    gchar              *state;

    if (window == NULL)
      continue;

    state = kiba_taskbar_get_wintype (display, window);

    /* adding newly created window */
    if (state != NULL  && g_strrstr (state, "_NET_WM_WINDOW_TYPE_NORMAL") &&
        window != NULL && g_list_find (taskbar_windows, window) == NULL)
    {
      KibaTaskbar     *taskbar;
      KibaTaskbarIcon *taskbar_icon;

      if (kiba_taskbar_skip_app (display, dock, window) ||
          kiba_taskbar_skip_hints (display, gdk_x11_drawable_get_xid (window)))
        continue;

      taskbar = kiba_taskbar_get (dock);
      if (taskbar == NULL)
        continue;

      taskbar_icon = kiba_taskbar_icon_new (dock);
      if (taskbar_icon == NULL)
        continue;

      taskbar_icon->object->plugin = kiba_get_plugin_struct (dock, "taskbar");

      if (taskbar_icon != NULL)
      {
        taskbar_windows          = g_list_append (taskbar_windows, window);
        taskbar_icon->app_window = window;

        kiba_taskbar_add_icon (dock, taskbar, taskbar_icon, g_random_int_range (0, dock->geometry.width/2),
                               g_random_int_range (0, dock->geometry.height/2), error);

        taskbar_icon->object->plugin->plugin_struct = taskbar;
        taskbar->taskbar_icons = g_list_append (taskbar->taskbar_icons, taskbar_icon);
      }
    }
  }

  XCloseDisplay (display);

  g_list_free (root_stack);
  g_list_free (taskbar_windows);

  return TRUE;
}
